(* Enumeration of bintrees by height (EXPERIMENTAL) 
 *  TO BE UPDATED with the version of EnmBinTree.v
 *
 *       
 *
 *)

(* Study of 3 styles of proofs of the following "enum" function (written in ML):

type bintree = Leaf | Node of  bintree*bintree ;;

let rec enum print n =
  if n=0 then
     print Leaf
  else 
    let pn = n-1 in
      enum (fun l -> enum (fun r -> print (Node (l,r))) pn) pn ;
      enum (fun l -> enum_lt (fun r -> print (Node (l,r))) pn) pn ;
      enum_lt (fun l -> enum (fun r -> print (Node (l,r))) pn) pn
and enum_lt print n =
  if n>0 then (
     enum print (n-1) ;
     enum_lt print (n-1) 
  );;


let rec print_tree t = 
  match t with
  | Leaf -> ()
  | Node(l,r) -> 
      print_string "(" ; 
      print_tree l; 
      print_string ")" ; 
      print_tree r ;; 

let aux = ref 0 in
 enum (fun t -> aux:=!aux+1 ; print_tree t; print_newline()) 3;
 !aux;;

*)

Set Implicit Arguments.
 
Require Import ISLLoops.
Require Import Arith.
Require Import Max.
Require Import MyTheoryList.
Import Wf_nat.

Lemma nat_le_ind: forall (P:nat -> Prop),
  (P 0) -> 
   (forall (n:nat), 
       (forall (m:nat),
          m <= n -> P m)
        -> (P (S n)))
     -> forall (n:nat), P n.
Proof.
  intros P H1 H2 n. pattern n. apply lt_wf_ind.
  clear n; intros n; case n; auto with arith.
Qed.

Inductive bintree : Set := Leaf: bintree | Node: bintree -> bintree -> bintree.

Fixpoint height (t:bintree) { struct t } : nat :=
  match t with
  | Leaf => 0
  | (Node l r) => S (max (height l) (height r))   
  end.

Inductive is_height: bintree -> nat -> Prop :=
 | Leaf_heightO: (is_height Leaf O)
 | Node_height_both: 
     forall (l r: bintree) (n:nat),
       (is_height l n) -> (is_height r n) -> (is_height (Node l r) (S n))
 | Node_height_left: 
     forall (l r: bintree) (n m:nat),
       (is_height l n) -> (is_height r m) -> m < n -> (is_height (Node l r) (S n))
 | Node_height_right: 
     forall (l r: bintree) (n m:nat),
       (is_height l n) -> (is_height r m) -> n < m -> (is_height (Node l r) (S m)).

Hint Resolve Leaf_heightO Node_height_both Node_height_left Node_height_right: bintree.

Lemma is_height_from_height:
  forall (t:bintree), (is_height t (height t)).
Proof.
  induction t; simpl; auto with bintree.
  case (lt_eq_lt_dec (height t1) (height t2)).
  intros H1; case H1.
  intros H2; rewrite max_r; auto with arith.
  apply Node_height_right with (n:=(height t1)); auto.
  intros H2; rewrite H2; rewrite max_r; auto with arith.
  apply Node_height_both; auto.
  rewrite <- H2; auto.
  intros H2; rewrite max_l; auto with arith.
  apply Node_height_left with (m:=(height t2)); auto.
Qed.

Lemma is_height_to_height:
  forall (t:bintree) (n:nat), (is_height t n) -> (height t)=n.
Proof.
  induction 1; simpl; subst; auto.
  rewrite IHis_height2.
  rewrite max_l; auto with arith.
  rewrite max_l; auto with arith.
  rewrite max_r; auto with arith.
Qed.

Lemma height_discr: 
  forall (t1 t2:bintree) (n m:nat), 
    (is_height t1 n) -> (is_height t2 m) -> n <> m -> t1 <> t2.
Proof.
  intros t1 t2 n m H1 H2 H3 H4.
  subst.
  rewrite <- (is_height_to_height H2) in H3.
  rewrite <- (is_height_to_height H1) in H3.
  tauto.
Qed. 
 
Hint Resolve is_height_to_height is_height_from_height: bintree.


Fixpoint eqbintree (t1 t2:bintree) { struct t1 }: bool :=
  match t1,t2 with
  | Leaf,Leaf => true
  | (Node l1 r1),(Node l2 r2) => Bool.andb (eqbintree l1 l2) (eqbintree r1 r2)
  | _,_ => false
  end. 

Lemma eqbintree_case: forall (t1 t2:bintree), (eqbintree t1 t2)=true \/ (eqbintree t1 t2)=false.
Proof.
  intros t1 t2; case (eqbintree t1 t2); auto.
Qed.

Lemma eq_eqbintree: forall (t1 t2:bintree), t1=t2 -> (eqbintree t1 t2)=true.
Proof.
  induction t1; simpl. 
  intros t2 H; inversion H; auto.
  intros t2 H; inversion H; auto. 
  rewrite IHt1_1; auto.
  rewrite IHt1_2; auto.
Qed.

Lemma eqbintree_eq: forall (t1 t2:bintree), (eqbintree t1 t2)=true -> t1=t2.
Proof.
  induction t1; simpl; intros t2; case t2; simpl; auto;
  try (intros; discriminate).
  intros b b0.
  case (eqbintree_case t1_1 b); intros H1; rewrite H1; simpl.
  rewrite (IHt1_1 b); auto.
  intros; rewrite (IHt1_2 b0); auto.
  intros; discriminate.
Qed.

Lemma noteq_eqbintree: forall (t1 t2:bintree), t1<>t2 -> (eqbintree t1 t2)=false.
Proof.
  intros t1 t2; case (eqbintree_case t1 t2); auto.
  intros H1; rewrite (eqbintree_eq _ _ H1).
  intros H2; case H2; auto.
Qed.

Lemma eqbintree_noteq: forall (t1 t2:bintree), (eqbintree t1 t2)=false -> t1 <> t2.
Proof.
  intros t1 t2 H1 H2.
  rewrite (eq_eqbintree H2) in H1.
  discriminate.
Qed.

(*************************)

Fixpoint enumBT (St:Type) (print: bintree -> DS St unit) (n:nat) {struct n} : (DS St unit) :=
  match n with
  | 0 => print Leaf
  | (S n1) => 
     (enumBT (fun l => enumBT (fun r => print (Node l r)) n1) n1 -;
      enumBT (fun l => enumBT_lt (fun r => print (Node l r)) n1) n1 -;
      enumBT_lt (fun l => enumBT (fun r => print (Node l r)) n1) n1)%ISL
  end
with enumBT_lt (St:Type) (print: bintree -> DS St unit) (n:nat) {struct n} : (DS St unit) :=
  match n with
  | 0 => Skip%ISL 
  | (S n1) => (enumBT print n1 -; enumBT_lt print n1)%ISL
  end.

(* Pour eviter d'introduire une version fonctionnelle de enumBT
 * 
 * on pourrait montrer que
 *
 *  (enumBT n f) raffine 
 *     let l = ref [] in
 *       (enumBT n (fun t => l:=t::!l)) ;
 *          (List.revIter l f)
 *
 *  en plus du lemme enumbintree. => Pas si simple !!!

Definition insertHead (t:bintree) : DS ((list bintree)*unit) unit := 
   Bind l <- !pL In pL :<- t::l.

Lemma enumlt_HOelim_aux:
  forall (St: Type) (p: bintree -> DS St unit) ,
     forall (n:nat), 
      (forall (m:nat) (st:St) (acc:list bintree), m < n -> 
        [< st >] (listRevIter acc p) -; (enumBT m p) 
                   >==> (Bind l <-/ (block acc (enumBT m insertHead -; !pL)) In listRevIter l p))
             -> forall (st:St) (acc:list bintree), ([< st >] (listRevIter acc p) -; (enumlt n p) >==> 
                     (Bind l <-/ (block acc (enumlt n insertHead -; !pL)) In listRevIter l p)).
Proof.
  induction n; simpl; intros.
  ISL_simplify_right; intros; apply bind_val_r_ref1; auto with isl.
  intros a; case a; simpl; auto with isl.
  ISL_assoc_left.
  ISL_enter_left.
  apply H; auto with arith.
  ISL_enter_right.
  ISL_trans.
  2: (apply block_monotony_simpl; ISL_unassoc_right; auto with isl).
  apply block_seq_ref2.
  ISL_unassoc_right.
  ISL_unassoc_left.
  ISL_monotony.
Qed.

Hint Resolve enumlt_HOelim_aux: bintree.

Lemma enumBT_HOelim_aux:
  forall (n:nat) (St: Type) (p: bintree -> DS St unit) (st:St) (acc: list bintree), 
       ([< st >]  ((listRevIter acc p) -; (enumBT n p)) 
                   >==> (Bind l <-/ (block acc (enumBT n insertHead -; !pL)) In listRevIter l p))
       /\        ([< st >]  
                   (Bind l <-/ (block acc (enumBT n insertHead -; !pL)) In listRevIter l p)
                   >==> (listRevIter acc p) -; (enumBT n p)).

Proof.
   intros n. elim n using nat_le_ind; simpl; intros; clear n.
   constructor 1.
   ISL_simplify_right;  auto with isl.
   ISL_simplify_left;  auto with isl.
   (* ISL_enter_right. *)
   constructor 1.
   assert (H1: forall m : nat,
      m <= n0 ->
      forall (St: Type) (p : bintree -> DS St unit) (st : St) (acc : list bintree),
      ([< st >] listRevIter acc p -;  enumBT m p >==> 
       Bind l <-/ block acc (enumBT m insertHead -;  !pL)
       In listRevIter l p)).
   intros m H1 St0 p0 st0 acc0; case (H m H1 St0 p0 st0 acc0); auto.
   assert (H2: forall m : nat,
      m <= n0 ->
      forall (St:Type) (p : bintree -> DS St unit) (st : St) (acc : list bintree),
      ([< st >] 
       Bind l <-/ block acc (enumBT m insertHead -;  !pL)
       In listRevIter l p
       >==> listRevIter acc p -;  enumBT m p)).
   intros m H2 St0 p0 st0 acc0; case (H m H2 St0 p0 st0 acc0); auto.
   clear H.
   ISL_enter_right.   
   apply block_monotony_simpl.
   ISL_enter_right.
   ISL_trans.
   apply H2 with (m:=n0) (acc:=(nil (A:=bintree))) (p:=(fun l : bintree =>
    enumBT n0 (fun r : bintree => insertHead (Node l r)) -; 
    enumlt n0
      (fun r : bintree =>
       insertHead (Node l r) -;  insertHead (Node r l)))); auto with arith. 
   ISL_simplify_left; auto with isl.
   auto with isl.
   ISL_monotony_left.
   intros; ISL_trans.
   2: apply H1 with (m:=n0) (acc:=(nil (A:=bintree))) (p:=(fun l : bintree =>
      enumBT n0 (fun r : bintree => p (Node l r)) -; 
      enumlt n0 (fun r : bintree => p (Node l r) -;  p (Node r l)))); auto with arith.
   ISL_simplify_right; auto with isl.

(*** PAS SI SIMPLE !!! ***)
*)

Lemma enumBT_lt_repeat_enumBT:
  forall (St:Type) (n:nat) (print: bintree -> DS St unit) (st:St), 
             ([< st >]
               (enumBT_lt print n) >==> 
                  (repeat_invoke (fun m => m < n) (enumBT print)))%ISL.     
Proof.
  induction n; simpl.
  auto with isl.
  intros print st.
  ISL_trans.
  2: apply repeat_invoke_stutter_left.
  ISL_enter_right.
  apply repeat_invoke_one_left with (x:=n).
  auto with arith.
  ISL_monotony.
  intros; ISL_trans.
  eapply IHn.
  apply repeat_invoke_guard_monotony.
  intros; auto with arith.
Qed.

(* height of all enumerated trees is n *)
Lemma enumBT_height_spec:
  forall (St:Type) (n:nat) (print: bintree -> DS St unit) (st:St), 
          ([< st >] (enumBT print n) >==> 
               (repeat_invoke (fun t => is_height t n) print))%ISL.
Proof.
  intros St n. elim n using nat_le_ind; simpl; clear n.
  (* cas de base *)
  intros print st; simpl.
  ISL_trans.
  2:apply repeat_invoke_one_left with (x:=Leaf).
  auto with isl.
  auto with bintree.
  (* cas recursif *)
  intros n IHn; simpl.
  intros print st.
  ISL_trans.
  2: apply repeat_invoke_stutter_left.
  ISL_monotony.
    (* cas both *)
    ISL_trans.
    apply IHn; auto with arith.
    ISL_trans.
    2: apply repeat_invoke_stutter_in_left.
    unfold repeat_invoke at 1.
    apply repeat_monotony.
    intros.
    ISL_simplify_left.
    intros l; ISL_simplify_left.
    intros; ISL_trans.
    apply IHn; auto with arith.
    unfold repeat_invoke.
    apply repeat_monotony.
    intros; ISL_simplify_left.
    intros r.
    ISL_simplify_right.
    intros; constructor 1 with (x:=(Node l r)).
    ISL_monotony.
    ISL_simplify_all.
    intuition auto with bintree.
    (* cas left *)
    intros;   ISL_trans.
    2: apply repeat_invoke_stutter_left.
    ISL_monotony.
    ISL_trans.
    apply IHn; auto with arith.
    ISL_trans.
    2: apply repeat_invoke_stutter_in_left.
    unfold repeat_invoke at 1.
    apply repeat_monotony.
    intros.
    ISL_simplify_left.
    intros l; ISL_simplify_left.
    intros; ISL_trans.
    apply enumBT_lt_repeat_enumBT.
    ISL_trans.
    2: apply repeat_invoke_stutter_in_left.
    unfold repeat_invoke at 1.
    apply repeat_monotony.
    intros; ISL_simplify_left.
    intros k; ISL_simplify_left; intros.
    ISL_trans.
    apply IHn.
    omega.
    unfold repeat_invoke; apply repeat_monotony.
    intros; ISL_simplify_left.
    intros r; ISL_simplify_right; intros.
    constructor 1 with (x:=(Node l r)).
    ISL_monotony.
    ISL_simplify_all.
    intuition eauto with bintree.
    (* cas right *)
    intros; ISL_trans.
    apply enumBT_lt_repeat_enumBT.
    ISL_trans.
    2: apply repeat_invoke_stutter_in_left.
    unfold repeat_invoke at 1.
    apply repeat_monotony.
    intros; ISL_simplify_left.
    intros k; ISL_simplify_left; intros.
    ISL_trans.
    2: apply repeat_invoke_stutter_in_left.
    ISL_trans.
    apply IHn.
    omega.
    unfold repeat_invoke at 1; apply repeat_monotony.
    intros; ISL_simplify_left.
    intros l; ISL_simplify_left.
    intros; ISL_trans.
    apply IHn.
    omega.
    unfold repeat_invoke; apply repeat_monotony.
    intros; ISL_simplify_left.
    intros r; ISL_simplify_right; intros.
    constructor 1 with (x:=(Node l r)).
    ISL_monotony.
    ISL_simplify_all.
    intuition eauto with bintree.
Qed.

Lemma enumBT_lt_height_spec:
  forall (St:Type) (n:nat) (print: bintree -> DS St unit) (st:St), 
             ([< st >] (enumBT_lt print n) >==> 
                 (repeat_invoke (fun t => (height t) < n) print))%ISL.     
Proof.
  intros St n print st.
  ISL_trans. 
  apply enumBT_lt_repeat_enumBT.
  ISL_trans.
  2: apply repeat_invoke_stutter_in_left.
  unfold repeat_invoke at 1.
  apply repeat_monotony.
  intros; ISL_simplify_left.
  intros m; ISL_simplify_left.
  intros; ISL_trans.
  apply enumBT_height_spec.
  apply repeat_invoke_guard_monotony.
  intros t H2; rewrite (is_height_to_height H2).
  auto.
Qed.

Lemma enumBT_lt_exact_once:
  forall (St:Type) (m n:nat), n < m -> 
    forall (print: bintree -> DS St unit) (st:St), 
      ([< st >] (enumBT_lt print m) >==>                
                 ( repeat_invoke (fun k => ~k=n) (enumBT print) -; 
                   enumBT print n -; 
                   repeat_invoke (fun k => ~k=n) (enumBT print)))%ISL.    
Proof.
  induction m.
  intros n H; inversion H.
  intros n H print st; simpl.
  inversion H.
  (* cas debut *)
  ISL_enter_right.
  apply repeat_invoke_zero_left.
  ISL_simplify_right.
  intros; ISL_monotony.
  intros; ISL_trans.
  apply enumBT_lt_repeat_enumBT.
  apply repeat_invoke_guard_monotony.
  intros; omega.
  (* cas apres *)
  ISL_enter_right.
  apply repeat_invoke_stutter_left.
  ISL_unassoc_right.
  ISL_enter_right.
  apply repeat_invoke_one_left with (x:=m).
  omega.
  ISL_monotony.
Qed.

Lemma repeat_invoke_diffNode_r: forall (St:Type) (l l' r:bintree) (print: bintree -> DS St unit) (st:St),  
  ([< st >]
    repeat_invoke (fun t' : bintree => t' <> r)
      (fun r0 : bintree => print (Node l' r0)) >==> 
    repeat_invoke (fun t' : bintree => t' <> Node l r) print)%ISL.
Proof.
   intros; unfold repeat_invoke; 
   apply repeat_monotony.
   intros; ISL_simplify_left.
   intros r'; ISL_simplify_left.
   intros; ISL_simplify_right; intros.
   constructor 1 with (x:=(Node l' r')).
   ISL_simplify_right.
   intuition auto with isl.
   injection H0; auto.
Qed.

Lemma repeat_invoke_diffNode_l: forall (St:Type) (l r:bintree)
  (print: bintree -> DS St unit) (g: bintree -> Prop) (st:St),  
  ([< st >]
    repeat_invoke (fun t' : bintree => t' <> l)
      (fun l0 => repeat_invoke g (fun r0  => print (Node l0 r0))) >==> 
    repeat_invoke (fun t' : bintree => t' <> Node l r) print)%ISL.
Proof.
   intros; ISL_trans.
   2: apply repeat_invoke_stutter_in_left.
   unfold repeat_invoke at 1.
   apply repeat_monotony.
   intros; ISL_simplify_left.
   intros l'; ISL_simplify_left.
   intros; unfold repeat_invoke.
   apply repeat_monotony.
   intros; ISL_simplify_left.
   intros r'; ISL_simplify_left.
   intros; ISL_simplify_right; intros.
   constructor 1 with (x:=(Node l' r')).
   ISL_simplify_right.
   intuition auto with isl.
   injection H1; auto.
Qed.

Hint Resolve repeat_invoke_diffNode_r: bintree.

(* each tree is enumerated only once *)
Lemma enumBT_exact_once:
  forall (St:Type) (n:nat) (t:bintree),  (is_height t n) ->
        forall (print: bintree -> DS St unit) (st:St),  
          ([< st >] (enumBT print n) >==> 
               ( repeat_invoke (fun t' => t'<>t) print -; 
                 print t -; 
                 repeat_invoke (fun t' => t'<>t) print))%ISL.
Proof.
  intros St n; elim n using nat_le_ind; simpl; clear n.
  (* cas de base *)
  intros t H; inversion_clear H.
  intros print st.
  ISL_enter_right.
  apply repeat_invoke_zero_left.  
  ISL_simplify_right.
  intros. apply refInEnv_trans with (sp2:=(print Leaf -; Skip)%ISL).
  apply bind_val_r_ref2; auto with isl.
  intros a; case a; auto with isl.
  ISL_monotony.
  (* cas recursif *)
   intros m IHn t X1; inversion_clear X1; simpl; clear t.
   (* cas both *)
   intros print st. 
   ISL_enter_left.
   eapply IHn with (t:=l); auto with arith.
   ISL_unassoc_left.
   ISL_enter_right.
   apply repeat_invoke_stutter_left.
   ISL_unassoc_right.
   ISL_monotony.
   ISL_trans.
   2: apply repeat_invoke_diffNode_l with (g:=(fun _:bintree => True)).
   apply repeat_invoke_monotony.
   intros; ISL_trans.
   apply enumBT_height_spec. 
   auto with isl.
   intros; ISL_unassoc_left.
   ISL_enter_left.
   eapply IHn with (t:=r); auto with arith.
   ISL_unassoc_left.
   ISL_monotony.
   auto with bintree.
   intros; ISL_unassoc_left.
   ISL_monotony; intros.
   ISL_trans.
   2: apply repeat_invoke_stutter_left.
   ISL_monotony.
   auto with bintree.
   intros; ISL_trans.
   2: apply repeat_invoke_stutter_left.
   ISL_monotony.
   ISL_trans.
   2: apply repeat_invoke_diffNode_l with (g:=(fun _:bintree => True)).
   apply repeat_invoke_monotony.
   intros; ISL_trans.
   apply enumBT_height_spec.
   auto with isl.
   intros; ISL_trans.
   2: apply repeat_invoke_stutter_left.
   ISL_monotony.
   ISL_trans.
   apply enumBT_height_spec.
   ISL_trans.
   2: apply repeat_invoke_stutter_in_left.
   unfold repeat_invoke at 1.
   apply repeat_monotony.
   intros; ISL_simplify_left.
   intros l'; ISL_simplify_left.
   intros; ISL_trans.
   apply enumBT_lt_height_spec.
   ISL_trans.
   2: apply repeat_invoke_diffNode_r with (l':=l').
   apply repeat_invoke_guard_monotony.
   intros t X1 X2; subst.
   rewrite (is_height_to_height H0) in X1.
   omega. 
   intros; ISL_trans.
   apply enumBT_lt_height_spec.
   ISL_trans.
   2: apply repeat_invoke_diffNode_l with (g:=(fun _:bintree => True)).
   intros; ISL_trans.
   2: apply repeat_invoke_guard_monotony with (g1:=(fun t => (height t) < m)).
   apply repeat_invoke_monotony.
   intros; ISL_trans. apply enumBT_height_spec. auto with isl.
   intros t X1 X2; subst.
   rewrite (is_height_to_height H) in X1.
   omega. 
   (* cas left *)
   intros print st.
   ISL_enter_right.
   apply repeat_invoke_stutter_left.
   ISL_unassoc_right.
   ISL_monotony.
   ISL_trans.
   apply enumBT_height_spec. 
   ISL_trans.
   2: apply repeat_invoke_stutter_in_left.
   unfold repeat_invoke at 1.
   apply repeat_monotony.
   intros; ISL_simplify_left.
   intros l'; ISL_simplify_left.
   intros; ISL_trans.
   apply enumBT_height_spec.
   ISL_trans.
   2: apply repeat_invoke_diffNode_r with (l':=l').
   apply repeat_invoke_guard_monotony.
   intros; eapply height_discr; eauto.
   omega. 
   intros.
   ISL_enter_left.
   eapply IHn with (t:=l); auto with arith.
   ISL_unassoc_left.
   ISL_enter_right.
   apply repeat_invoke_stutter_left.  
   ISL_unassoc_right. ISL_monotony; intros.
   ISL_trans.
   2: apply repeat_invoke_diffNode_l with (g:=(fun _:bintree => True)).
   apply repeat_invoke_monotony.
   intros; ISL_trans.
   apply enumBT_lt_height_spec.
   auto with isl.
   ISL_unassoc_left.
   ISL_enter_left.
   eapply enumBT_lt_exact_once with (n:=m0); auto.
   ISL_unassoc_left.
   ISL_enter_right.
   apply repeat_invoke_stutter_left.
   ISL_unassoc_right.
   ISL_monotony.
   ISL_trans.
   2: apply repeat_invoke_diffNode_r with (l':=l).
   ISL_trans.
   2: apply repeat_invoke_stutter_in_left.
   unfold repeat_invoke at 1.
   apply repeat_monotony.
   intros; ISL_simplify_left.
   intros k; ISL_simplify_left.
   intros; ISL_trans.
   apply enumBT_height_spec.
   apply repeat_invoke_guard_monotony.
   intros; eapply height_discr; eauto.
   intros; ISL_unassoc_left.
   ISL_enter_left.
   apply IHn with (t:=r); auto.
   omega.
   ISL_unassoc_left.
   ISL_monotony; intros.
   auto with bintree.
   ISL_unassoc_left.
   ISL_monotony.   
   intros; ISL_trans.
   2: apply repeat_invoke_stutter_left.
   ISL_monotony; auto with bintree.
   intros; ISL_trans.
   2: apply repeat_invoke_stutter_left.
   ISL_monotony.
   ISL_trans.
   2: apply repeat_invoke_diffNode_r with (l':=l).
   unfold repeat_invoke at 1.
   ISL_trans.
   2: apply repeat_invoke_stutter_in_left.
   apply repeat_monotony.
   intros; ISL_simplify_left.
   intros k; ISL_simplify_left.
   intros; ISL_trans.
   apply enumBT_height_spec.
   apply repeat_invoke_guard_monotony.
   intros; eapply height_discr; eauto.
   intros; ISL_trans.
   2: apply repeat_invoke_stutter_left.
   ISL_monotony.
   ISL_trans.
   2: apply repeat_invoke_diffNode_l 
     with (g:=(fun t:bintree => True)).      
   apply repeat_invoke_monotony.
   intros; ISL_trans.
   apply enumBT_lt_height_spec.
   apply repeat_invoke_guard_monotony; auto.
   intros; ISL_trans.
   apply enumBT_lt_height_spec.
   ISL_trans.
   2: apply repeat_invoke_diffNode_l 
     with (g:=(fun t:bintree => is_height t m)).      
   ISL_trans.
   2: apply repeat_invoke_guard_monotony
     with (g1:=(fun t : bintree => height t < m)).
   apply repeat_invoke_monotony.
   intros. apply enumBT_height_spec.
   intro y; intros. eapply height_discr with (n:=(height y)); eauto.
   apply is_height_from_height; auto.
   omega.   
   (* cas right *)
   intros print st.
   ISL_enter_right.
   apply repeat_invoke_stutter_left.
   ISL_unassoc_right.
   ISL_monotony.
   ISL_trans.
   apply enumBT_height_spec. 
   ISL_trans.
   2: apply repeat_invoke_diffNode_l 
     with (g:=(fun t:bintree => is_height t m)).      
   ISL_trans.
   2: apply repeat_invoke_guard_monotony
     with (g1:=(fun t : bintree => is_height t m)).
   apply repeat_invoke_monotony.
   intros. apply enumBT_height_spec.
   intros. eapply height_discr; eauto.
   omega.   
   intros. ISL_enter_right.
   apply repeat_invoke_stutter_left.
   ISL_unassoc_right.
   ISL_monotony.
   ISL_trans.
   apply enumBT_height_spec.
   ISL_trans.
   2: apply repeat_invoke_diffNode_l 
     with (g:=(fun t:bintree => (height t) < m)).      
   ISL_trans.
   2: apply repeat_invoke_guard_monotony
     with (g1:=(fun t : bintree => is_height t m)).
   apply repeat_invoke_monotony.
   intros. apply enumBT_lt_height_spec.
   intros. eapply height_discr; eauto.
   omega.   
   intros.
   ISL_trans.
   eapply enumBT_lt_exact_once with (n:=n); auto.
   ISL_enter_right.
   apply repeat_invoke_stutter_left.  
   ISL_unassoc_right. ISL_monotony; intros.
   ISL_trans.
   2: apply repeat_invoke_stutter_in_left.
   unfold repeat_invoke at 1.
   apply repeat_monotony.
   intros st1; ISL_simplify_left.
   intros k; ISL_simplify_left.
   intros. 
   ISL_trans. apply enumBT_height_spec.
   ISL_trans.
   2: apply repeat_invoke_diffNode_l 
     with (g:=(fun t:bintree => is_height t m)).      
   ISL_trans.
   2: apply repeat_invoke_guard_monotony
     with (g1:=(fun t:bintree => is_height t k)).
   apply repeat_invoke_monotony.
   intros; apply enumBT_height_spec.
   intros; eapply height_discr; eauto.
   ISL_enter_left.   
   eapply IHn with (t:=l); auto with arith.
   ISL_unassoc_left.
   ISL_enter_right.
   apply repeat_invoke_stutter_left.  
   ISL_unassoc_right. ISL_monotony; intros.
   ISL_trans.
   2: apply repeat_invoke_diffNode_l 
     with (g:=(fun t:bintree => is_height t m)).      
   apply repeat_invoke_monotony.
   intros.
   apply enumBT_height_spec.
   ISL_unassoc_left.
   ISL_enter_left.   
   eapply IHn with (t:=r); auto with arith.
   ISL_unassoc_left.
   ISL_monotony.
   apply repeat_invoke_diffNode_r with (l':=l).
   intros; ISL_unassoc_left.
   ISL_monotony.
   intros.
   ISL_trans.
   2: apply repeat_invoke_stutter_left.
   ISL_monotony.
   apply repeat_invoke_diffNode_r.
   intros; ISL_trans.
   2: apply repeat_invoke_stutter_left.
   ISL_monotony.
   ISL_trans.
   2: apply repeat_invoke_diffNode_l 
     with (g:=(fun t:bintree => is_height t m)).      
   apply repeat_invoke_monotony.
   intros; apply enumBT_height_spec.
   intros; ISL_trans.
   2: apply repeat_invoke_diffNode_l 
     with (g:=(fun t:bintree => is_height t m)).      
   ISL_trans.
   2: apply repeat_invoke_stutter_in_left.
   unfold repeat_invoke at 1.
   apply repeat_monotony.
   intros; ISL_simplify_left.
   intro k; ISL_simplify_left; intros.
   ISL_trans.
   apply enumBT_height_spec.
   ISL_trans.
   2: apply repeat_invoke_guard_monotony with
     (g1:=(fun t : bintree => is_height t k)).
   apply repeat_invoke_monotony.
   intros; apply enumBT_height_spec.
   intros; eapply height_discr; eauto.
Qed. 


Lemma bintree_exists: forall (n:nat), exists t:bintree, (height t)=n.
Proof.
  induction n.
  constructor 1 with (x:=Leaf).
  simpl; auto.
  case IHn.
  intros t H; constructor 1 with (x:=(Node t Leaf)).
  simpl; rewrite H.
  case n; auto.
Qed.

Lemma repeat_invoke_cons: forall (A:Set) (P: A-> Prop) (l:list A),
   ([< l >]
      (repeat_invoke P
       (fun x:A => Bind m <- ! allPl In allPl :<- (cons x m)))
     >==>  (upd_abspost allPl
              (fun m r => exists m':(list A), m=(app m' l) /\ AllS P m')))%ISL.
Proof.
  intros A P l; unfold repeat_invoke, repeat.
  ISL_simplify_left.
  (* cas de base *)
    intros n; generalize l; clear l; elim n; clear n; simpl.
    intros l H; apply updabspost2wp with (old_st:=l); simpl; intuition auto.
    constructor 1 with (x:=(nil (A:=A))); simpl.
    intuition auto with lists.
  (* cas recursif *)
    intros n HR l.
    ISL_simplify_left. (* ICI *)
    intros a H; ISL_simplify_left.
    ISL_trans. apply HR.
    apply updabspost_monotony; auto with place.
    intros st2 r H0; case H0; clear H0.
    intros m H0.
    constructor 1 with (x:=(m ++ (a::nil))).
    rewrite <- ass_app; simpl.
    intuition auto with lists.
Qed.

(* A first-order "version" of the theorem, 
 * using only results of "version 2".
 *)
Lemma enumbintree: forall (St:Type) (n:nat),
  ([<nil>] 
    (enumBT (fun t => Bind m <- ! allPl In allPl :<- t::m) n)
      >==> (upd_abspost allPl 
              (fun l r => enumlist (fun t => (height t)=n) l)))%ISL.
Proof.
  intros St n; ISL_trans. 
  apply updabspost_split
    with (v:=(allPl (A:=list bintree)))
         (q1:=(fun l (r:unit) =>
                forall t : bintree, List.In t l -> height t = n))
         (q2:= (fun l (r:unit) =>
                (forall (t:bintree), (height t)=n -> 
                 exists l1:list bintree, exists l2:list bintree,
                 (l= l1 ++ (t::l2)) 
                 /\ ~(List.In t l1) /\ ~(List.In t l2)))).
  (* cond 1 *)
  ISL_trans. apply enumBT_height_spec.     
  ISL_trans. apply repeat_invoke_cons.
  apply updabspost_monotony; auto with place.
  intros st2 r H; case H; clear H.
  intros l; rewrite <- app_nil_end.
  intuition auto; subst.
  apply is_height_to_height.
  pattern t; eapply allS_in_ext; eauto.
  (* cond 2 *)
  apply updabspost_splitforall
    with (p:= (fun t => (height t)=n))
         (q:=(fun (t:bintree) l (r:unit) =>
           exists l1 : list bintree,
           (exists l2 : list bintree, l = l1 ++ t :: l2 /\ ~ List.In t l1 /\ ~ List.In t l2))).
  apply bintree_exists.
  intros t H.
  ISL_trans. apply enumBT_exact_once with (t:=t).
  rewrite <- H; apply is_height_from_height.
  ISL_enter_left.
  apply repeat_invoke_cons.
  ISL_simplify_left. (* ICI *)
  intros r l H0; case H0; clear r H0.
  intros l0; rewrite <- app_nil_end.
  intuition subst.
  ISL_simplify_left.
  ISL_simplify_left.
  ISL_simplify_left.
  ISL_trans.
  apply repeat_invoke_cons.
  apply updabspost_monotony; auto with place.
  intros l r H; case H; clear r H.
  intros m; intuition subst.
  constructor 1 with (x:=m).
  constructor 1 with (x:=l0).
  constructor 1.
  auto.
  constructor 1; apply allS_not_in; auto.
  (* fin *)
  apply updabspost_monotony; auto with place.
  intros; apply enumlist_ext1;
  intuition auto with lists.
Qed.

(**** Version 3: Study of specifications by "monitoring" ****
* but it does not really work...
* maybe because data-refinement is currently missing in DSL 
*)

Lemma enumBT_lt_monotony_aux:
  forall (St: Type) (p1 p2: bintree -> DS St unit) ,
    (forall (t:bintree) (st:St), [< st >] p1 t >==> p2 t)%ISL
     -> forall (n:nat), 
      (forall (m:nat) (st:St), m < n -> 
        [< st >] (enumBT p1 m) >==> (enumBT p2 m))%ISL  
             -> forall (st:St), ([< st >] (enumBT_lt p1 n) >==> (enumBT_lt p2 n))%ISL.
Proof.
  induction n; simpl; intros; auto with isl bintree.
  ISL_monotony.
Qed.

Hint Resolve enumBT_lt_monotony_aux: bintree.

Lemma enumBT_monotony:
  forall (St: Type) (n:nat) (p1 p2: bintree -> DS St unit) ,
    (forall (t:bintree) (st:St), [< st >] p1 t >==> p2 t)%ISL
     -> forall (st:St), ([< st >] (enumBT p1 n) >==> (enumBT p2 n))%ISL.
Proof.
  intros St n. elim n using nat_le_ind; simpl; intros; clear n; auto with isl bintree arith.
  (* cas recursif *)
  ISL_monotony.
  intros; ISL_monotony; auto with bintree arith.
Qed.

Hint Resolve enumBT_monotony: bintree.

Lemma enumBT_lt_monotony:
  forall (St: Type) (p1 p2: bintree -> DS St unit) ,
    (forall (t:bintree) (st:St), [< st >] p1 t >==> p2 t)%ISL
     -> forall (n:nat) (st:St), ([< st >] (enumBT_lt p1 n) >==> (enumBT_lt p2 n))%ISL.
Proof.
  intros; auto with bintree arith.
Qed.

Lemma enumBT_lt_embed_monotony_aux:
  forall (St St1 St2: Type) (v1:(Place St St1)) (v2:(Place St St2)) 
     (p1: bintree -> DS St1 unit) (p2: bintree -> DS St2 unit),
    (forall (t:bintree) (st:St), [< st >] (Embed[v1] (p1 t)) >==> (Embed [v2] p2 t))%ISL
     -> forall (n:nat), 
      (forall (m:nat) (st:St), m < n -> 
        [< st >] (Embed [v1] (enumBT p1 m)) >==> (Embed [v2] (enumBT p2 m)))%ISL  
             -> forall (st:St), ([< st >] (Embed [v1] (enumBT_lt p1 n)) >==> (Embed [v2] (enumBT_lt p2 n)))%ISL.
Proof.
  induction n; simpl; intros.
  ISL_simplify_all.
  autorewrite with placeeq.
  auto with isl bintree.
  ISL_unassoc_left.
  ISL_embed_intro_left.
  ISL_enter_left.
  apply H0; auto with arith.
  ISL_unassoc_left.
  ISL_unassoc_right.
  ISL_monotony.
  intros; ISL_simplify_left.
  ISL_trans.
  apply IHn; auto with isl arith.
  auto with isl.
Qed.

Hint Resolve enumBT_lt_embed_monotony_aux: bintree.

Lemma enumBT_embed_monotony:
  forall (St St1 St2: Type) (v1:(Place St St1)) (v2:(Place St St2)) (n:nat)
     (p1: bintree -> DS St1 unit) (p2: bintree -> DS St2 unit),
    (forall (t:bintree) (st:St), [< st >] (Embed [v1] (p1 t)) >==> (Embed [v2] (p2 t)))%ISL
     -> forall (st:St), 
        ([< st >] (Embed [v1] (enumBT p1 n)) >==> (Embed [v2] (enumBT p2 n)))%ISL.  
Proof.
  intros St St1 St2 v1 v2 n. elim n using nat_le_ind; simpl; intros; clear n; auto.
  (* cas recursif *)
  ISL_unassoc_left; ISL_unassoc_right.
  ISL_embed_intro_left.
  ISL_embed_intro_right.
  ISL_monotony.
  intros st2 zzz.
  ISL_unassoc_left.
  ISL_unassoc_right.
  ISL_embed_intro_left.
  ISL_embed_intro_right. 
  ISL_monotony.
  auto with bintree isl arith.
  intros; auto with isl arith bintree.
Qed.

Hint Resolve enumBT_embed_monotony: bintree.

Lemma enumBT_lt_embed_monotony:
  forall (St St1 St2: Type) (v1:(Place St St1)) (v2:(Place St St2)) 
     (p1: bintree -> DS St1 unit) (p2: bintree -> DS St2 unit),
    (forall (t:bintree) (st:St), [< st >] (Embed [v1] (p1 t)) >==> (Embed [v2] (p2 t)))%ISL
  -> forall (n:nat) (st:St), ([< st >] (Embed [v1] (enumBT_lt p1 n)) >==> (Embed [v2] (enumBT_lt p2 n)))%ISL.
Proof.
  intros.
  apply enumBT_lt_embed_monotony_aux.
  auto.
  intros; apply enumBT_embed_monotony. 
  auto.
Qed.


Lemma enumBT_lt_height_aux:
  forall (St: Type) (n:nat) (print: bintree -> DS St unit),
     (forall (m:nat) (st:St), m < n -> 
        [< st >] (enumBT print m) >==> (enumBT (fun t => Ensure (is_height t m) /; print t) m))%ISL  
             -> forall (st:St), ([< st >]
                   (enumBT_lt print n) >==> 
                      (enumBT_lt (fun t => Ensure ((height t) < n) /; print t) n))%ISL.     
Proof.
  induction n; simpl.
  auto with isl.
  intros print H st.
  ISL_monotony.
  ISL_trans. apply H.
  omega.
  apply enumBT_monotony.
  intros; ISL_monotony.
  ISL_simplify_all; intuition auto.
  rewrite (is_height_to_height H1).
  omega.
  intros; ISL_trans.
  apply IHn.
  intros; auto with arith bintree.
  apply enumBT_lt_monotony.
  intros; ISL_monotony.
  ISL_simplify_all; intuition auto with arith.
Qed.

Hint Resolve enumBT_lt_height_aux: bintree.

Lemma enumBT_height:
  forall (St: Type) (n:nat) (print: bintree -> DS St unit) (st:St), 
          ([< st >] (enumBT print n) >==> 
                    (enumBT (fun t => Ensure (is_height t n) /; print t) n))%ISL.
Proof.
  intros St n. elim n using nat_le_ind; simpl; clear n.
  (* cas de base *)
  intros; ISL_simplify_right.
  intuition auto with isl bintree.  
  (* cas recursif *)
  intros n IHn print st; simpl.
  ISL_monotony.
    (* cas both *)
    ISL_trans. apply IHn; auto with arith.
    apply enumBT_monotony.
    intros l st0; ISL_simplify_left.
    intros; ISL_trans. apply IHn; auto with arith.
    apply enumBT_monotony.
    intros r st1; ISL_monotony.
    ISL_simplify_all; auto with bintree.
    (* cas left *)
    intros; ISL_monotony.
    ISL_trans. apply IHn; auto with arith.
    apply enumBT_monotony.
    intros l st0; ISL_simplify_left.
    intros; ISL_trans. apply enumBT_lt_height_aux; auto with arith.
(*    intros; ISL_trans. apply IHn; auto with arith.
    auto with bintree isl. *)
    apply enumBT_lt_monotony.
    intros r st1; ISL_monotony.
    ISL_simplify_all; 
    intuition eauto with bintree.
    (* cas right *)
    intros; ISL_trans. apply enumBT_lt_height_aux; auto with arith.
(*    intros; ISL_trans. apply IHn; auto with arith.
    auto with bintree isl. *)
    apply enumBT_lt_monotony.
    intros l st1; ISL_simplify_left.
    intros; ISL_trans. apply IHn; auto with arith.
    apply enumBT_monotony.
    intros r st3; ISL_monotony.
    ISL_simplify_all; 
    intuition eauto with bintree.
Qed.

Hint Resolve enumBT_height: bintree.

Lemma enumBT_lt_height:
  forall (St: Type) (n:nat) (print: bintree -> DS St unit) (st:St), 
       ([< st >] (enumBT_lt print n) >==> 
                      (enumBT_lt (fun t => Ensure ((height t) < n) /; print t) n))%ISL.     
Proof.
  intros; auto with bintree arith.
Qed.

(*

Definition snd3Pl (A B C:Type):  (Place (prodT A (prodT B C)) B)
  := (pR _ _)#(pL _ _).

Definition csnd3Pl (A B C:Type):  (Place (prodT A (prodT B C)) (prodT A C)).
 intros A B C.
 refine (Build_Place 
            (fun e => match e with (pairT x y) => (pairT x (sndT y)) end)
            (fun (e:(prodT A (prodT B C))) x => match x with (pairT a b) => (pairT a (pairT (fstT (sndT e)) b)) end) 
            _
            _  
            _
            );
abstract (
 ( intros e x; case x; simpl; auto ) ||
 ( intros e; case e; simpl; intros a p; case p; simpl; auto ) ||
 ( intros y; case y; simpl; 
   intros x p y0; case y0; simpl; auto )
).
Defined.

Definition mid3Pl (A B C:Type): (sumPl (snd3Pl A B C) (csnd3Pl A B C)).
 intros A B C.
 refine (Build_sumPl _ _
            (fun x y => match y with (pairT a b) => (pairT a (pairT x b)) end)
            _
            _  
            ); 
abstract (intros e; case e; simpl; auto).
Defined.


Lemma block_compose3:
 forall (R A B St:Type) (x:B) (sp: (DS (prodT B St) R)) (sp0:DS (prodT A St) R) (sti:prodT A St),
   ([< sti >] (block (mid3Pl _ _ _) x (Embed [(pR A _)] sp)) >==> sp0)%ISL
      -> ([< sti >] (Embed [(pR A St)] (block (pairsumPl B St) x sp)) >==> sp0)%ISL.
Proof.
  unfold block, bind, set, refInEnv.
  intros R A B St x sp sp0 sti; case sti; clear sti.
  intros a st; simpl; intro H.
  eapply refines_trans.
  2: apply H.
  eapply refines_trans.
  apply FSL.bind_assoc_ref1.
  eapply refines_trans.
  2:apply FSL.bind_assoc_ref2.
  simpl; autorewrite with placeeq.
  apply FSL.bind_monotony; auto with wprefines.
  intros b; case b; clear b; simpl.
  intros p r; case p; simpl; intros.
  intros; apply refines_fold; simpl.
  auto.
Qed.

Definition dejaVu (St:Type) := (pL bool St).
Implicit Arguments dejaVu [St].
Definition screen (St:Type) := (pR bool St).
Implicit Arguments screen [St].
Definition monitoring (St:Type) := (pairsumPl bool St). 
Implicit Arguments monitoring [St]. 

Section Monitor.

Variable A: Type.
Variable eqA: A -> A -> bool.

Hypothesis eqA_true_eq: forall (x y:A), (eqA x y)=true -> x=y.
Hypothesis eqA_false_neq: forall (x y:A), (eqA x y)=false -> x<>y.

Lemma eqA_case: forall (x y:A), (eqA x y)=true \/ (eqA x y)=false.
Proof.
  intros x y; case (eqA x y); simpl; auto.
Qed.

Lemma neq_eqA_false: forall (x y:A), x<>y -> (eqA x y)=false.
Proof.
  intros x y; case (eqA_case x y); simpl; auto.
  intros H; rewrite (eqA_true_eq H).
  intros X; case X; auto.
Qed.

Lemma eq_eqA_true: forall (x y:A), x=y -> (eqA x y)=true.
Proof.
  intros x y; case (eqA_case x y); simpl; auto.
  intros H X; case (eqA_false_neq H).
  auto.
Qed.

Definition monitorIn (St: Type) (f: A -> DS St unit) 
                       (x y:A): (DS (prodT bool St) unit)
 := ([dejaVu](if (eqA x y) 
              then ( 
                  (Bind  d <- !allPl In
                   Ensure (d=false) ) -;
                  allPl :<- true 
              ) else (
                  Skip
              )) -;
     (Embed [screen] (f y)))%ISL.

Definition monitorOut (St: Type) (test: DS (prodT bool St) unit): (DS St unit)
 := (block monitoring false (test -; Bind d <- !dejaVu In Ensure (d=true)))%ISL.

Lemma intro_monitorOut: 
  forall (St: Type) (sp: DS St unit) (test: DS (prodT bool St) unit) (st:St),
         ([< st >] sp >==> (monitorOut test))%ISL
     ->  ([< (pairT false st) >] 
             ([screen] sp -; dejaVu :<- true) >==> test)%ISL.  
Proof.
  unfold monitorOut; intros.
  ISL_enter_left.
  apply H.
  ISL_embed_intro_left.
  ISL_enter_left.
  apply embed_block_left.
  auto with isl. 
  ISL_unassoc_left.
  ISL_simplify_left.
  ISL_unassoc_left.
  ISL_unassoc_left.
  ISL_trans.
  2:apply skip_r_ref1.
  ISL_monotony.
  intros st2 x; case st2; case x; simpl.
  intros; ISL_simplify_left.
  intuition subst.
  ISL_Trivial.
Qed.
 
Lemma elim_monitorOut: 
  forall (St: Type) (sp: DS St unit) (test: DS (prodT bool St) unit) (st:St),
        ([< (pairT false st) >] ([screen] sp -; dejaVu :<- true) >==> test)%ISL -> 
         ([< st >] sp >==> (monitorOut test))%ISL.
Proof.
  unfold monitorOut; intros.
  apply block_enter_right.
  apply refInEnv_trans with 
   (sp2:=(([screen]sp -;  dejaVu :<- true) -; (Bind d <- !dejaVu In Ensure (d=true)) -; dejaVu :<- false)%ISL).
  ISL_unassoc_right.
  ISL_monotony.
  intros st2 x; case x; ISL_Trivial.
  ISL_enter_left.
  apply H.
  ISL_unassoc_right.
  ISL_monotony.
  ISL_Trivial.  
Qed.

Variable sp1 sp2: forall (St:Type), (A -> DS St unit) -> (DS St unit).

Hypothesis sp1_embed_monotony:
  forall (St St1 St2: Type) (v1:(Place St St1)) (v2:(Place St St2))
     (p1: A -> DS St1 unit) (p2: A -> DS St2 unit),
    (forall (x:A) (st:St), [< st >] (Embed [v1] (p1 x)) >==> (Embed [v2] (p2 x)))%ISL
     -> forall (st:St), 
        ([< st >] (Embed [v1] (sp1 p1)) >==> (Embed [v2] (sp1 p2)))%ISL.  

Hypothesis sp2_embed_monotony:
  forall (St St1 St2: Type) (v1:(Place St St1)) (v2:(Place St St2))
     (p1: A -> DS St1 unit) (p2: A -> DS St2 unit),
    (forall (x:A) (st:St), [< st >] (Embed [v1] (p1 x)) >==> (Embed [v2] (p2 x)))%ISL
     -> forall (st:St), 
        ([< st >] (Embed [v1] (sp2 p1)) >==> (Embed [v2] (sp2 p2)))%ISL.  

Hypothesis sp1_monotony:
  forall (St: Type) 
     (p1: A -> DS St unit) (p2: A -> DS St unit),
    (forall (x:A) (st:St), [< st >] (p1 x) >==> (p2 x))%ISL
     -> forall (st:St), 
        ([< st >] (sp1 p1) >==> (sp1 p2))%ISL.  

Hypothesis sp2_monotony:
  forall (St: Type) 
     (p1: A -> DS St unit) (p2: A -> DS St unit),
    (forall (x:A) (st:St), [< st >] (p1 x) >==> (p2 x))%ISL
     -> forall (st:St), 
        ([< st >] (sp2 p1) >==> (sp2 p2))%ISL.  

Variable a b:A.

Variable f: A -> A -> A.
Hypothesis f_a_b_uniq: forall (x y:A), (f x y)=(f a b) -> x=a /\ y=b.

(*

Lemma continuation_monitorIn_both:
  (forall (St:Type) (print: A -> DS St unit) (st:St),
   [< st >] (sp2 print) >==> 
              (monitorOut
                  (sp2 (monitorIn print a))))%ISL ->
  (forall (St:Type) (print: A -> DS St unit) (st:St),
   [< st >] (sp1 print) >==> 
              (monitorOut
                  (sp1 (monitorIn print b))))%ISL ->
  (forall (St:Type) (print: A -> DS St unit) (st:St),
   [< st >](sp2 (fun a' => (sp1 (fun b' => print (f a' b'))))) >==> 
              (monitorOut 
                  (sp2 (fun a' => (sp1 (monitorIn (fun b' => print (f a' b')) (f a b)))))))%ISL.
Proof.
  intros H1 H2 St p st.
  unfold monitorOut.
  apply block_enter_right.
  ISL_enter_left.
  apply sp2_monotony with (p2:=(fun a' : A => 
   (monitorOut (sp1 (monitorIn (fun b' : A => p (f a' b')) b))))).
  intros; apply H2.
  ISL_enter_left.
  apply H1.
  unfold monitorOut.
  unfold monitoring.
  apply block_compose3.
  apply block_enter_left.
  ISL_unassoc_left.
(**
Est-ce qu'il ne faudrait pas 
du raffinement de donnee !!!!!
**)



  ISL_unassoc_left.
  ISL_unassoc_right.
  ISL_unassoc_right.

(*  (* essai1 ? *)
  ISL_trans.
  apply H1.
  unfold monitorOut at 1.
  intros; apply block_enter_left.
  ISL_unassoc_left.
  ISL_enter_left.
*)

Qed.

*)

End Monitor.

Lemma enumBT_height_discr: 
  forall (St: Type) (print: bintree -> DS St unit) (t:bintree) (n:nat) (st:(prodT bool St)),
      (height t) <> n ->
           ([< st >]
             (Embed [screen] (enumBT print n))  >==> (enumBT (monitorIn eqbintree print t) n))%ISL.  
Proof.
  intros St print t n st H.
  ISL_enter_left.
  apply enumBT_height.
  ISL_trans.
  2:apply skip_r_ref1.
  apply enumBT_embed_monotony.
  intros t0 st0. unfold monitorIn.
  ISL_unassoc_left.
  ISL_simplify_left.
  intros H0.
  assert (X:t <> t0).
  eapply height_discr; auto with bintree.
  rewrite (is_height_to_height H0); auto.
  rewrite (noteq_eqbintree X).
  ISL_unassoc_right.
  ISL_simplify_right.
  case st0; simpl; intros; ISL_simplify_left.
  ISL_unassoc_right.
  ISL_monotony.
Qed.

Lemma enumBT_lt_height_discr: 
  forall (St: Type) (print: bintree -> DS St unit) (t:bintree) (n:nat) (st:(prodT bool St)),
      n <= (height t) ->
           ([< st >]
             (Embed [screen] (enumBT_lt print n))  >==> (enumBT_lt (monitorIn eqbintree print t) n))%ISL.  
Proof.
  intros St print t n st H.
  ISL_enter_left.
  apply enumBT_lt_height.
  ISL_trans.
  2:apply skip_r_ref1.
  apply enumBT_lt_embed_monotony.
  intros t0 st0. unfold monitorIn.
  ISL_unassoc_left.
  ISL_simplify_left.
  intros H0.
  assert (X:t <> t0).
  eapply height_discr; auto with bintree.
  omega.
  rewrite (noteq_eqbintree X).
  ISL_unassoc_right.
  ISL_simplify_right.
  case st0; simpl; intros; ISL_simplify_left.
  ISL_unassoc_right.
  ISL_monotony.
Qed.


       
Lemma enumBT_lt_exact_once_aux:
  forall (St: Type) (print: bintree -> DS St unit) (t:bintree) (n:nat),
      (height t) < n ->
         (forall (st:St),
           [< (pairT false st) >] 
             ([screen] (enumBT print (height t)) -; dejaVu :<- true) >==> 
                  (enumBT (monitorIn eqbintree print t) (height t)))%ISL  
             -> forall (st:St), 
                 ([< (pairT false st) >]
                    ([screen] (enumBT_lt print n) -; dejaVu :<- true) >==> 
                       (enumBT_lt (monitorIn eqbintree print t) n))%ISL.     
Proof.
  induction n; simpl.
  intros H; inversion H.
  intros H; inversion H.
  (* cas debut *)
  intros H0 st; subst.
  ISL_enter_right.
  apply H0.
  ISL_unassoc_left.
  ISL_unassoc_right.
  ISL_monotony.
  intros; ISL_simplify_right.
  intros; ISL_trans. 
  2: apply enumBT_lt_height_discr; auto with arith.
  ISL_simplify_left.
  refine (set_permut_sumPl_left monitoring _ _).
  ISL_monotony.
  intros st3 x0; case x0; ISL_Trivial.
  subst.
  intros H0 st.
  ISL_enter_right.
  apply enumBT_height_discr.
  omega.
  ISL_unassoc_left.
  ISL_unassoc_right.
  ISL_monotony.
  intros; ISL_simplify_right.
  auto with arith.
Qed.

Axiom Essai: False.

Lemma enumBT_exact_once:
  forall (n:nat) (t:bintree), (is_height t n) 
   -> forall (St: Type) (print: bintree -> DS St unit),
         (forall (st:St),
           [< st >] 
             (enumBT print n) >==> 
                  (monitorOut (enumBT (monitorIn eqbintree print t) n)))%ISL.
Proof.
  intros n. elim n using nat_le_ind; simpl; clear n.
  (* cas de base *)
  intros t H; inversion H.
  unfold monitorOut, monitorIn; simpl.
  intros; apply block_enter_right.
  ISL_unassoc_right.
  ISL_unassoc_right.
  ISL_simplify_right.
  intuition auto.
  ISL_unassoc_right.
  ISL_simplify_left.
  ISL_enter_right.
  assert (X: ([< pairT false st >] dejaVu :<- true >==>  allPl :<- pairT true st)%ISL).
  elim Essai. (* ICI !!! *)
  apply X.
  refine (set_permut_sumPl_left monitoring _ _).
  ISL_monotony.
  intros st2 x; case x; ISL_Trivial.
  (* cas recursif *)
  intros n IHn t H; inversion H.
    (* cas both *)
    subst; intros St print st; simpl.
    apply elim_monitorOut.

    unfold monitorOut.
    apply block_enter_right.
    ISL_unassoc_left.
    ISL_unassoc_right.
    ISL_unassoc_right.    
    ISL_embed_intro_left.
    (*******)
    apply refInEnv_trans with (sp2:=
    (enumBT
     (fun l0 : bintree =>
      enumBT (fun r0 : bintree => monitorIn print (Node l r) (Node l0 r0)) n) n
      -;  (Bind d <- !dejaVu In Ensure d = true)) -;
 (enumBT
    (fun l0 : bintree =>
     enumBT_lt (fun r0 : bintree => monitorIn print (Node l r) (Node l0 r0)) n)
    n -; 
  enumBT_lt
    (fun l0 : bintree =>
     enumBT (fun r0 : bintree => monitorIn print (Node l r) (Node l0 r0)) n) n) 
 Bind a1 <- (Bind d <- !dejaVu In Ensure d = true)
 In dejaVu :<- false -;  Val a1)%ISL
  ).

    ISL_enter_left.    
    ISL_enter_left.
    apply IHn with (t:=l); auto with arith.
    unfold monitorOut.
    apply embed_block_left with (u:=(pairsumPl bool St)) (x:=false) (st:=st).
    auto with isl.

    ISL_unassoc_left.
    ISL_enter_left.
    apply IHn with (t:=r); auto with arith.
    unfold monitorOut.

    auto with isl.
    SL_embed_intro_left.
    ISL_monotony    ISL_trans.

*)
