(*

A (partial) proof of the following ML code:

let rec eratosthene f n =
  if n=2
  then
    (f 2)
  else
    let isprime=ref true in
      eratosthene (fun p ->
                         (if n mod p = 0 then isprime:=false) ;
                         (f p))
                  (n-1) ;
      if !isprime then (f n);;

*)

Set Implicit Arguments.

Require Import MyTheoryList.
Require Import ISLLoops.

Lemma AllS_dec:  forall (A:Set) (P: A -> Prop) (P_dec: forall (a:A), {(P a)} + {~(P a)})
 (l:list A), {(AllS P l)} + {exists a:A, (List.In a l) /\ ~(P a)}.
Proof.
  induction l.
  constructor 1; auto with lists.
  case (P_dec a); intros H1.
  case IHl; intros H2.
  constructor 1; auto with lists.
  constructor 2; case H2.
  intros x H3; constructor 1 with (x:=x).
  simpl.
  intuition auto with lists.
  constructor 2; simpl; intuition eauto.
Qed.

Section Prime.

(* Here, I assume a "divides" test *)

Variable divides: nat -> nat -> bool.

(* Below, I assume that the predicate "primelist n l" means that l is a list 
 * containing successive prime numbers from 2 to n in the reverse order.
 *)
Inductive primelist: nat -> (list nat) -> Prop :=
| two_prime: (primelist 2 (cons 2 nil))
| next_prime: forall (n:nat) (l:list nat), 
    (primelist n l) -> 
       (AllS (fun x => (divides x (S n))=false) l) ->
           (primelist (S n) (cons (S n) l))
| next_not_prime: forall (n p:nat) (l:list nat), 
    (primelist n l) -> (List.In p l) -> (divides p (S n))=true ->
           (primelist (S n) l).

Hint Resolve two_prime next_prime: isl.

(* Hence, here is the specification of Eratosthene sieve *)
Definition absEratosthene (St:Type) (print: nat -> DS St unit) (n:nat) : (DS St unit)
 := (Bind l <- (Any (list nat))
     In Ensure (primelist n l) -;
       listRevIter l print)%ISL.

(* Here is the closure used in the recursive call *)
Definition monitorIn (St: Type) (f: nat -> DS St unit)
                       (n p:nat): (DS (bool*St) unit)
 := ([pL](if (divides p n)
              then 
                  allPl :<- false
              else
                  Skip
              ) -;
     (Embed [pR] (f p)))%ISL.


Lemma eratosthene_aux:
  forall (l:list nat) (n:nat) (St: Type) (print: nat -> DS St unit) (st:St) (b:bool), 
       ([< (b,st) >] listRevIter l (monitorIn print n)
         >==>  ([pR] (listRevIter l print) -; (Bind b <-/ any bool In pL :<- b)))%ISL.
Proof.
  induction l; simpl.
  intros; ISL_Trivial.
  intros; ISL_enter_left.
  apply IHl.
  ISL_unassoc_right.
  ISL_unassoc_left.
  ISL_monotony.
  intros; ISL_simplify_left.
  intros; ISL_simplify_left.
  unfold monitorIn.
  apply refInEnv_trans with 
  (sp2:=([pL] (Bind b0 <-/ Any bool In allPl :<- b0) -; Embed [pR]print a)%ISL).
  ISL_monotony.
  case (divides a n); ISL_Trivial.
  ISL_simplify_left.
  intros; ISL_simplify_left.
  ISL_assoc_right.
  apply refInEnv_trans with
   (sp2:=(pL :<- b -; [pR]print a -;  Bind b0 <-/ Any bool In pL :<- b0)%ISL).
  apply set_permut_sumPl_left with (v1:=pL (A:=bool) (B:=St)).
  apply pairsumPl.
  ISL_monotony.
  intros; case x0.
  ISL_Trivial.
  ISL_monotony.
  ISL_Trivial.
Qed.

Lemma eratosthene_nextprime:
  forall (n:nat) (l:list nat) (St: Type) (print: nat -> DS St unit) (st:St), 
    (AllS (fun x => divides x (S n) = false) l) ->  
       ([< (true,st) >] listRevIter l (monitorIn print (S n))
         >==>  ([pR] (listRevIter l print) -; pL :<- true))%ISL.
Proof.
  induction 1; simpl.
  intros; ISL_Trivial.
  ISL_enter_left.
  apply IHAllS.
  ISL_unassoc_right.
  ISL_unassoc_left.
  ISL_monotony.
  intros; ISL_simplify_left.
  unfold monitorIn.
  rewrite H.
  ISL_simplify_left.
  ISL_monotony.
  intros; case x0.
  ISL_Trivial.
Qed.

Lemma eratosthene_next_notprime:
  forall (n:nat) (l:list nat) (St: Type) (print: nat -> DS St unit) (st:St) (p:nat), 
    (List.In p l) -> (divides p (S n) = true) ->  
       ([< (true,st) >] listRevIter l (monitorIn print (S n))
         >==>  ([pR] (listRevIter l print) -; pL :<- false))%ISL.
Proof.
  induction l; simpl.
  intros; tauto.
  intros St print st p H X.
  ISL_unassoc_right.
  ISL_monotony_right.
  intros; apply refInEnv_trans with 
  (sp2:=((pL :<- false -; Embed [pR](print a))%ISL)).
  auto with isl.
  ISL_simplify_left.
  apply refInEnv_trans with 
  (sp2:=(pL :<- true -; [pR]print a -;  pL :<- false)%ISL).
  apply set_permut_sumPl_left with (v1:=pL (A:=bool) (B:=St)).
  apply pairsumPl.
  ISL_monotony.
  intros; case x; ISL_Trivial.
  ISL_monotony.
  ISL_Trivial.
  
  unfold monitorIn at 2.
  set (b:=divides a (S n)).
  cut (divides a (S n)=b); auto.
  case b; simpl; clear b.

  intros; ISL_enter_left.
  apply  eratosthene_aux.
  ISL_unassoc_left.
  ISL_monotony.
  intros; ISL_simplify_left.
  intros; ISL_simplify_left.
  ISL_simplify_left.
  ISL_assoc_right.
  ISL_assoc_right.
  ISL_simplify_right; auto with isl.

  case H; clear H.
  intros; subst.
  rewrite X in H0; discriminate H0.
  intros. ISL_enter_left. apply IHl with (p:=p); intuition auto.
  ISL_assoc_right.
  ISL_monotony.
  intros; ISL_simplify_left.
  auto with isl.
Qed.

Lemma absEratosthene_prop:
  forall (n:nat), 2 <= n ->
 forall (St: Type) (print: nat -> DS St unit) (st:St),
  ([< st >] 
    (block true
       (absEratosthene (monitorIn print (S n)) n -; 
        If !pL Then (Embed [pR](print (S n))) Else Skip))
    >==> (absEratosthene print (S n)))%ISL.
Proof.
  intros.
  apply block_enter_left.
  ISL_unassoc_left.  
  unfold absEratosthene at 1.
  ISL_unassoc_left.
  ISL_simplify_left.
  intros l; ISL_simplify_left.
  ISL_unassoc_left.
  ISL_simplify_left.
  intros; ISL_simplify_left.
  unfold absEratosthene.
  ISL_unassoc_right.
  ISL_simplify_right.
  intros X; clear X.
  generalize (AllS_dec (fun x => (divides x (S n))=false)).
  intros H1.
  lapply H1; clear H1.
  intros H1; case (H1 l); clear H1.
  intros H1.
  constructor 1 with (x:=cons (S n) l).
  ISL_unassoc_right.
  ISL_simplify_right.
  intros X; clear X.
  constructor 1.
  auto with isl.
  ISL_simplify_left.
  ISL_simplify_left.
  ISL_unassoc_right.
  ISL_enter_left.
  apply eratosthene_nextprime; auto.
  ISL_unassoc_left.
  ISL_monotony.
  intros; ISL_simplify_left.
  ISL_unassoc_left.
  ISL_simplify_left.
  ISL_unassoc_left.
  ISL_monotony.
  intros st3 x0; case x0.
  ISL_Trivial.
  intros H1; case H1; clear H1.
  intros p.
  constructor 1 with (x:=l).
  ISL_unassoc_right.
  ISL_simplify_right.
  intros X; clear X.
  constructor 1.
  apply next_not_prime with (p:=p);
  intuition auto. 
  generalize H3.
  case (divides p (S n)); auto.
  ISL_simplify_left.
  ISL_simplify_left.
  ISL_enter_left.
  apply eratosthene_next_notprime with (p:=p); intuition auto.
  generalize H3.
  case (divides p (S n)); auto.
  ISL_unassoc_left.
  ISL_monotony.
  intros; ISL_simplify_left.
  ISL_unassoc_left.
  ISL_simplify_left.
  case x; ISL_Trivial.
  intros a; case (divides a (S n)); auto.
  constructor 2; discriminate.
Qed. 


(* Here is an ISL translation of "eratosthene" *)
Fixpoint eratosthene (St:Type) (print: nat -> DS St unit) (n:nat) {struct n} : (DS St unit) :=
  match n with
  | 0 => Skip%ISL
  | 1 => Skip%ISL
  | 2 => print 2
  | (S pn) =>
     (block true
       (eratosthene (monitorIn print (S pn)) pn -; 
        If !pL Then (Embed [pR](print (S pn))) Else Skip))%ISL
  end.

(* here is the proof of the program *)
Theorem eratosthene_proof: forall (n:nat), 2 <= n ->
 forall (St: Type) (print: nat -> DS St unit) (st:St),
  ([< st >] (eratosthene print n) >==> (absEratosthene print n))%ISL.
Proof.
  induction n.
  intros H; inversion_clear H.
  generalize IHn; clear IHn.
  case n; clear n.
  intros IHn H. absurd (2<=1).
  omega.
  auto.
  intros n; case n; clear n.
  intros IHn H St print st.
  simpl. 
  unfold absEratosthene.
  intros; ISL_simplify_right.
  intros X; clear X.
  constructor 1 with (x:=(2::nil)).
  ISL_simplify_right. 
  intuition auto with isl.
  ISL_simplify_right. 
  intros; ISL_simplify_left. 
  intros; ISL_simplify_left. 
  auto with isl.
  intros n IHn H St print st.
  change 
   ([< st >]
     (block true
       (eratosthene (monitorIn print (S (S (S n)))) (S (S n)) -; 
        If !pL Then (Embed [pR](print (S (S (S n))))) Else Skip))
     >==> (absEratosthene print (S (S (S n)))))%ISL.
  lapply IHn.
  2: omega.
  cut (2 <= (S (S n))).
  2: omega.
  generalize (S (S n)).
  clear IHn H n; intros n X IHn.
  ISL_trans.
  2: apply absEratosthene_prop; auto.
  apply block_monotony.
  ISL_monotony.
Qed.

End Prime.

