Set Implicit Arguments.
Require Export FSL.

Section ABSPOST.


Variable A: Type.

Variable q: A -> Prop.

Definition absPost: (DS A)
  := bind (any A) (fun a => bind (ensure (q a)) (fun _ => val a)).

Lemma abspost2wp: forall (s:DS A), (sdemon s q) -> (refines s absPost).
Proof.
  intros sp H1; unfold absPost.
  apply refines_fold; simpl.
  intros; eapply sdemon_weaken; eauto.
Qed.

Lemma intro_abspost:
  forall (sp1 sp2:(DS A)),
   (refines sp1 absPost)
      -> (refines (bind sp1 (fun a => bind (ensure (q a)) (fun _ => (val a)))) sp2)       
  -> (refines sp1 sp2).
Proof.
  intros sp1 sp2 H1 H. eapply refines_trans.
  2:apply H.
  apply refines_fold.
  simpl; intros H3.
  lapply (refines_unfold H1).
  intros H4; apply sdemon_fold.
  eapply sdemon_weaken.
  apply H4.
  auto.
  intros a H5.
  apply angel_fold with (a:=a); intuition auto.
  case (sdemon_unfold _ _ a H4 H5).
  simpl.
  intros x H6; decompose [ex Logic.and] H6.
  subst. simpl.  intuition auto.
  simpl; intuition auto. 
Qed.

(** This rule is a generalization of the preceding one.
*)
Lemma intro_bind_abspost_l:
  forall (B:Type) (sp1:DS A) (sp2: A -> DS B) (sp3: DS B),
    (refines sp1 absPost)
        -> (refines (bind sp1 (fun a => 
                             (bind (ensure (q a)) 
                                      (fun _ => sp2 a))))
                       sp3)
  -> (refines (bind sp1 sp2) sp3).
Proof.
  intros B sp1 sp2 sp3 H1 H3.
  eapply refines_trans.
  2:apply H3.
  apply refines_trans with
   (s2:=(bind (bind sp1 (fun a => bind (ensure (q a)) (fun _ => val a))) (fun b => sp2 b))).
   apply bind_monotony.
   eapply intro_abspost.
   apply H1.
   auto with wprefines.
   auto with wprefines.
   eapply refines_trans.
   apply bind_assoc_ref1.
   apply bind_monotony; auto with wprefines.
   intros a; apply bind_simpl_left; simpl.
   intros; apply bind_simpl_right; simpl.
   intuition auto with wprefines.
Qed.

(* dual lemma of the preceding one *)
Lemma intro_bind_abspost_r:
  forall (B:Type) (sp1:DS A) (sp2: A -> DS B) (sp3: DS B),
    (refines sp1 absPost)
        -> (refines sp3 
                        (bind sp1 (fun a => 
                             (bind (require (q a)) 
                                      (fun _ => sp2 a)))))
  -> (refines sp3 (bind sp1 sp2)).
Proof.
  intros B sp1 sp2 sp3 H1 H3.
  eapply refines_trans.
  apply H3.
  apply intro_bind_abspost_l. 
  auto.
  apply bind_monotony; auto with wprefines.
Qed.

End ABSPOST.


(** *** Rules for Hoare specifications.
*)

Lemma abspost_monotony: forall (A:Type) (q1 q2: A -> Prop),
   (forall (a:A), (q1 a) -> (q2 a)) -> 
    (refines (absPost q1) (absPost q2)).
Proof.
  intros; apply abspost2wp. 
  simpl.
  auto.
Qed.

(*
Lemma abspost_split: forall (A St:Type) (q1 q2: St -> A -> Prop) 
  (st:St) (sp:(K St A)),
   (refInEnv st sp (absPost q1))
     -> (refInEnv st sp (absPost q2))
      ->
    (refInEnv st sp (absPost (fun (stf:St) (r:A) => (q1 stf r)  /\ (q2 stf r)))).
Proof.
  intros A St q1 q2 st sp H1 H2.
  eapply intro_abspost.
  apply H2.
  SSL_enter_left.
  apply H1.
  intros; eapply abspost2wp. 
  simpl.
  intuition auto.
Qed.

Lemma abspost_splitforall: forall (A B St:Type) (p: B -> Prop) (q: B -> St -> A -> Prop) 
  (st:St) (sp:(K St A)),
   (exists x:B, (p x)) -> 
   (forall (x:B), (p x) -> (refInEnv st sp (absPost (q x)))) ->
    (refInEnv st sp (absPost (fun (stf:St) (r:A) => forall (x:B), (p x) -> (q x stf r)))).
Proof.
  intros.
  apply abspost2wp. 
  case H; clear H.
  intros x H; apply sdemon_fold.  
  lapply (refines_unfold (H0 x H)); simpl; intros; eauto with wipi.
  intros a; case a; clear a.
  intros s a H1 x0 H2; intros.
  lapply (refines_unfold (H0 x0 H2)); simpl; auto.
  intros H3; lapply (sdemon_unfold _ _ (pairT s a) H3); auto.
  clear H3; intros H3.
  decompose [ex Logic.and] H3.
  inversion H6. 
  subst; auto.
Qed.
*)

Unset Implicit Arguments.
(* end hide *)
