Set Implicit Arguments.

Require Export FSLAbsPost.

Definition pred (n:nat): DS nat
 := match n with
    | 0 =>  abort nat 
    | (S p) => val p
    end.

Fixpoint minus (n m:nat) {struct m}: DS nat
  := match m with
     | 0 => val n
     | (S p) => bind (pred n) (fun n' => minus n' p)
     end. 


Lemma minus_test1:
  forall (n:nat), refines (minus (500+n) 500) (val n).
Proof.
  intros n; apply refines_fold.
  vm_compute.
  intuition; auto.
Qed.

Lemma minus_test2:
  forall (n:nat) (sp: DS nat),
    refines sp (minus 100 (500 + n)).
Proof.
  intros n sp; apply refines_fold.
  vm_compute.
  intuition; auto.
Qed.

(*
Goal forall (n:nat), refines (minus 100 (500+n)) (any nat).
  intros n sp; apply refines_fold.
  vm_compute.
  intuition; auto.
Abort.  
*)

Lemma minus_test3:
  forall (n:nat),
    ~refines (minus 100 (500+n)) (any nat).
Proof.
  intros n H; generalize (refines_unfold H).
  vm_compute.
  auto.
Qed.

Lemma minus_correctness: forall (m n:nat), m <= n ->
  refines (minus n m)
          (absPost (fun k => n=k+m)).
Proof.
  induction m; simpl.
  intros; apply abspost2wp; simpl. auto.
  intros n; case n; simpl.
  intros H; inversion H.
  clear n; intros n H; apply bind_simpl_left; simpl.
  eapply refines_trans.
  apply IHm. omega.
  apply abspost2wp; simpl. intros; omega.
Qed.

Lemma minus_exact: forall (m n:nat), refines (bind (require (m<=n)) (fun _ => absPost (fun k => n=k+m)))
                                             (minus n m).
Proof.
  induction m; simpl.
  intros; apply refines_fold; simpl. intuition; intros; omega.
  intros n; apply bind_simpl_right; case n; simpl; auto.
  intros; eapply refines_trans.
  2:apply IHm.
  apply bind_monotony.
  apply refines_fold; simpl; intuition omega.
  intros; apply abspost2wp; simpl. intros; omega.
Qed.


Lemma minus_plus_l: forall (n m:nat), refines (minus (n+m) m) (val n).
Proof.
  intros; eapply refines_trans.
  apply minus_correctness; omega. 
  apply refines_fold.
  simpl. intros; omega.
Qed.

Lemma plus_minus_r: forall (n m:nat), m <= n -> 
  refines (bind (minus n m) (fun k => val (k+m))) (val n).
Proof.
  intros; eapply refines_trans.
  eapply bind_monotony_left. apply minus_correctness; auto. 
  apply refines_fold.
  simpl; auto.
Qed.

Lemma minus_minus: forall (n m p:nat), refines (bind (minus n m) (fun k => minus k p)) (minus n (m+p)).
Proof.
   intros; eapply refines_trans.
   2: apply minus_exact.
   apply bind_simpl_right.
   simpl; intuition.
  intros; eapply refines_trans.
  eapply bind_monotony_left. apply minus_correctness; omega.
  apply bind_simpl_left; simpl.
  intros; eapply refines_trans.
  apply minus_correctness; omega.
  intros; apply abspost2wp; simpl. intros; omega.
Qed.

Require Import List.

Definition foldl (A:Type) (f: A -> A -> DS A) (init:A) (l:list A) : DS A
 := fold_left (fun acc b => bind acc (fun a => f a b)) l (val init).

Open Scope list_scope.

Lemma minus_ho_test: refines (foldl minus 27 (3 :: 5 :: 2 :: nil)) (val 17).
Proof.
  intros; apply refines_fold; simpl.
  auto.
Qed. 

Lemma pred_impl: forall n, impl (pred n).
Proof.
  intros n; case n; simpl; auto with wpimpl.
Qed.

Hint Resolve pred_impl: wpimpl.

Lemma minus_impl: forall n m, impl (minus n m).
Proof.
  intros n m; generalize m n; clear m n; induction m; simpl; auto with wpimpl. 
Qed.

Hint Resolve minus_impl: wpimpl.


Extraction Inline pred minus.

Extract Inductive sumbool => "bool" [ "true" "false" ].
(* Extract Inductive DS "'a" => "unit" [ "()" ].*)

Extraction minus_impl. 

(*** Exemple de points fixe ***)

Set Implicit Arguments.

Require Export FSLTarski.
Require Export FSLAbsPost.

Require Export Arith.

Definition div_eq  (b:nat) (div: nat -> DS nat) (a:nat)
 := if (le_gt_dec b a) then 
      bind (minus a b) (fun r0 => 
           bind (div r0) (fun r => val (S r))) 
    else (val 0).

(* div in partial correctness *)
Definition div (a b:nat):= sfixF (div_eq b) a.

Definition div_abs a b q := q*b <= a < (S q)*b.

Lemma div_spec:
   forall a b, refines (div a b) (absPost (div_abs a b)).
Proof.
  unfold div; intros; sfixF_mmagic.
  unfold div_eq, div_abs.
  intros x; case (le_gt_dec b x).
  intros H. eapply refines_trans.
   eapply bind_monotony_left.
   apply minus_correctness; auto.
   apply abspost2wp; simpl.
   intros; subst; omega.
  intros; apply abspost2wp; simpl. omega.
Qed.

Lemma div_eq_monotony:
  forall b (f g: nat -> DS nat), (forall x, refines (f x) (g x)) -> 
    forall a,  refines (div_eq b f a) (div_eq b g a).
Proof.
  intros; unfold div_eq; intros.
  case (le_gt_dec b a); auto with wprefines.
Qed.

Hint Resolve div_eq_monotony: div.

Lemma fold_left_monotony: 
 forall (A B:Type) (f g: DS A -> B -> DS A), 
   (forall x1 x2 y, refines x1 x2 -> refines (f x1 y) (g x2 y))
    -> forall (l:list B) (x y:DS A), refines x y -> refines (fold_left f l x) (fold_left g l y).
Proof.
  induction l; simpl; auto with wprefines.
Qed.

Hint Resolve fold_left_monotony: wprefines.

Lemma foldl_monotony: 
 forall (A:Type) (f g: A -> A -> DS A), 
   (forall x y, refines (f x y) (g x y))
    -> forall (init:A) (l:list A), refines (foldl f init l) (foldl g init l).
Proof.
  unfold foldl; intros; auto with wprefines.
Qed.

Lemma div_ho_test: refines (foldl div 1234 (10 :: 5 :: 2 :: nil)) (val 12).
Proof.
  intros; unfold div.
  eapply refines_trans.
  eapply foldl_monotony with (g:=fun a b => iter a (div_eq b) (sfixF (div_eq b)) a).
  intros; apply itersfixF_postfix; auto with div.  
  apply refines_fold.
  vm_compute.
  auto.
Qed.


(* div in total correctness *)
Require Import FSLFixF.
Require Import Zwf.

Lemma div_cd: forall b, 0<b -> (call_decreasing (div_eq b) lt).
Proof.
  unfold call_decreasing.
  intros b H divr a; unfold div_eq.
   case (le_gt_dec b a); auto with wprefines.
   intros;
    eapply intro_bind_abspost_l.
    apply minus_correctness; auto.
    apply bind_monotony; auto with wprefines.
    intros r;
     eapply refines_trans.
     2: apply bind_assoc_ref2.
     apply bind_monotony; auto with wprefines.
     apply refines_fold; simpl.
     intuition.
Qed.

Hint Resolve div_cd: div.

Lemma div_eq_impl: 
 forall b (f: nat -> DS nat), 
  (forall a, impl (f a))
  -> forall a, impl ((div_eq b) f a).
Proof.
  unfold div_eq.
  intros b divr H a; case (le_gt_dec b a); auto with wpimpl.
Qed.

Hint Resolve div_eq_impl: div.

Lemma div_impl: forall a b, 0 < b -> impl (div a b).
Proof.
  intros; unfold div; eapply sfixF_impl; auto with div.
  apply lt_wf.
Qed.

Extraction Inline div div_eq_impl.

Extraction "div.ml" div_impl.

Lemma div_exact:
   forall a b, 0<b -> refines (absPost (div_abs a b)) (div a b).
Proof.
  unfold div; intros; sfixF_mabort; auto with div.
  apply lt_wf.
  unfold div_eq, div_abs.
  intros x; case (le_gt_dec b x).
  intros. 
  (* TODO: emballer la preuve "technique" ci-dessous dans un lemme gal  *)
  eapply intro_bind_abspost_r.
  apply minus_correctness; auto.  
  apply bind_simpl_right; intros D; apply impl_angel_fold with (H:=(minus_impl x b)) (H1:=D);
  generalize (extract (minus_impl x b D)); clear D.
  (* FIN TODO *)
  intros n; apply bind_simpl_right; simpl; intuition subst.
  apply refines_fold; simpl.
  intros X; clear X.
  intros y; case y; simpl.
  intros; assert False; intuition.
  intros py Hy; constructor 1 with (x:=py).
  omega.
  intros; apply refines_fold; simpl.
  intros D a0; case a0; simpl; auto.
  intros n; generalize (n*b).
  intros; cut False; intuition.
Qed.
