Set Implicit Arguments.

Require Export FSLTarski.
Require Export FSLAbsPost.

Require Export ZArith.
Open Scope Z_scope.

Definition _100 : Z := 100.
Definition _11 : Z := 11.
Definition _10 : Z := 10.

Definition f91_eq  (f91: Z-> (DS Z)) (n:Z)
 := if (Z_le_gt_dec n _100) then 
         bind (f91 (n+ _11)) f91
      else (val (n- _10)).

(* f91 in partial correctness *)
Definition f91_p := sfixF f91_eq.


Definition f91_spec (n:Z) : Z :=
   if (Z_le_gt_dec n 100) then 91 else (n-10).

Lemma f91_correctness:
   forall n, refines (f91_p n) (val (f91_spec n)).
Proof.
  unfold f91_p; intros; sfixF_mmagic.
  unfold f91_eq, f91_spec, _100, _11, _10.
  intros x; case (Z_le_gt_dec x 100); auto with wprefines.
  intros H; case (Z_le_gt_dec (x+11) 100); auto with wprefines.
  intros; apply refines_fold; simpl; auto.
  intros; apply refines_fold; simpl.
  case (Z_le_gt_dec (x+11-10) 100); auto. 
  intros; omega.
Qed.

Require Import FSLFixF.
Require Import Zwf.

(* f91 in total correctness: we need an other fixpoint equation *)
Definition f91_eq2  (f91: Z-> (DS Z)) (n:Z)
 := if (Z_le_gt_dec n _100) then 
         bind (f91 (n+ _11)) (fun r => 
            bind (require (r = f91_spec (n+11))) (fun _ =>
               f91 r))
      else (val (n- _10)).

Definition f91 := sfixF f91_eq2.

(* First, we prove the equivalence between f91 and f91_p *)

Lemma f91_eq2_monotony:
  forall (f g: Z -> (DS Z)), (forall x, refines (f x) (g x)) -> 
    forall x,  refines (f91_eq2 f x) (f91_eq2 g x).
Proof.
  intros; unfold f91_eq2; intros.
  case (Z_le_gt_dec x _100); auto with wprefines.
Qed.

Hint Resolve f91_eq2_monotony: f91.

(* the trivial part *)
Lemma f91_triv:
   forall n, refines (f91_p n) (f91 n).
Proof.
  unfold f91_p; intros. sfixF_mmagic.
  intros x; unfold f91.
  eapply refines_trans.
  2: apply sfixF_prefix; auto with f91.
  unfold f91_eq2 at 2, f91_eq at 1.
  case (Z_le_gt_dec x _100); auto with wprefines.
Qed.   

Lemma f91_eq_monotony:
  forall (f g: Z -> (DS Z)), (forall x, refines (f x) (g x)) -> 
    forall x,  refines (f91_eq f x) (f91_eq g x).
Proof.
  intros; unfold f91_eq; intros.
  case (Z_le_gt_dec x _100); auto with wprefines.
Qed.

Hint Resolve f91_eq_monotony: f91.

(* the less trivial part *)
Lemma f91_refines_f91_p:
   forall n, refines (f91 n) (f91_p n).
Proof.
  unfold f91; intros; sfixF_mmagic.
  intros x; unfold f91_p.
  eapply refines_trans.
  2:apply sfixF_prefix; auto with f91.
  unfold f91_eq at 2, f91_eq2.
  case (Z_le_gt_dec x _100); auto with wprefines.
  intros; apply intro_bind_abspost_l with (q:=fun r => r=f91_spec(x+11)).
  unfold absPost; eapply bind_simpl_right; simpl.
  intros D; clear D.
  eapply ex_intro.
  eapply refines_trans.
  eapply f91_correctness.
  eauto with wprefines.
  auto with wprefines.
Qed.

(* now, we prove that f91 terminates *)
Definition max (x y:Z) : Z :=
  if (Z_le_gt_dec x y) then y else x.

Definition f91_var (y x:Z) := 
  (Zwf 0 (max 0 (101-y)) (max 0 (101-x))).  

Lemma f91_var_wf:well_founded f91_var.
Proof.
  unfold f91_var.
  apply wf_inverse_image with (f:=fun x => max 0 (101-x)).
  apply Zwf_well_founded.
Qed.

Lemma f91_cd: call_decreasing f91_eq2 f91_var.
Proof.
  unfold call_decreasing.
  intros f x; unfold f91_eq2, _100, _11, _10.
   case (Z_le_gt_dec x 100); auto with wprefines.
   intros;
    eapply refines_trans.
    2: apply bind_assoc_ref2.
   intros; apply bind_simpl_right; simpl.
   intros D; clear D.
   intuition.
   unfold f91_var, Zwf, max.
   case (Z_le_gt_dec 0 (101-x)); intuition (try omega).
   case (Z_le_gt_dec 0 (101-(x+11))); intros; omega.
   apply bind_monotony; auto with wprefines.
   intros; apply bind_simpl_right; simpl.
   intuition subst.
   apply bind_monotony; auto with wprefines.
   apply refines_fold; simpl.
   unfold f91_var, f91_spec, Zwf, max.
  case (Z_le_gt_dec 0 (101-x)); intuition (try omega).
  case (Z_le_gt_dec (x+11) 100); intuition (try omega).
  case (Z_le_gt_dec 0 (101 - 91)); intuition (try omega).
  case (Z_le_gt_dec 0 (101 - (x + 11 - 10))); intuition (try omega).
Qed.

Hint Resolve f91_cd: f91.

Lemma f91_eq2_impl: 
 forall (f: Z -> DS Z), 
  (forall n, impl (f n))
  -> forall n, impl (f91_eq2 f n).
Proof.
  unfold f91_eq2.
  intros; case (Z_le_gt_dec n _100); auto with wpimpl.
Qed.

Hint Resolve f91_eq2_impl: f91.

Lemma f91_impl: forall n, impl (f91 n).
Proof.
  intros; unfold f91; eapply sfixF_impl; auto with f91.
  2: apply f91_cd.
  apply f91_var_wf.
Qed.

Extraction Inline f91_eq2_impl.

Extract Inductive sumbool => "bool" [ "true" "false" ].
Extract Inductive unit => "unit" [ "()" ].

Extraction (* "f91.ml" *) f91_impl.
