(* Without the hypothesis of monotony (Tarski), fixpoints may not
 * exist.  Here is an example.
 *)

Set Implicit Arguments.

Require Export FSL.

Lemma magic_not_refines_abord: ~(refines (abort unit) (magic unit)).
Proof.
  intros H; apply (refines_unfold H); simpl; auto.
Qed.

(** example of non-monotonic **)

Definition absurd1_eq (x:(DS unit))
 := (bind (require (refines (abort unit) x))
          (fun _ => (bind (ensure (refines x (magic unit))))
                   (fun _ => x))).

Section Absurd1.

Variable absurd1: (DS unit).

Hypothesis absurd1_fix1: 
 (refines (absurd1_eq absurd1) absurd1).

Hypothesis absurd1_fix2: 
 (refines absurd1 (absurd1_eq absurd1)).

Lemma abort_refines_absurd1: (refines (abort unit) absurd1).
Proof.
  eapply refines_trans.
  2: apply  absurd1_fix1; auto with wprefines.
  unfold absurd1_eq.
  apply bind_simpl_right; simpl; intuition auto with wprefines.
  apply bind_simpl_right; simpl; intuition auto with wprefines.
  eapply refines_trans.
  apply absurd1_fix2.  
  unfold absurd1_eq.
  apply bind_simpl_left; simpl; intuition auto with wprefines.
Qed.

Hint Resolve abort_refines_absurd1: wprefines.

Lemma absurd1_refines_magic: (refines absurd1 (magic unit)).
Proof.
  eapply refines_trans.
  apply absurd1_fix2.  
  unfold absurd1_eq.
  apply bind_simpl_left; simpl; intuition auto with wprefines.
Qed.

Lemma absurd1_proof: False.
Proof.
  apply  magic_not_refines_abord.
  eapply refines_trans.
  apply abort_refines_absurd1.
  apply absurd1_refines_magic.
Qed.


End Absurd1.


(* Here is a counter-example that the converse property of sfixDS2_refines_G_sfixDS1 
does not hold.
*)
Require Export FSLTarski.



Definition Id (x:DS unit) : DS unit := x.

Definition Abort (x:DS unit) : DS unit := (abort unit).

Section Absurd2.

Lemma sfixId_magic: (refines (sfixDS Id) (magic unit)).
Proof.
  intros; apply sfixDS_mostmagic.
  simpl; auto with wprefines.
Qed.

Lemma abort_Abort: forall (x:DS unit), 
 (refines (abort unit) (Abort x)).
Proof.
   simpl; auto with wprefines.
Qed.

Hypothesis G_sfixDS1_refines_sfixDS2:
 (forall (x y:DS unit), (refines x y) -> (refines (Id x) (Id y)))
  -> (forall (x y:DS unit), (refines x y) -> (refines (Abort x) (Abort y)))
  -> (forall (x: DS unit), (refines (Id (Abort x)) (Abort (Id x))))
    -> (refines (Abort (sfixDS Id)) (sfixDS Id)).

Lemma absurd2: False.
Proof.
  apply  magic_not_refines_abord.
  eapply refines_trans.
  apply abort_Abort.
  eapply refines_trans.
  apply G_sfixDS1_refines_sfixDS2; simpl; auto with wprefines.
  apply sfixId_magic.
Qed.


End Absurd2.


