Set Implicit Arguments.

Require Export FSLFixF.


Section Fix.

Variable A B: Type.

Variable F: (DS A) -> (DS A).
Variable obs: (DS B).
Variable R: B -> B -> Prop.

Hypothesis obs_refines_any: (refines obs (any B)).


Hypothesis F_monotony:
  forall (x y:(DS A)), (refines x y) -> (refines (F x) (F y)).

Hypothesis Rwf : (well_founded R).


Definition auxF (f:(B -> (DS A))) (x:B) := 
  (F (bind obs (fun y => (bind (ensure (R y x)) (fun _ => (f y)))))).

Lemma auxF_monotony :
  forall (f g: B -> (DS A)),
   (forall (x:B), (refines (f x) (g x)))
      -> forall (x:B), (refines (auxF f x) (auxF g x)).
Proof.
  unfold auxF; intros; auto with wprefines.
Qed.

Lemma auxF_decreasing: (call_decreasing auxF R).
Proof.
  unfold call_decreasing, auxF; auto 15 with wprefines. 
Qed.

Hint Resolve auxF_monotony auxF_decreasing: wprefines.

Definition fixDS:=(bind obs (fixF auxF Rwf)).

Lemma fixDS_postfix:
   (refines fixDS (F fixDS)).
Proof.
  unfold fixDS; apply refines_trans with (s2:=(bind (any B) (fixF auxF Rwf)));
  auto with wprefines.
  apply any_refines; intros x.
  eapply refines_trans. 
  apply fixF_postfix; auto with wprefines.
  unfold auxF at 1; auto with wprefines.
Qed.

Theorem fixDS_mostmagic:
  forall (fixO: (DS A)), 
    (refines (F fixO) fixO)
     -> (refines fixDS fixO).
Proof.
  intros fixO H; unfold fixDS.
  apply refines_trans with (s2:=(bind (any B) (fixF auxF Rwf))); 
  auto with wprefines.
  apply any_refines; intros x.
  apply fixF_mostmagic with (fixO:=(fun (x:B) => fixO));
  auto with wprefines.
  intros y; unfold auxF.
  eapply refines_trans. 2: apply H.
  apply F_monotony.
  apply refines_trans with (s2:=(bind (any B) (fun _ => fixO))).
  auto with wprefines.
  auto with wprefines.
Qed.

Definition obs_decreasing :=
  forall (f:B -> (DS A)), (refines (F (bind obs f)) 
                                 (bind obs 
                                    (fun x => 
                                        (F (bind obs 
                                             (fun y =>
                                               (bind (ensure (R y x))
                                                   (fun _ => (f y))))))))).

Hypothesis F_decreasing: obs_decreasing.

Lemma fixDS_prefix:
   (refines (F fixDS) fixDS).
Proof.
  unfold fixDS; eapply refines_trans.
  unfold obs_decreasing in F_decreasing; apply F_decreasing.
  apply bind_monotony; auto with wprefines.
  intros x; eapply refines_trans.
  2: apply fixF_prefix; auto with wprefines.
  unfold auxF at 2; auto with wprefines.
Qed. 

Lemma iterfixDS_prefix:
  forall (n:nat), 
     (refines (iter n F fixDS) fixDS). 
Proof.
  induction n; simpl; auto with wprefines.
  eapply refines_trans.
  2: apply fixDS_prefix.
  auto.
Qed.

Section Inv_FixDS_ensure.
                                                                                                
Variable inv: (DS A) -> Type.
                                                                                                
Hypothesis inv_compat: forall (s1 s2: (DS A)),
  (inv s1) -> (refines s1 s2) -> (refines s2 s1) -> (inv s2).
                                                                                                
Hypothesis inv_obs: forall (f: B -> (DS A)),
  (forall (b:B), (inv (f b))) -> (inv (bind obs f)).
                                                                                                
Hypothesis inv_ensure: forall (Q: Prop) (s:(DS A)),
      (Q -> (inv s)) -> (inv (bind (ensure Q) (fun _ => s))).
                                                                                                
Hypothesis inv_F: (forall (s: (DS A)), (inv s) -> inv (F s)).
                                                                                                
Lemma inv_fixDS_ensure: inv fixDS.
Proof.
  unfold fixDS; apply inv_obs; intros x.
  pattern (fixF auxF Rwf), x; apply inv_fixF_ensure; clear x;
  eauto with wprefines.
  unfold auxF; auto.
Qed.
                                                                                                
End Inv_FixDS_ensure.

Section Inv_FixDS.
 
Variable inv: (DS A) -> B -> Type.
 
Hypothesis inv_obs: forall (f: B -> (DS A)) (x:B),
  (inv (f x) x) -> (inv (bind obs f) x).
 
Hypothesis obs_inv: forall (f: B -> (DS A)) (x:B),
  (inv (bind obs f) x) -> (inv (f x) x).
 
Hypothesis inv_compat: forall (s1 s2: B -> (DS A)) (x:B),
  (inv (s1 x) x) -> (refines (s1 x) (s2 x)) -> (refines (s2 x) (s1 x))
   -> (inv (s2 x) x).

Hypothesis inv_F: forall (s: (DS A)), (forall (x:B), (inv s x)) -> 
 forall (x:B), inv (F s) x.
 
Lemma inv_fixDS_require:
  (forall (Q: Prop) (s:(DS A)) (x:B),
      (Q -> (inv s x)) -> (inv (bind (require Q) (fun _ => s)) x))
      -> forall (x:B), inv fixDS x.
Proof.
  intros inv_require; unfold fixDS. intros x; apply inv_obs.
  pattern (fixF auxF Rwf), x; apply inv_fixF_require; clear x;
  auto with wprefines.
  intros; eapply inv_compat; eauto.
  unfold auxF; intros.
  apply obs_inv with (f:=fun x => (F (bind obs (fun y : B => bind (ensure (R y x)) (fun _ : unit => f y))))).
  unfold obs_decreasing in F_decreasing.
  apply inv_compat with
   (s2:=fun (_:B) => (bind obs (fun x => (F (bind obs (fun y : B => bind (ensure (R y x)) (fun _ : unit => f y)))))))
   (s1:=fun (_:B) => (F (bind obs f))); auto with wprefines.
  apply refines_trans with (s2:=(bind (any B) (fun x => (F (bind obs (fun y : B
=> bind (ensure (R y x)) (fun _ : unit => f y)))))));
  auto with wprefines.
Qed.
                                                                                
Lemma inv_fixDS_ensure2:
  (forall (Q: Prop) (s:(DS A)) (x:B),
      (Q -> (inv s x)) -> (inv (bind (ensure Q) (fun _ => s)) x))
      -> forall (x:B), inv fixDS x.
Proof.
  intros inv_ensure; unfold fixDS; intros x; apply inv_obs.
  pattern (fixF auxF Rwf), x; apply inv_fixF_ensure; clear x;
  auto with wprefines.
  intros; eapply inv_compat; eauto.
  unfold auxF; auto.
Qed.
                                                                                
End Inv_FixDS.


End Fix.


Section Monotony.

Variable A B1 B2: Type.

Variable F1 F2: (DS A) -> (DS A).

Variable obs1: (DS B1).
Hypothesis obs1_refines_any: (refines obs1 (any B1)).

Variable obs2: (DS B2).
Hypothesis obs2_refines_any: (refines obs2 (any B2)).

Hypothesis F1_monotony:
  forall (x y:(DS A)), (refines x y) -> (refines (F1 x) (F1 y)).

Hypothesis F2_monotony:
  forall (x y:(DS A)), (refines x y) -> (refines (F2 x) (F2 y)).

Variable R1: B1 -> B1 -> Prop.
Variable R2: B2 -> B2 -> Prop.
Hypothesis R1wf : (well_founded R1).
Hypothesis R2wf : (well_founded R2).

Hypothesis F1_refines_F2:
  forall (x: (DS A)), (refines (F1 x) (F2 x)).

Hypothesis F_decreasing2: (obs_decreasing F2 obs2 R2).

Lemma fixDS1_refines_fixDS2: (refines (fixDS F1 obs1 R1wf) (fixDS F2 obs2 R2wf)).
Proof.
  apply fixDS_mostmagic; auto.
  eapply refines_trans.
  2: apply fixDS_prefix; auto.
  auto.
Qed.

End Monotony.


Section Morph.

Variable A1 A2 B1 B2: Type.

Variable F1: (DS A1) -> (DS A1).
Variable F2: (DS A2) -> (DS A2).
Variable G: (DS A1) -> (DS A2).

Hypothesis F1_monotony:
  forall (x y:(DS A1)), (refines x y) -> (refines (F1 x) (F1 y)).

Hypothesis F2_monotony:
  forall (x y:(DS A2)), (refines x y) -> (refines (F2 x) (F2 y)).

Hypothesis G_monotony:
  forall (x y:(DS A1)), (refines x y) -> (refines (G x) (G y)).

Variable obs1: (DS B1).
Hypothesis obs1_refines_any: (refines obs1 (any B1)).

Variable obs2: (DS B2).
Hypothesis obs2_refines_any: (refines obs2 (any B2)).
 
Variable R1: B1 -> B1 -> Prop.
Hypothesis R1wf : (well_founded R1).

Variable R2: B2 -> B2 -> Prop.
Hypothesis R2wf : (well_founded R2).

Lemma fix2_refines_G_fix1:
 (obs_decreasing F1 obs1 R1)
  -> (forall (x:(DS A1)), (refines (F2 (G x)) (G (F1 x))))
    -> (refines (fixDS F2 obs2 R2wf) (G (fixDS F1 obs1 R1wf))).
Proof.
  intros F1_decreasing F2G_refines_GF1; apply fixDS_mostmagic; auto.
  eapply refines_trans.
  apply F2G_refines_GF1.
  apply G_monotony.
  apply fixDS_prefix; auto.
Qed.

(*
Hypothesis F2_decreasing: (obs_decreasing F2 R2).

Hypothesis GF1_refines_F2G: 
  (forall (x:(DS A1)), (refines (G (F1 x)) (F2 (G x)))).

Hypothesis G_decreasing: (obs_decreasing G R1).

Lemma G_fix1_refines_fix2:
      (refines (G (fixDS F1 R1wf)) (fixDS F2 R2wf)).
Proof.
  unfold fixDS at 1. unfold obs_decreasing in G_decreasing.
   eapply refines_trans.
   apply G_decreasing.
   apply any_refines; intros x.
   apply refines_trans with  (s2:=(G (iterWF F1 (R1wf x)))).
   apply G_monotony.
   case (R1wf x).
   intros x0 h0; apply any_refines; intros y0; 
   apply bind_simpl_left; simpl; intros h1.
   eapply refines_trans.
   apply iterWF_refines_F; auto.
   apply F1_monotony.
   apply any_refined_bind with (a:=y0).
   apply any_refined_bind with (a:=h1).
   apply iterWF_non_acc_dep; auto.
   generalize (R1wf x); intro r;  
   elim r using Acc_ind_dep; clear r x; simpl.
   intros x H1 H2.
   eapply refines_trans.
   apply GF1_refines_F2G.
   eapply refines_trans.
   2: apply fixDS_prefix; auto.
   apply F2_monotony.
*) 

End Morph.


Unset Implicit Arguments.
