(*** Exemples de points fixes ***)

Set Implicit Arguments.

Require Export FSLTarski.

(*** even1 ***)
Definition even1_eq (ev: (DS nat)) 
 := (or (val O) (bind ev (fun n => (val (S (S n)))))).

Definition even1 := (sfixDS even1_eq).

Lemma even1_eq_monotony:
  forall (x y:(DS nat)), (refines x y) -> (refines (even1_eq x) (even1_eq y)).
Proof.
  intros; unfold even1_eq; auto with wprefines.
Qed.

Hint Resolve  even1_eq_monotony: wprefines.

Lemma sixteen_refines_even1: (refines (val 16) even1).
Proof.
  unfold even1; eapply refines_trans.
  2:apply iterFsfixDS_refines_sfixDS with (n:=9); auto with wprefines.
  intros; apply refines_fold; simpl; intuition auto.
Qed.

(*** odd1 ***)

Definition odd1_eq (ev: (DS nat)) 
 := (or (val 1) (bind ev (fun n => (val (S (S n)))))).

Definition odd1 := (sfixDS odd1_eq).

Lemma odd1_eq_monotony:
  forall (x y:(DS nat)), (refines x y) -> (refines (odd1_eq x) (odd1_eq y)).
Proof.
  intros; unfold odd1_eq; auto with wprefines.
Qed.

Hint Resolve odd1_eq_monotony: wprefines.

Lemma odd1_refines_Seven1:
  (refines odd1 (bind even1 (fun n => (val (S n))))).
Proof.
  unfold odd1, even1; 
  apply sfixDS2_refines_G_sfixDS1
    with (G:= (fun even1 => (bind even1 (fun n => (val (S n)))))); 
  auto with wprefines.
  unfold odd1_eq, even1_eq.
  intros x.
  apply refines_or_elim; auto with wprefines.
  apply refines_trans with (s2:= (bind (val O)  (fun n : nat => val (S n))));
  auto with wprefines.
  apply bind_simpl_right; auto with wprefines; simpl; auto with wprefines.
  apply refines_trans with (s2:= 
    (bind (bind x (fun n : nat => val (S (S n)))))
        (fun n : nat => val (S n)));
  auto with wprefines.
  eapply refines_trans.
  2: apply bind_assoc_ref2.
  eapply refines_trans.
  apply bind_assoc_ref1.
  apply bind_monotony; auto with wprefines.
  intros; apply refines_fold; simpl; auto.
Qed.

Lemma even1_refines_Sodd1:
  (refines even1 (or (val 0) (bind odd1 (fun n => (val (S n)))))).
Proof.
  unfold odd1, even1; 
  apply sfixDS2_refines_G_sfixDS1
    with (G:=(fun odd1 =>  (or (val 0) (bind odd1 (fun n => (val (S n)))))));
  auto with wprefines.
  unfold odd1_eq, even1_eq.
  intros x.
  apply refines_or_elim; auto with wprefines.
  apply refines_or_intro_r.
  eapply refines_trans.
  apply bind_or_distr_r_1.
  apply refines_or_elim.
  apply refines_trans with (s2:= (bind (val 1)  (fun n : nat => val (S n))));
  auto with wprefines.
  apply refines_fold; simpl; auto.
  apply refines_trans with (s2:= 
    (bind (bind x (fun n : nat => val (S (S n)))))
        (fun n : nat => val (S n)));
  auto with wprefines.
  eapply refines_trans.
  2: apply bind_assoc_ref2.
  eapply refines_trans.
  apply bind_assoc_ref1.
  apply bind_monotony; auto with wprefines.
  intros; apply refines_fold; simpl; auto.
Qed.

Hint Resolve even1_refines_Sodd1: wprefines.


Lemma Seven1_refines_odd1:
  (refines (bind even1 (fun n => (val (S n)))) odd1).
Proof.
  apply refines_trans with (s2:=
   (bind (or (val 0) (bind odd1 (fun n => (val (S n))))))
             (fun n => (val (S n)))).
  auto with wprefines.
  eapply refines_trans.
  apply bind_or_distr_r_1.
  apply refines_or_elim.
  unfold odd1; eapply refines_trans.
  2:apply sfixDS_prefix; auto with wprefines.
  fold odd1; unfold odd1_eq; simpl.
  apply bind_simpl_left; simpl; auto with wprefines.
  unfold odd1; eapply refines_trans.
  2:apply sfixDS_prefix; auto with wprefines.
  fold odd1; unfold odd1_eq; simpl.
  apply refines_or_intro_r.
  eapply refines_trans.
  apply bind_assoc_ref1.
  apply bind_monotony; auto with wprefines.
  intros; apply refines_fold; simpl; auto.
Qed.

Require Omega.

(*** even2 ***)
Lemma even1_refines_ext:
  (refines even1 (bind (any nat) 
                   (fun n => val (n+n)))).
Proof.
  unfold even1; apply sfixDS_mostmagic; auto with wprefines.
  unfold even1_eq; simpl.
    apply refines_or_elim.
    apply any_refined_bind with (a:=0).
    apply  refines_fold; simpl. firstorder auto.
  eapply refines_trans.
  apply bind_assoc_ref1.
  apply any_refines; intros n.
  apply bind_simpl_left; simpl.
  apply any_refined_bind with (a:=(S n)).
  apply refines_fold; simpl; auto.
Qed.

Lemma ext_refines_even1:
  (refines (bind (any nat) (fun n => val (n+n))) even1).
Proof.
  intros; apply any_refines; intros n. eapply refines_trans.
  2: unfold even1; apply iterFsfixDS_refines_sfixDS with (n:=(S n));
  auto with wprefines.
  fold even1; subst.
  elim n; simpl; clear n.
  unfold even1_eq; auto with wprefines.
  intros n HR; unfold even1_eq at 1.
  apply refines_or_intro_r.
  apply refines_trans with 
    (s2:=(bind (val (n + n)) (fun n0 : nat => val (S (S n0)))));
  auto with wprefines.
  apply refines_fold; simpl.
  intros; omega.
Qed.

Definition even2_eq (ev: (DS nat)) 
 := (or (or (val O) (val (S (S O)))) 
        (bind ev (fun n => (bind ev (fun m => (val (plus n m))))))).

Definition even2 := (sfixDS even2_eq).

Lemma even2_eq_monotony:
  forall (x y:(DS nat)), (refines x y) -> (refines (even2_eq x) (even2_eq y)).
Proof.
  intros; unfold even2_eq; auto with wprefines.
Qed.

Hint Resolve even2_eq_monotony:  wprefines.

Lemma sixteen_refines_even2: (refines (val 16) even2).
Proof.
  eapply refines_trans.
  2: unfold even2; apply sfixDS_prefix; auto with wprefines.
  fold even2; unfold even2_eq.
  apply refines_or_intro_r.
  cut  (refines (val 8) even2).
  intros H.
  apply refines_trans with 
    (s2:= (bind (val 8) 
      (fun n : nat => bind (val 8) (fun m : nat => val (n + m)))));
  auto with wprefines.
  intros; apply refines_fold; simpl; auto.
  (*** 8 ***)
  eapply refines_trans.
  2: unfold even2; apply sfixDS_prefix; auto with wprefines.
  fold even2; unfold even2_eq.
  apply refines_or_intro_r.
  cut  (refines (val 4) even2).
  intros H.
  apply refines_trans with 
    (s2:= (bind (val 4) 
      (fun n : nat => bind (val 4) (fun m : nat => val (n + m)))));
  auto with wprefines.
  intros; apply refines_fold; simpl; auto.
  (*** 4 ***)
  eapply refines_trans.
  2: unfold even2; apply sfixDS_prefix; auto with wprefines.
  fold even2; unfold even2_eq.
  apply refines_or_intro_r.
  cut  (refines (val 2) even2).
  intros H.
  apply refines_trans with 
    (s2:= (bind (val 2) 
      (fun n : nat => bind (val 2) (fun m : nat => val (n + m)))));
  auto with wprefines.
  intros; apply refines_fold; simpl; auto.
  (*** 2 ***)
  eapply refines_trans.
  2: unfold even2; apply sfixDS_prefix; auto with wprefines.
  fold even2; unfold even2_eq; auto with wprefines.
Qed.

Lemma even2_refines_ext:
  (refines even2 (bind (any nat) 
                   (fun n => (bind (ensure (exists y:nat, n=2*y))
                     (fun _ => (val n)))))).
Proof.
  unfold even2; apply sfixDS_mostmagic; auto with wprefines.
  unfold even2_eq; simpl.
    apply refines_or_elim.
    apply refines_or_elim.
    apply any_refined_bind with (a:=0).
    apply  refines_fold; simpl. firstorder auto.
    constructor 1 with (x:=0); simpl; auto.
    apply any_refined_bind with (a:=2).
    apply  refines_fold; simpl. firstorder auto.
    constructor 1 with (x:=1); simpl; auto.
  eapply refines_trans.
  apply bind_assoc_ref1.
  apply any_refines; intros n1.
  apply bind_simpl_left; simpl.
  intros H; case H; clear H; intros p1 H.
  eapply refines_trans.
  apply bind_assoc_ref1.
  apply any_refines; intros n2.
  apply bind_simpl_left; simpl.
  intros H1; case H1; clear H1; intros p2 H1.
  apply any_refined_bind with (a:=n1+n2).
  apply refines_fold; simpl; intro H0; firstorder auto.
  clear H0; subst; constructor 1 with (x:=(p1 + p2)); simpl.
  omega.
Qed.

Lemma ext_refines_even2:
  (refines (bind (any nat) (fun n => val (n+n))) even2).
Proof.
  intros; apply any_refines; intros n; simpl.
  case n.
  eapply refines_trans.
  2: unfold even2; apply sfixDS_prefix; auto with wprefines.
  simpl; unfold even2_eq; auto with wprefines.
  clear n; intros n; elim n; simpl; clear n.
  eapply refines_trans.
  2: unfold even2; apply sfixDS_prefix; auto with wprefines.
  simpl; unfold even2_eq; auto with wprefines.
  intros n HR.
  eapply refines_trans.
  2: unfold even2; apply sfixDS_prefix; auto with wprefines.
  fold even2; simpl; unfold even2_eq at 1; auto with wprefines.
  apply refines_or_intro_r.
  apply refines_trans with (s2:=(bind (val 2) 
    (fun n0 : nat => bind (val (S (n + S n))) 
     (fun m : nat => val (n0 + m))))).
  apply refines_fold; simpl; intros; omega.
  apply bind_monotony; auto with wprefines.
  eapply refines_trans.
  2: unfold even2; apply sfixDS_prefix; auto with wprefines.
  fold even2; simpl; unfold even2_eq at 1; auto with wprefines.
Qed.

(*** even3 ***)

Definition odd3_eq (ev:(DS nat)) (odd:(DS nat))
 := (or (val 1) (bind ev (fun n => (bind odd (fun m => (val (plus n m))))))).

Lemma odd3_eq_monotony:
  forall (ev1 ev2 od1 od2:(DS nat)), 
    (refines ev1 ev2) -> 
       (refines od1 od2) -> 
         (refines (odd3_eq ev1 od1) (odd3_eq ev2 od2)).
Proof.
  intros; unfold odd3_eq; auto with wprefines.
Qed.

Hint Resolve odd3_eq_monotony sfixDS_monotony: wprefines.

Definition odd3 (ev:(DS nat)) := (sfixDS (odd3_eq ev)).

Lemma odd3_monotony: forall (x y:(DS nat)), (refines x y) -> (refines (odd3 x) (odd3 y)).
Proof.
  intros; unfold odd3.
  auto with wprefines.
Qed.

Hint Resolve odd3_monotony: wprefines.

(*
Lemma odd3_refines_ext: forall (ev:(DS nat)),
  (refines (odd3 ev) (bind (require ((pre ev) /\ (angel ev (fun _ => True))))
                       (fun _ => (bind (any nat)
                         (fun n => 
                          (bind (ensure 
                           (refines (val n) 
                             (bind (any nat) (fun y => (bind ev (fun z => (val (y*z+1))))))))
                            (fun _ => (val n)))))))).
Proof.
  intros ev; apply bind_simpl_right; simpl; intuition auto; clear H1.
  unfold odd3; apply sfixDS_mostmagic; auto with wprefines.
  unfold odd3_eq; simpl.
    apply refines_or_elim.
    apply any_refined_bind with (a:=1).
    apply  refines_fold; simpl. intuition auto.
    apply any_refined_bind with (a:=0).
    apply bind_simpl_right; simpl.
    intros; eauto with wipi wprefines.
  eapply refines_trans. (* ici *)
  apply bind_assoc_ref1.
  apply any_refines; intros n1.
  apply bind_simpl_left; simpl.
  intros H; case H; clear H; intros p1 H.
  eapply refines_trans.
  apply bind_assoc_ref1.
  apply any_refines; intros n2.
  apply bind_simpl_left; simpl.
  intros H1; case H1; clear H1; intros p2 H1.
  apply any_refined_bind with (a:=n1+n2).
  apply refines_fold; simpl; intro H0; firstorder auto.
  clear H0; subst; constructor 1 with (x:=(p1 + p2)); simpl.
  omega.
Qed.
*)

Definition even3_eq (ev: (DS nat)) 
 := (or (val O) 
          (bind (odd3 ev)
             (fun n => (bind (odd3 ev)
                           (fun m => (val (plus n m))))))).

Lemma even3_eq_monotony:
  forall (x y:(DS nat)), 
    (refines x y) -> (refines (even3_eq x) (even3_eq y)).
Proof.
  intros; unfold even3_eq.
  auto 20 with wprefines.
Qed.

Unset Implicit Arguments.
