(* begin hide *)
Set Implicit Arguments.
 
Require Export ISLfriendly.
(* end hide *)
(** ** Module [ISLBlock]:  introduction of local mutable variables%\label{sec:defblock}%

In this module, I define operations to define blocks (introduction of
local mutable variables).

*)

(** I first define, a macro [Embed]. *)
Notation "'Embed' [ v ] sp" := (bind v sp (val (A:=_) (St:=_)))
  (at level 95, v, sp at level 90): ISL_scope.
Notation "'Embed' sp" := (Embed [allPl] sp)%ISL
  (at level 95, sp at level 90): ISL_scope.
(* begin hide *)

Lemma embed_change_left: forall (St St1 A B:Type) 
 (v:Place St St1) (sp1: DS St1 B) (sp2: DS St B) (st:St),
    ([< st >] (Embed [v] sp1) >==> sp2)%ISL
      ->  ([< st >] (Bind x <-[v] sp1 In (Val x)) >==> sp2)%ISL.
Proof.
 intros; ISL_trans.
 2: apply H.
 ISL_monotony.
Qed.   

Lemma embed_change_right: forall (St St1 A B:Type) 
 (v:Place St St1) (sp1: DS St1 B) (sp2: DS St B) (st:St),
    ([< st >] sp2 >==> (Embed [v] sp1))%ISL
      ->  ([< st >] sp2 >==> (Bind x <-[v] sp1 In (Val x)))%ISL.
Proof.
 intros; ISL_trans.
 apply H.
 ISL_monotony.
Qed.   

Lemma embed_intro_left: forall (St St1 A B:Type) 
 (v:Place St St1) (sp1: DS St1 A) (sp2: A -> DS St B) (sp3:DS St B) (st:St),
    ([< st >] (Bind x <- (Embed [v] sp1) In sp2 x) >==> sp3)%ISL
      ->  ([< st >] (Bind x <-[v]sp1 In sp2 x) >==> sp3)%ISL.
Proof.
 intros; ISL_trans.
 2: apply H.
 ISL_unassoc_right.
 ISL_monotony.
 intros; ISL_simplify_right.
 intros; auto with isl.
Qed.   

Lemma embed_elim_left: forall (St St1 A B:Type) 
 (v:Place St St1) (sp1: DS St1 A) (sp2: A -> DS St B) (sp3:DS St B) (st:St),
    ([< st >] Bind x <-[v]sp1 In sp2 x  >==> sp3)%ISL
      ->  ([< st >] (Bind x <- (Embed [v] sp1) In sp2 x) >==> sp3)%ISL.
Proof.
 intros; ISL_trans.
 2: apply H.
 ISL_unassoc_left.
 ISL_monotony.
 intros; ISL_simplify_left.
 intros; auto with isl.
Qed.   

Lemma embed_intro_right: forall (St St1 A B:Type) 
 (v:Place St St1) (sp1: DS St1 A) (sp2: A -> DS St B) (sp3:DS St B) (st:St),
    ([< st >] sp3 >==> (Bind x <- (Embed [v] sp1) In sp2 x))%ISL
      ->  ([< st >] sp3 >==> (Bind x <-[v]sp1 In sp2 x))%ISL.
Proof.
 intros; ISL_trans.
 apply H.
 ISL_unassoc_left.
 ISL_monotony.
 intros; ISL_simplify_left.
 intros; auto with isl.
Qed.   

Lemma embed_elim_right: forall (St St1 A B:Type) 
 (v:Place St St1) (sp1: DS St1 A) (sp2: A -> DS St B) (sp3:DS St B) (st:St),
    ([< st >] sp3 >==> (Bind x <-[v]sp1 In sp2 x))%ISL
     -> ([< st >] sp3 >==> (Bind x <- (Embed [v] sp1) In sp2 x))%ISL.
Proof.
 intros; ISL_trans.
 apply H.
 ISL_unassoc_right.
 ISL_monotony.
 intros; ISL_simplify_right.
 intros; auto with isl.
Qed.   

Ltac ISL_embed_intro_left :=  refine (embed_intro_left _ _ ).
Ltac ISL_embed_intro_right :=  refine (embed_intro_right _ _ ).

Lemma skip_r_ref2: forall (St A:Type) (sp: DS St A) (st:St),
    ([< st >] sp >==> (Embed sp))%ISL.
Proof.
 intros; apply bind_val_r_ref2; auto with isl.
Qed.   

Lemma skip_r_ref1: forall (St A:Type) (sp: DS St A) (st:St),
    ([< st >] (Embed sp) >==> sp)%ISL.
Proof.
 intros; apply bind_val_r_ref1; auto with isl.
Qed.   

(* end hide *)
                                                                                                 
Section ISLBlock.
                                                                                                 
(** Here, [v] is the ``new'' variable introduced in the block, [bkv] is the place of the ``old'' state. Hence, the state of the block is of type [sumPl v bkv]. In the future, it is maybe a good idea to simplify all this stuff, and to construct directly a pair of states.
*)
 
Variable St St2 A: Type.
Variable v: Place St A.
Variable bkv: Place St St2.
                                                                                                 
Definition block (B: Type) (u:sumPl v bkv) (x:A) (sp: (DS St B)): (DS St2 B)
  := fun (st:St2) =>
       (FSL.bind (sp (mkEnv u x st)) (fun p => 
                     FSL.val (pairT (read bkv (fstT p)) (sndT p)))).

(* begin hide *)
Lemma block_enter_left_aux:
 forall (B:Type) (u:sumPl v bkv) (x:A) (sp1: (DS St B)) (sp2:(DS St2 B)) (st:St2),
   ([< (mkEnv u x st) >] sp1
             >==> (Bind a <-[bkv] sp2 In (Bind x' <-/ (Any A) In v :<- x') -; (Val a)))%ISL
    -> ([< st >] (block u x sp1) >==> sp2)%ISL.
Proof.
  unfold refInEnv, block; simpl.
  intros B u x sp1 sp2 st H.
  apply refines_trans with
    (s2:= (FSL.bind
       ((Bind a <-[bkv] sp2 In
          (Bind x' <-/ (Any A) In v :<- x') -; (Val a))%ISL
           (mkEnv u x st))
     (fun p : prodT St B =>
      FSL.val (pairT (read bkv (fstT p)) (sndT p))))).
  apply FSL.bind_monotony; auto with wprefines.
  unfold bind; eapply refines_trans.
  apply FSL.bind_assoc_ref1.
  eapply refines_trans.
  2: apply FSL.bind_val_r_ref1.
  autorewrite with placeeq.
  apply FSL.bind_monotony; auto with wprefines.
  intros a; case a.
  intros; simpl;
  intros; apply refines_fold; simpl;
  intros; autorewrite with placeeq;
  auto.
Qed.
(* end hide *)
              
(** Here is a proof rule when [block] in the left part of a refinement goal. The refinement goal is simply lifted in a bigger state, where the new variable is simply not changed by the right part of the goal. A similar rule can be applied when [block] is in the right part of the goal.
*)                                                                  
Lemma block_enter_left:
  forall (B:Type) (u:sumPl v bkv) (x:A) (sp1: (DS St B)) (sp2:(DS St2 B)) (st:St2),
   ([< mkEnv u x st >] (Bind a <- sp1 In v :<- x -; Val a) >==> (Embed [bkv] sp2))%ISL
    -> ([< st >] (block u x sp1) >==> sp2)%ISL.
Proof.
  intros. apply block_enter_left_aux.
  ISL_embed_intro_right. 
  ISL_enter_right.
  apply H.
  ISL_unassoc_right.
  apply bind_val_r_ref2; auto with isl.
  intros a st0; ISL_simplify_all.
  constructor 1 with (x:=(read v st0)).
  intros; autorewrite with placeeq.
  auto.
Qed.
 
(* begin hide *)
Lemma block_enter_right:
  forall (B:Type) (u:sumPl v bkv) (x:A) (sp1: (DS St B)) (sp2:(DS St2 B)) (st:St2),
   ([< mkEnv u x st >] (Embed [bkv] sp2) >==> (Bind a <- sp1 In v :<- x -; Val a))%ISL
    -> ([< st >] sp2 >==> (block u x sp1))%ISL.
Proof.
  unfold refInEnv, block; simpl.
  intros B u x sp1 sp2 st H.
  apply refines_trans with
    (s2:= (FSL.bind
       ((Embed [bkv] sp2)%ISL (mkEnv u x st))
     (fun p : prodT St B =>
      FSL.val (pairT (read bkv (fstT p)) (sndT p))))).
  unfold bind; eapply refines_trans.
  apply FSL.bind_val_r_ref2.
  autorewrite with placeeq.
  unfold bind; simpl.
  eapply refines_trans.
  2: apply FSL.bind_assoc_ref2.
  autorewrite with placeeq.
  apply FSL.bind_monotony; auto with wprefines.
  intros a; case a; simpl; auto.
  intros; apply refines_fold; simpl.
  autorewrite with placeeq; auto.
  apply refines_trans with
    (s2:= (FSL.bind
      ((Bind a <- sp1 In v :<- x -;  Val a)%ISL (mkEnv u x st)) 
        (fun p : prodT St B =>
        FSL.val (pairT (read bkv (fstT p)) (sndT p))))).
  apply FSL.bind_monotony; auto with wprefines.
  unfold bind; simpl.
  eapply refines_trans.
  apply FSL.bind_assoc_ref1.
  apply FSL.bind_monotony; auto with wprefines.
  intros a; case a; simpl; auto.
  intros; apply refines_fold; simpl.
  rewrite <- (mkEnv_readLR u s).
  autorewrite with placeeq; auto.
Qed.

Lemma embed_block_right:
  forall (B:Type) (u:sumPl v bkv) (x:A) (sp: (DS St B)) (sp0:DS St B) (sti:St),
   ([< sti >] sp0 >==> (v :<- x -; Bind a <- sp In v :<- (read v sti) -; Val a))%ISL
   -> ([< sti >] sp0 >==> (Embed [bkv] (block u x sp)))%ISL.
Proof.
  unfold block, bind, set, refInEnv.
  intros; simpl.
  eapply refines_trans.
  apply H.
  eapply refines_trans.
  2:apply FSL.bind_assoc_ref2.
  simpl; autorewrite with placeeq.
  apply FSL.bind_simpl_left; simpl. 
  apply FSL.bind_monotony; auto with wprefines.
  intros a; case a; simpl.
  intros; apply refines_fold; simpl.
  rewrite <- (mkEnv_readLR u s).
 autorewrite with placeeq; auto.
Qed.

Lemma embed_block_left:
  forall (B:Type) (u:sumPl v bkv) (x:A) (sp: (DS St B)) (sp0:DS St B) (sti:St),
   ([< sti >] (v :<- x -; Bind a <- sp In v :<- (read v sti) -; Val a) >==> sp0)%ISL
      -> ([< sti >] (Embed [bkv] (block u x sp)) >==> sp0 )%ISL.
Proof.
  unfold block, bind, set, refInEnv.
  intros; simpl.
  eapply refines_trans.
  2: apply H.
  eapply refines_trans.
  apply FSL.bind_assoc_ref1.
  simpl; autorewrite with placeeq.
  apply FSL.bind_simpl_right; simpl; intros.
  apply FSL.bind_monotony; auto with wprefines.
  intros a; case a; simpl.
  intros; apply refines_fold; simpl.
  rewrite <- (mkEnv_readLR u s).
  autorewrite with placeeq; auto.
Qed.
(* end hide *)


(** The following rule expresses monotony of block w.r.t refinement. *)
Lemma block_monotony:
 forall (B:Type) (u:sumPl v bkv) (x:A) (sp1: (DS St B)) (sp2:(DS St B)) (st:St2),
   ([< mkEnv u x st >] (Bind a <- sp1 In v :<- x -; Val a) >==> (Bind a <- sp2 In v :<- x -; Val a))%ISL
    -> ([< st >] (block u x sp1) >==> (block u x sp2))%ISL.
Proof.
  intros. apply block_enter_left.
  apply embed_block_right.
  ISL_simplify_right. 
  autorewrite with placeeq.
  intros; ISL_simplify_left; auto.
Qed.

(** In these rules for handling blocks, data-refinement misses. *)

End ISLBlock.

