Set Implicit Arguments.

Require Export ISLAbsPost.
Require Export ISLLoops.
Require Import FSLFixF.

Section WHILE.

Variable St : Type.

Variable cond : DS St bool.
Variable body : DS St unit.

Section While.

Definition while: (DS St unit) :=
  (Bind n <-/ (Any nat)
   In (forN n
          (Bind b <- cond
           In (Ensure (b = true) /; body)))
       -;  Bind b <- cond 
           In Ensure (b = false))%ISL.

Lemma while_refines_invspec :
 forall (invspec : DS St unit),
 (forall (st:St), 
    ([< st >] (If cond Then (body -; invspec) Else Skip) >==> invspec)%ISL) -> 
  forall (st:St), 
    ([< st >] while >==> invspec)%ISL.
Proof.
  unfold while ; intros invspec H st.
  ISL_simplify_left; intros n.
  generalize st; elim n; clear n st; simpl.
  (* cas de base *)
  intros st; ISL_trans.
  2: eapply H.
  ISL_simplify_left.
  ISL_monotony.
  intros st' b; case b; simpl.
  ISL_simplify_all; intros; discriminate.
  ISL_Trivial.
  (* HR*)
  intros n HR st.
  ISL_unassoc_left.
  ISL_monotony_left; simpl.
  apply HR.
  ISL_unassoc_left.
  ISL_trans.
  2:apply H.
  ISL_monotony.
  intros st' b; ISL_unassoc_left.
  ISL_simplify_left.
  intros; subst; simpl.
  auto with isl.
Qed.


Lemma ifwhile_refines_while:
  forall (st:St), ([< st >] (If cond Then (body -; while) Else Skip) >==> while)%ISL.
Proof.
  unfold while; intros.
  apply if_elim. 
  apply refInEnv_trans with (sp2:=(
   (Bind b <- cond In Ensure b = true /; body) -; 
       Bind n <-/ Any nat
       In forN n (Bind b0 <- cond In Ensure b0 = true /;  body) -; 
          Bind b0 <- cond In Ensure b0 = false)%ISL).
  ISL_unassoc_right; auto with isl.
  ISL_monotony.
  intros. ISL_unassoc_right. auto with isl.
  ISL_trans. apply any_permut_bind with 
   (f:= (fun (_:unit) (n:nat) =>forN n (Bind b0 <- cond In Ensure b0 = true /;  body) -; 
    Bind b0 <- cond In Ensure b0 = false)%ISL).
  auto with nonempty.
  ISL_simplify_left; intros n.
  ISL_simplify_right.
  intros X; clear X.
  constructor 1 with (x:=(S n)).
  simpl.
  ISL_unassoc_right; auto with isl.
  ISL_simplify_right.
  intros X; clear X.
  constructor 1 with (x:=O); simpl.
  ISL_simplify_right.
  intros; ISL_monotony.
  intros; ISL_Trivial. 
Qed.

Lemma while_refines_ifwhile:
  forall (st:St), ([< st >] while >==> (If cond Then (body -; while) Else Skip))%ISL.
Proof.
  unfold while; intros.
  ISL_simplify_left; intros n.
  case n; simpl; clear n.
  ISL_simplify_left.
  ISL_monotony.
  intros st2 x; case x; simpl; ISL_simplify_all; (auto || intros; discriminate).
  intros n; ISL_unassoc_left.
  ISL_unassoc_left.
  ISL_monotony.
  intros st2 x; ISL_unassoc_left.
  ISL_simplify_left; intros; subst; simpl.
  ISL_monotony.
  intros; ISL_simplify_right.
  intros; constructor 1 with (x:=n).
  auto with isl.
Qed.

End While.

(** notion "classique" d'invariant **) 

Section WhileAbs.

Variable inv: St -> Prop.

Definition inv_preservation: Prop :=
   forall (st:St), (inv st) 
      -> ([< st >] If cond Then body Else Skip >==> 
    (upd_abspost allPl (fun new_st result => (inv new_st))))%ISL.

Lemma inv_preservation_pre:
  inv_preservation -> forall (st:St), (inv st) 
   -> (pre ((If cond Then body Else Skip)%ISL st)).
Proof.
  unfold inv_preservation, refInEnv, refines; intros.
  abstract firstorder auto.
Qed. 

Definition while_abs: (DS St unit) :=
 (( Bind st <- !allPl In Req (inv st)) -; 
   (Bind tmp_st <-/ (Any St) In allPl :<- tmp_st) -;
      Bind b <- cond
      In Bind new_st <- ! allPl
           In Ensure ((inv new_st) /\ b=false))%ISL.

Lemma while_refines_while_abs: inv_preservation -> forall (st:St), ([< st >] while >==> while_abs)%ISL.
Proof.
  intros H1 st; apply while_refines_invspec; clear st.
  intros st;  unfold while_abs at 2.
  ISL_simplify_right.
  intuition auto.
  apply if_elim.
  (* cas boucle *)
  apply refInEnv_trans with
   (sp2:=((If cond Then body Else Skip) -; while_abs)%ISL).
  ISL_unassoc_right.
  ISL_monotony.
  intros st0 b; ISL_simplify_left.
  intros H; subst; simpl; auto with isl.
  ISL_enter_left.
  apply H1; auto.
  ISL_simplify_left. 
  unfold while_abs.
  intros; ISL_simplify_left.
  ISL_simplify_left.
  intuition auto.
  ISL_simplify_left.
  ISL_simplify_left; intros st1.
  ISL_simplify_right.
  intros; ISL_simplify_right.
  intros; constructor 1 with (x:=st1).
  ISL_assoc_left.
  ISL_assoc_left.
  ISL_simplify_left.
  auto with isl.
  (* cas arret *)
  ISL_simplify_right; simpl.
  intros H; clear H.
  constructor 1 with (x:=st).
  ISL_simplify_left.
  apply refInEnv_trans with
   (sp2:=(Bind b <- cond In Ensure (b = false))%ISL).
  ISL_monotony.
  intros; ISL_Trivial.
  eapply intro_updabspost with (v:=(allPl (A:=St))).
  ISL_trans.
  2:apply H1; auto.
  ISL_monotony.
  intros; ISL_simplify_all; intros; subst; simpl; intuition auto.
  simpl.
  ISL_unassoc_left; auto with isl.
  ISL_monotony.
  intros; ISL_Trivial.
Qed.

End WhileAbs.


Section While2.

Variable inv: St -> Prop.

Hypothesis ipH: (inv_preservation inv).

Variable R: St -> St -> Prop.
Hypothesis Rwf : (well_founded R).
  
Definition variant_decreasing: Prop :=
  forall (st:St), (inv st) -> ([< st >] (Bind b <- cond In Ensure (b=true) /; body) >==> 
    (upd_abspost allPl (fun new_st result => (R new_st st))))%ISL.

Hypothesis vdH: variant_decreasing.

Definition while_loop (wh:(DS St unit)): (DS St unit) 
  := ((Bind st <- !allPl In (Req (inv st))) -; If cond Then body -; wh Else Skip)%ISL.

Lemma while2_decreasing: (call_decreasing while_loop R).
Proof.
  unfold call_decreasing, while_loop.
  intros wh st; apply bind_simpl_both; simpl; auto with isl place.
  intuition auto.
  apply if_elim.
  apply refInEnv_trans with (sp2:=((Bind b <- cond In Ensure b = true /;  body)-; wh)%ISL).
  ISL_unassoc_right; ISL_monotony.
  intros; ISL_unassoc_right; auto with isl.
  eapply intro_bind_updabspost with (v:=(allPl (A:=St))).
  apply vdH; auto.
  ISL_unassoc_left.
  ISL_monotony.
  intros; ISL_unassoc_left.
  ISL_simplify_left.
  intros; subst; simpl.
  ISL_monotony.
  intros; ISL_simplify_left.
  intros; ISL_simplify_left.
  unfold ensure, refInEnv.
  apply FSL.bind_simpl_right; simpl.
  intuition auto with wprefines.
  ISL_monotony.
  intros; ISL_simplify_left.
  intros; subst; simpl.
  ISL_Trivial.
Qed.

Lemma while2_monotony:
 forall (sp1 sp2: (DS St unit)),
  (forall x : St, refines (sp1 x) (sp2 x))
  -> forall (st:St),
    refines (while_loop sp1 st) (while_loop sp2 st).
Proof.
  intros sp1 sp2 H1 st.
  unfold while_loop.
  ISL_monotony.
  intros; apply if_monotony; auto with isl.
  intros; ISL_monotony.
  intros st3 dum; apply (H1 st3).
Qed.

Definition while2 := (fixF while_loop Rwf).

Hint Resolve while2_decreasing while2_monotony: isl.

Lemma while2_refines_while:
  forall (st:St), ([< st >] while2 >==> ( (Bind st <- !allPl In (Req (inv st))) -; while))%ISL.
Proof. 
  unfold while2.  
  intros st; unfold refInEnv; apply fixF_mostmagic;
  auto with isl.
  clear st; intros st; 
  change ([< st >] (while_loop ((Bind st <- !allPl In (Req (inv st))) -; while)) 
    >==> (((Bind st <- !allPl In (Req (inv st)))) -; while))%ISL.
  unfold while_loop.
  ISL_simplify_both.
  intuition auto.
  ISL_trans.
  2: apply ifwhile_refines_while.
  apply if_elim.
  apply refInEnv_trans with
   (sp2:=
    ((Bind b <- cond
      In Ensure b = true /; 
       body)-; 
      (Bind st0 <- !allPl In Req inv st0) -; while)%ISL).
   repeat (intros; ISL_unassoc_right; ISL_monotony).
   eapply intro_bind_updabspost with (v:=(allPl (A:=St))).
   ISL_trans.
   2: apply ipH.
   ISL_monotony.
   intros; ISL_simplify_left; intros; subst; simpl; auto with isl.
   auto.
   simpl; ISL_unassoc_left.
   ISL_monotony.
   intros.
   ISL_unassoc_left.
   ISL_simplify_left; intros; subst; simpl.
   ISL_monotony.
   intros.
   ISL_simplify_left.
   intros; ISL_simplify_left.
   ISL_simplify_left.
   intuition auto.
   ISL_simplify_left.
   auto with isl.
   ISL_monotony.
   intros; ISL_simplify_left; intros; subst.
   simpl; auto with isl.
Qed.

Lemma while2_decaps:
  forall (st:St), (refInEnv st (while_loop while2) while2).
Proof.
  unfold while2.  
  intros st; unfold refInEnv; apply fixF_prefix; auto with isl.
Qed.

Section While2_rect.

Variable P: (DS St unit) -> St -> Type.

Hypothesis P_compat: forall (st:St) (sp1 sp2: (DS St unit)),
  (P sp1 st)
    -> (refInEnv st sp1 sp2) 
      -> (refInEnv st sp2 sp1)
     ->  (P sp2 st).

Lemma P_while_require:
   (forall (st:St) (Q: St -> Prop) (sp:(DS St unit)),
      ((Q st) -> (P sp st)) -> (P ((Req (Q st)) /; sp)%ISL) st)
   -> (forall (wh: (DS St unit)), (forall (st:St), (P wh st))
             -> forall (st:St), P (while_loop wh) st)
   -> forall (st:St), P while2 st.
Proof.
  unfold while2.
  intros P_require P_preserv st.
  apply inv_fixF_require; eauto with isl.
  clear st; intros st Q sp H; eapply P_compat.
  apply P_require with (Q:=Q) (sp:=sp); auto.
  unfold refInEnv, require, bind, lift.
  intros; apply FSL.require_refined.
  intros; apply FSL.bind_simpl_left; simpl.
  intuition auto with wprefines.
  ISL_simplify_right.
  intuition auto; unfold refInEnv; auto with wprefines.
Qed.

Lemma P_while_ensure:
   (forall (st:St) (Q: St -> Prop) (sp:(DS St unit)),
      ((Q st) -> (P sp st)) -> (P ((Ensure (Q st)) /; sp)%ISL) st)
   -> (forall (wh: (DS St unit)), (forall (st:St), (P wh st))
             -> forall (st:St), P (while_loop wh) st)
   -> forall (st:St), P while2 st.
Proof.
  unfold while2.
  intros P_require P_preserv st.
  apply inv_fixF_ensure; eauto with isl.
  clear st; intros st Q sp H; eapply P_compat.
  apply P_require with (Q:=Q) (sp:=sp); auto.
  ISL_simplify_left; simpl.
  intros; unfold refInEnv; auto with wprefines.
  unfold refInEnv, ensure, bind, lift.
  intros; apply FSL.ensure_refines.
  intros; apply FSL.bind_simpl_right; simpl.
  intuition auto with wprefines.
Qed.

End While2_rect.

Lemma while_refines_while2:
  forall (st:St), ([< st >] ((Bind st <- !allPl In (Req (inv st))) -; while) >==> while2)%ISL.
Proof.
  intros st; pattern while2, st.
  apply P_while_require; clear st.
  intros; ISL_trans; eauto.
  intros; ISL_simplify_right; intuition auto.
  intros wh H st.
  unfold while_loop.
  ISL_simplify_right.
  intuition auto.
  clear H2.
  ISL_simplify_left; intuition auto.
  ISL_simplify_left.
  ISL_trans.
  apply while_refines_ifwhile.
  repeat (intros; ISL_monotony).
  intros.
  ISL_trans. 2:eauto.
  ISL_simplify_right; auto with isl.
Qed.

Lemma while2_refines_while_abs:
  forall (st:St), ([< st >] while2 >==> (while_abs inv))%ISL.
Proof.
  intros; apply refInEnv_pre_intro; simpl.
  intros H1; intuition auto.
  clear H0.
  eapply refInEnv_trans.
  apply while2_refines_while.
  ISL_simplify_left; intuition auto.
  ISL_simplify_left.
  apply while_refines_while_abs; auto.
Qed.

End While2.

Section WhileWF.

Variable inv: St -> Prop.

Variable R: St -> St -> Prop.

Definition whileWF:=
 ((Bind st <- !allPl In
   Req ((inv_preservation inv) /\ (variant_decreasing inv R) /\ (well_founded R) /\ (inv st))) -;
   while)%ISL.

Definition whileWF2:=
 ((Req ((inv_preservation inv) /\ (variant_decreasing inv R) /\ (well_founded R))) /;
  Bind Rwf <-/ (Any (well_founded R)) In
     (while2 inv Rwf))%ISL.

Lemma whileWF_ref1: forall (st:St), (refInEnv st whileWF whileWF2).
Proof.
  intros; unfold whileWF, whileWF2; ISL_simplify_right.
  intuition auto.
  ISL_simplify_right.
  intros; constructor 1 with (x:=H3).
  ISL_trans.
  2: apply while_refines_while2; auto.
  ISL_monotony.
  ISL_Trivial.
Qed.

Lemma whileWF_ref2: forall (st:St), (refInEnv st whileWF2 whileWF).
Proof.
  intros; unfold whileWF, whileWF2. ISL_simplify_right.
  intuition auto.
  ISL_simplify_left.
  intuition auto.
  ISL_simplify_left.
  intros; ISL_trans.
  apply while2_refines_while; auto.
  ISL_simplify_left.
  intuition auto with isl.
Qed.

Section WhileWF_ind.

Variable P: (DS St unit) -> St -> Prop.

Hypothesis P_compat: forall (st:St) (sp1 sp2: (DS St unit)),
  (P sp1 st)
    -> (refInEnv st sp1 sp2) 
      -> (refInEnv st sp2 sp1)
     ->  (P sp2 st).

Lemma P_whileWF_require:
   (forall (st:St) (Q: St -> Prop) (sp:(DS St unit)),
      ((Q st) -> (P sp st)) -> (P ((Req (Q st))  /; sp)%ISL) st)
   -> (forall (while: (DS St unit)), (forall (st:St), (P while st))
             -> forall (st:St), P (while_loop inv while) st)
   -> forall (st:St), P whileWF st.
Proof.
  intros P_require P_preserv st.
  eapply P_compat.
  2: apply whileWF_ref2.
  unfold whileWF2.
  apply P_require with (Q:=(fun (_:St) => inv_preservation inv /\ variant_decreasing inv R /\ well_founded R)).
  intros H; decompose [Logic.and] H.
  eapply P_compat.
  apply P_while_require with (inv:=inv) (Rwf:=H3); auto.
  ISL_simplify_right; simpl; intros.
  constructor 1 with (x:=H3); auto with isl.
  ISL_simplify_left; simpl; intros.
  unfold while2, refInEnv.
  apply fixF1_refines_fixF2; auto with wprefines.
  intros; apply while2_monotony; auto.
  intros; apply while2_monotony; auto.
  intros; apply while2_decreasing; auto.
  apply whileWF_ref1.
Qed.

End WhileWF_ind.

Lemma whileWF_spec: 
   forall (st:St), ([< st >] whileWF >==> 
     ((Req (inv_preservation inv) /\ (variant_decreasing inv R) /\ (well_founded R)) /;
      (while_abs inv)))%ISL.
Proof.
  intros st; ISL_simplify_right; intuition auto.
  unfold whileWF.
  apply refInEnv_pre_intro; simpl.
  intros.
  ISL_simplify_left; intuition auto.
  ISL_simplify_left. 
  apply while_refines_while_abs; auto.
Qed.

Lemma whileWF_unfold_left:
 forall (st:St), ([< st >] (If cond Then body -; whileWF Else Skip) >==> whileWF)%ISL.
Proof.
  intros; ISL_trans. 
  2: apply whileWF_ref2.
  unfold whileWF2; ISL_simplify_right.
  intuition auto.
  ISL_simplify_right.
  intros D; clear D.
  constructor 1 with (x:=H3).
  ISL_trans. 2: apply while2_decaps.
  unfold while_loop.
  ISL_simplify_right.
  intros; ISL_monotony.
  intros; ISL_monotony.
  intros. ISL_trans.
  apply whileWF_ref1.
  unfold whileWF2; ISL_simplify_left; intuition auto with nonempty.
  ISL_simplify_left.
  intros;  unfold while2, refInEnv.
  apply fixF1_refines_fixF2; auto with wprefines.
  intros; apply while2_monotony; auto.
  intros; apply while2_monotony; auto.
  intros; apply while2_decreasing; auto.
  auto.
Qed.

Lemma invspec_refines_whileWF :
 forall (invspec : DS St unit),
 (forall (st:St), 
    ([< st >] invspec >==> (while_loop inv invspec))%ISL) -> 
  forall (st:St), ([< st >] invspec >==> whileWF)%ISL.
Proof.
  intros invspec H st.
  pattern whileWF, st. apply P_whileWF_require; clear st; auto with isl.
  intros; ISL_trans; eauto.
  intros; ISL_simplify_right; intuition auto.
  intros; ISL_trans.
  apply H.
  unfold while_loop.
  ISL_monotony.
Qed.

End WhileWF.

End WHILE.

Unset Implicit Arguments.
