Set Implicit Arguments.

Require Export ISLfriendly.
Require Import List.

Fixpoint lookup (St A : Set) (test : A -> DS St bool) (l : list A) {struct l} : 
 DS St A :=
  match l with
  | nil => (Abort)%ISL
  | a :: m => (If (test a) Then (Val a) Else (lookup test m))%ISL
  end.  

(* si test est "fonctionnel pur" *)
Lemma lookup_when_pur_reject :
 forall (St A : Set) (test : A -> DS St bool) (p : A -> St -> bool),
 (forall (a : A) (s : St), 
  ([< s >] test a >==>  ((Ext  s => (p a s)) @-  ! allPl))%ISL)
 -> 
 forall (l : list A) (s : St),
 ([< s >] lookup test l >==> 
 (Bind s' <- ! allPl
  In Req (forall a : A, List.In a l -> p a s' = false) /;  Abort))%ISL.
Proof.
  simpl in |- *; intros St A test p H l; elim l; simpl in |- *. 
  (* On fait la preuve par rcurrence sur "l" *)
  (* cas de base *) ISL_Trivial.
  (* Recu: On injecte l'hypothse sur "test" par monotonie et on simplie la formule *)
  intros a m HR s.
  ISL_enter_left. eauto.
  ISL_unassoc_left.
  ISL_simplify_both.
  intros; ISL_simplify_left.
  ISL_simplify_right.
  (* On utilise la prcondition pour simplifier la formule *)
  intros; assert (H2 : p a s = false). intuition auto. rewrite H2.
  (* on injecte l'hypothse de HR *)
  ISL_trans; eauto.
  (* il suffit de vrifier que la prcondition est vrifie... *)
  ISL_Trivial.
Qed.

Lemma lookup_when_pur_accept :
 forall (St A : Set) (test : A -> DS St bool) (p : A -> St -> bool),
 (forall (a : A) (s : St), ([< s >] test a >==>  ((Ext s => (p a s)) @- ! allPl))%ISL) -> 
 forall (l : list A) (s : St),
 ([< s >] lookup test l >==> 
  (Bind s <-  ! allPl
   In Req (exists a : A, List.In a l /\ p a s = true) /; 
       Bind a <-/ Any A In Ensure List.In a l /\ p a s = true /;  Val a))%ISL.
Proof.
  intros St A test p H l; elim l; simpl in |- *. (* On fait la preuve par rcurrence sur "l" *)
  (* cas de base: la prcondition n'est pas vrifie. *) 
  intros; ISL_simplify_right; firstorder auto.
  (* Recu: On injecte l'hypothse sur "test" par monotonie et on simplie la formule *)
  intros a m HR s.
  ISL_enter_left. eauto.
  ISL_unassoc_left.
  ISL_simplify_both.
  intros; ISL_simplify_left.
  (* On introduit la prcondition, on simplifie cette prcondition *)
  intros; ISL_simplify_right.
  intuition auto.
  cut (p a s = true \/ (exists a0 : A, List.In a0 m /\ p a0 s = true)).
  2: decompose [Logic.and ex] H2; intuition eauto; rewrite H1; intuition eauto.
  (* On raisonne par cas sur la valeur renvoye par "(p a s)" *)
  assert (H1 : exists b : bool, b = p a s). eauto.
  case H1; intros b; case b; clear H2 H1; intro H2; rewrite <- H2; simpl in |- *;
   intros. 
   (* cas "(p a s)=true" *) 
   ISL_simplify_right. intros; constructor 1 with (x := a); ISL_Trivial.
   (* cas "(p a s)=false" *)
   case H1; intros H4. discriminate H4. (* On simplifie la prcondition en hypothse *)
   ISL_trans. eauto.  (* On injecte l'hypothse de rcurrence *)
   clear H2 H3; ISL_Trivial.
Qed. 

Fixpoint forN (St : Set) (n : nat) (i : DS St unit) {struct n} :
 DS St unit :=
  match n with
  | O => (Val tt)%ISL
  | S m => (i -;  forN m i)%ISL
  end.

Lemma lookup_non_pur_reject :
 forall (St A : Set) (test : A -> DS St bool) (p : A -> bool)
   (modify : DS St unit),
 (forall (a : A) (s : St), ([< s >] test a >==>  (modify -;  Val p a))%ISL) -> 
 forall (l : list A) (s : St),
 ([< s >] lookup test l >==> 
  (Req (forall a : A, List.In a l -> p a = false) /; 
  forN (length l) modify -; Abort))%ISL.
Proof.
  simpl in |- *; intros St A test p modify H l; elim l; simpl in |- *.
  (* On fait la preuve par rcurrence sur "l" *)
  (* cas de base *) ISL_Trivial.
  (* Recu: On injecte l'hypothse sur "test" par monotonie et on simplie la formule *)
  intros a m HR s.
  ISL_enter_left. eauto.
  ISL_simplify_right; intros. assert (H2 : p a = false). intuition auto. rewrite H2.
  ISL_unassoc_left. 
  ISL_unassoc_right. 
  ISL_monotony.
  intros; ISL_simplify_left. ISL_trans; eauto.
  ISL_simplify_left.
  constructor 1. intuition auto.
  auto with isl.
Qed.

Lemma lookup_non_pur_accept :
 forall (St A : Set) (test : A -> DS St bool) (p : A -> bool)
   (modify : DS St unit),
 (forall (a : A) (s : St), ([< s >] test a >==>  (modify -;  Val p a))%ISL) -> 
 forall (l : list A) (s : St),
 ([< s >] lookup test l >==> 
  (Req (exists a : A, List.In a l /\ p a = true) /; 
   Bind n <-/  Any nat
   In forN (S n) modify -; 
        Bind a <-/  Any A
        In Ensure List.In a l /\ p a = true /; 
           Ensure nth n l a = a /;  Val a))%ISL.
Proof.
  intros St A test p modify H l; elim l; simpl in |- *. 
  (* On fait la preuve par rcurrence sur "l" *)
  (* cas de base: la prcondition n'est pas vrifie. *) 
  intros; ISL_simplify_right; firstorder auto.
  (* Recu: On injecte l'hypothse sur "test" par monotonie et on simplie la formule *)
  intros a m HR s.
  ISL_enter_left. eauto.
  ISL_unassoc_left.
  (* On introduit la prcondition, on simplifie cette prcondition *)
  ISL_simplify_right; intros.
  cut (p a = true \/ (exists a0 : A, List.In a0 m /\ p a0 = true)).
  2: decompose [Logic.and ex] H0; intuition eauto; rewrite H0; intuition eauto.
  (* On raisonne par cas sur la valeur renvoye par "(p a s)" *)
  assert (H2 : exists b : bool, b = p a). eauto.
  case H2; intros b; case b; clear H2; intro H2; rewrite <- H2; simpl in |- *;
   intros.
  (* cas "(p a s)=true" *) 
  ISL_monotony_left. intros; ISL_simplify_left. eauto with isl.
  ISL_simplify_right. constructor 1 with (x := 0); simpl in |- *.
  ISL_unassoc_right. ISL_monotony.
  intros; ISL_simplify_right; intros. ISL_simplify_right. constructor 1 with (x := a).
  ISL_Trivial.
  (* cas "(p a s)=false" *)
  ISL_monotony_left. intros; ISL_simplify_left. eauto with isl.
  ISL_monotony_left.      
   intros; ISL_simplify_left.
   intuition auto. discriminate H0.
  auto with isl.
  auto with isl.
  ISL_trans.
  eapply any_permut_bind with
                       (f := 
                         fun (_ : unit) n =>
                         ((modify -;  forN n modify) -; 
                         Bind a0 <-/  Any A
                         In Ensure List.In a0 m /\ p a0 = true /; 
                            Ensure nth n m a0 = a0 /;  Val a0)%ISL).
  auto with nonempty.
  ISL_simplify_left; intros n; ISL_simplify_right. intros; constructor 1 with (x := S n).
  ISL_unassoc_right; simpl in |- *. 
  ISL_monotony.
  intros; ISL_monotony.
  intros; ISL_simplify_both; ISL_Trivial.
Qed. 

(** exemple d'instanciations **)
 
Definition sumbool2bool (p q : Prop) (sb : {p} + {q}) : bool :=
  match sb with
  | left _ => true
  | right _ => false
  end.

Variable A : Set.
Hypothesis eq_A_dec : forall x y : A, {x = y} + {x <> y}.

Definition eq_A_bool (x y : A) : bool := sumbool2bool (eq_A_dec x y).

Definition test_pur (b : A) : DS A bool :=
  ((Ext a => (eq_A_bool a b)) @-  ! allPl)%ISL.

Lemma lookup_eqA :
 forall (a : A) (l : list A),
 ([< a >] lookup test_pur l >==> 
 (if sumbool2bool (In_dec eq_A_dec a l) then Val a else Abort))%ISL.
Proof.
   simpl in |- *; intros a l.
   (* On raisonne par cas sur "In" *)
   case (In_dec eq_A_dec a l); simpl.
   (* cas b=true *)
   intro H0; ISL_trans.
   eapply lookup_when_pur_accept with (p := fun x a => eq_A_bool a x).
   unfold test_pur in |- *; simpl in |- *; intros; auto with isl.
   ISL_simplify_left; ISL_simplify_left; constructor 1.
   constructor 1 with (x := a); simpl in |- *; intuition auto.
   unfold eq_A_bool in |- *; case (eq_A_dec a a); simpl in |- *;
    intuition auto.
   ISL_simplify_all; intros H1 a0.
   unfold eq_A_bool in |- *; case (eq_A_dec a a0); simpl in |- *.
   intros; subst; auto.
   intros H4 H5. absurd (false = true). 
   intros H6; discriminate H6.
   intuition auto.
   (* cas b=false *)
   intro H0; ISL_trans.
   eapply lookup_when_pur_reject with (p := fun x a => eq_A_bool a x).
   unfold test_pur in |- *; simpl in |- *; intros; auto with isl.
   ISL_simplify_all; intuition auto.
Qed.

Definition test_count (x y : A) : DS nat bool :=
  ((Bind ns <- ((Ext n => (S n)) @-  ! allPl) In allPl :<- ns) -;  
  Val eq_A_bool x y)%ISL.

Lemma lookup_test_count :
 forall (a : A) (l : list A) (n : nat),
 ([< n >] lookup (test_count a) l >==> 
 (Req List.In a l /; 
  Bind n' <-/  Any nat
  In allPl :<- S n' + n  -;  
     Ensure nth n' l a = a /;  Val a))%ISL. 
Proof.
  intros a l n; ISL_trans.
  eapply
   lookup_non_pur_accept with
                         (p := fun x => eq_A_bool a x)
                         (modify := ((Bind ns <- ((Ext n => (S n)) @-  (get allPl)) In allPl :<- ns)%ISL)).
  unfold test_pur in |- *; simpl in |- *; intros; auto with isl.
  ISL_simplify_right.
  intros; ISL_simplify_left; constructor 1. constructor 1 with (x := a); intuition auto.
  unfold eq_A_bool in |- *; case (eq_A_dec a a); simpl in |- *;  firstorder auto. 
  ISL_simplify_both.
  intros dum m.
  ISL_monotony.
  ISL_simplify_left.
  ISL_simplify_left.
  generalize n; elim m; clear m dum n H l a; simpl in |- *.
  intros; ISL_Trivial.
  intros m HR n. ISL_simplify_left. ISL_simplify_left.
  simpl; ISL_trans. eapply HR. rewrite <- plus_n_Sm. auto with isl.
  intros; ISL_simplify_left; intros a0. ISL_simplify_left.  
  unfold eq_A_bool in |- *; case (eq_A_dec a a0); simpl in |- *.
  intros H1 H2; rewrite <- H1. auto with isl.
  intros H1 H2; case H2; intros H3 H4; discriminate H4.
Qed.
