Set Implicit Arguments.

Require Export SSLAbsPost.
Require Import MyTheoryList.


Fixpoint forN (St:Type) (n : nat) (i : K St unit) {struct n} : K St unit :=
  match n with
  | O => Skip%SSL
  | S m => (i -;  forN m i)%SSL
  end.

Lemma forN_monotony:
  forall (St:Type) (n:nat) (i j:K St unit),
    (forall (st:St), [< st >] i >==> j)%SSL
    -> (forall (st:St),
         [< st >] (forN n i) >==> (forN n j))%SSL.
Proof.
  induction n.
  simpl; auto with isl.
  intros; simpl.
  SSL_monotony.
Qed.

Hint Resolve forN_monotony: isl.

Lemma forN_commut_ref1:
  forall (St:Type) (n:nat) (i:K St unit) (st:St),
         ([< st >] (i-; forN n i) >==> (forN n i -; i))%SSL.
Proof.
  induction n.
  simpl. intros; SSL_simplify_right.
  intros; apply bind_val_r_ref1; auto with isl.
  intros a; case a; auto with isl.
  intros; simpl.
  SSL_unassoc_right.
  SSL_monotony.
Qed.

Hint Resolve forN_commut_ref1: isl.

Definition repeat (St:Type) (i : K St unit) : K St unit :=
  (Bind m <- (any nat) In (forN m i))%SSL.

Lemma repeat_monotony:
  forall (St:Type) (i j:K St unit),
    (forall (st:St), [< st >] i >==> j)%SSL
    -> (forall (st:St),
         [< st >] (repeat i) >==> (repeat j))%SSL.
Proof.
  intros; unfold repeat.
  SSL_monotony.
Qed.

Hint Resolve repeat_monotony: isl.

Lemma repeat_stutter_left:
  forall (St:Type) (i:K St unit) (st:St),
         ([< st >] (repeat i -; repeat i)  >==> (repeat i))%SSL.
Proof.
  intros; unfold repeat.
  SSL_unassoc_left.
  SSL_simplify_left.
  intros a; SSL_simplify_left.
  induction a; simpl.
  SSL_simplify_left; auto with isl.
  SSL_trans.
  2:apply IHa.
  SSL_enter_left.
  eapply forN_commut_ref1.
  SSL_unassoc_left.
  SSL_monotony.
  intros.
  apply refInEnv_trans with (sp2:=(Bind m <- any nat In i-; forN m i)%SSL).
  apply any_permut_bind with (f:=(fun (x:unit) m => forN m i)%SSL).
  auto with nonempty.
  SSL_simplify_left.
  intros n; SSL_simplify_right.
  intros; constructor 1 with (x:=(S n)).
  auto with isl.
Qed.

Hint Resolve repeat_stutter_left: isl.

Lemma repeat_stutter_in_left:
  forall (St:Type) (i:K St unit) (st: St),
         ([< st >] (repeat (repeat i))  >==> (repeat i))%SSL.
Proof.
  intros; unfold repeat at 1.
  SSL_simplify_left.
  intros n;
  SSL_simplify_left; generalize st; clear st; elim n; simpl.
  unfold repeat; intros; SSL_simplify_right.
  intros; constructor 1 with (x:=0).
  simpl; auto with isl.
  intros m HR st; SSL_trans.
  2: apply repeat_stutter_left.
  SSL_monotony.
Qed.


Lemma repeat_zero_left:
  forall (St:Type) (i:K St unit) (st: St),
         ([< st >] Skip  >==> (repeat i))%SSL.
Proof.
  intros St i st; unfold repeat.
  SSL_simplify_right; intros.
  constructor 1 with (x:=0).
  simpl; auto with isl.
Qed.

Lemma repeat_one_left:
  forall (St:Type) (i:K St unit) (st: St),
         ([< st >] i  >==> (repeat i))%SSL.
Proof.
  intros St i st; unfold repeat.
  SSL_simplify_right; intros.
  constructor 1 with (x:=1).
  simpl.
  apply bind_val_r_ref2; auto with isl.
  intros a; case a; auto with isl.
Qed.

(*
Definition repeat_invoke (St A:Type) (g: A -> Prop) (i: A -> K St unit)
 := (repeat (Bind x <-/ (Any A) In
           Ensure (g x) /;
           i x))%SSL.

Lemma repeat_invoke_guard_monotony:
  forall (St A:Type) (g1 g2: A -> Prop) (i:A -> K St unit),
    (forall (x:A), (g1 x) -> (g2 x))%SSL
    -> (forall (st:St),
         [< st >] (repeat_invoke g1 i) >==> (repeat_invoke g2 i))%SSL.
Proof.
  intros; unfold repeat_invoke.
  apply repeat_monotony.
  intros.
  SSL_monotony.
  intros; SSL_monotony.
  SSL_Trivial.
Qed.

Hint Resolve repeat_invoke_guard_monotony: isl.

Lemma repeat_invoke_monotony:
  forall (St A:Type) (g: A -> Prop) (i j:A -> K St unit),
    (forall (st:St) (x:A), (g x) -> [< st >] (i x) >==> (j x))%SSL
    -> (forall (st:St),
         [< st >] (repeat_invoke g i) >==> (repeat_invoke g j))%SSL.
Proof.
  intros; unfold repeat_invoke.
  apply repeat_monotony.
  intros.
  SSL_monotony.
  intros; SSL_simplify_left.
  intros; SSL_simplify_right.
  intros; intuition auto.
Qed.

Hint Resolve repeat_invoke_monotony: isl.

Lemma repeat_invoke_stutter_left:
  forall (St A:Type) (g: A -> Prop) (i:A -> K St unit)  (st:St),
         ([< st >] (repeat_invoke g i -; repeat_invoke g i)  >==> (repeat_invoke g i))%SSL.
Proof.
  intros; unfold repeat_invoke.
  auto with isl.
Qed.

Lemma repeat_invoke_stutter_in_left:
  forall (St A:Type) (g: A -> Prop) (i:A -> K St unit)  (st:St),
         ([< st >] (repeat (repeat_invoke g i))  >==> (repeat_invoke g i))%SSL.
Proof.
  intros; unfold repeat_invoke.
  apply repeat_stutter_in_left.
Qed.

Lemma repeat_invoke_zero_left:
  forall (St A:Type) (g: A -> Prop) (i: A-> K St unit) (st: St),
         ([< st >] Skip  >==> (repeat_invoke g i))%SSL.
Proof.
  intros; unfold repeat_invoke.
  apply repeat_zero_left.
Qed.

Hint Resolve repeat_invoke_zero_left: isl.

Lemma repeat_invoke_one_left:
  forall (St A:Type) (g: A -> Prop) (i: A -> K St unit)
    (x:A) (st: St), (g x) -> ([< st >] (i x)  >==> (repeat_invoke g i))%SSL.
Proof.
  intros; unfold repeat_invoke.
  SSL_trans.
  2:apply repeat_one_left.
  SSL_simplify_right.
  intros; constructor 1 with (x:=x).
  SSL_simplify_right.
  intuition auto.
  auto with isl.
Qed.
*)

Open Scope SSL_scope.

Fixpoint listRevIter (St:Type) (A:Set) (l:list A) (p: A -> K St unit) { struct l} : (K St unit) 
  := match l with
     | nil => skip
     | x::m => listRevIter m p -; p x
     end.

Lemma listRevIter_monotony:
  forall (St:Type) (A:Set) (l:list A) (p1 p2: A-> K St unit),
    (forall (x:A) (st:St), [< st >] p1 x >==> p2 x)%SSL
    -> forall (st:St),
         ([< st >] (listRevIter l p1) >==> (listRevIter l p2))%SSL.
Proof.
  induction l.
  simpl; auto with isl.
  intros; simpl.
  SSL_monotony.
Qed.

Hint Resolve listRevIter_monotony: isl.

Lemma listRevIter_app_ref1:
  forall (St:Type) (A:Set) (l1 l2:list A) (p: A -> K St unit) (st:St),
         ([< st >] (listRevIter (l1++l2) p) >==> (listRevIter l2 p) -; (listRevIter l1 p))%SSL.
Proof.
  induction l1; simpl.
  intros; apply bind_val_r_ref2; auto with isl.
  intros a; case a; auto with isl. 
  intros; SSL_enter_left.
  apply IHl1.
  SSL_unassoc_left.
  auto with isl.
Qed.

Lemma listRevIter_app_ref2:
  forall (St:Type) (A:Set) (l1 l2:list A) (p: A -> K St unit) (st:St),
         ([< st >] (listRevIter l2 p) -; (listRevIter l1 p) >==> (listRevIter (l1++l2) p) )%SSL.
Proof.
  induction l1; simpl.
  intros; apply bind_val_r_ref1; auto with isl.
  intros a; case a; auto with isl. 
  intros. SSL_assoc_left. SSL_enter_left.
  apply IHl1.
  auto with isl.
Qed.

Lemma listRevIter_composeIn_app_ref1:
  forall (St:Type) (A B:Set) (f: A -> list B) (p: B -> K St unit) (l:list A) (st:St),
   ([< st >] (listRevIter l (fun x => listRevIter (f x) p)) >==> 
     (listRevIter (fold_right (fun x m => (f x) ++ m) nil l) p))%SSL.
Proof.
  induction l; simpl.
  auto with isl.
  intros st; SSL_enter_left.
  apply IHl.
  SSL_trans.
  apply listRevIter_app_ref2.
  auto with isl.
Qed.

Lemma listRevIter_composeIn_cons:
  forall (St:Type) (A B:Set) (f: A -> B) (p: B -> K St unit)  (l:list A) (st:St),
         ([< st >] (listRevIter l (fun x:A => p (f x))) >==> 
                     (listRevIter (fold_right (fun x m => (f x)::m) nil l) p))%SSL.
Proof.
  intros; SSL_trans.
  2: apply listRevIter_composeIn_app_ref1 with (f:=(fun x => (f x)::nil)).
  simpl.
  apply listRevIter_monotony.
  intros. SSL_simplify_right.
  auto with isl.
Qed.

Lemma listRevIter_composeIn_cons_app:
  forall (St:Type) (A:Set) (p: A -> K St unit) (f: A -> A -> A) (l1 l2:list A) (st:St),
         ([< st >] (listRevIter l1 (fun x => listRevIter l2 (fun y => p (f x y)))) >==> 
                     (listRevIter (fold_right 
                                     (fun x m1 => 
                                        ((fold_right (fun y m2 => (f x y)::m2) nil l2) ++ m1)) nil l1) 
                                            p))%SSL.
Proof.
  intros; SSL_trans.
  2: apply listRevIter_composeIn_app_ref1 with (f:=fun x => (fold_right (fun y m2 => (f x y)::m2) nil l2)).
  apply listRevIter_monotony.
  intros; apply listRevIter_composeIn_cons.
Qed.

Unset Implicit Arguments.

