Set Implicit Arguments.

Require Export SSLAbsPost.
Require Export SSLLoops.
Require Import FSLFixF.

Definition feasible (St A:Type) (sp:K St A) (st:St) 
  := [< st >] skip >==> (bind sp (fun _ => set st)). 

Lemma feasible_unfold: forall (St A: Type) (sp: K St A) st,
   (feasible sp st) -> (pre (sp st))  -> (angel (sp st) (fun _ => True)).
Proof.
  unfold refInEnv; intros St A sp st H H0;  lapply (refines_unfold H); simpl.
  intuition.
  eapply angel_weaken.
  eauto.
  auto.
  eapply sdemon_weaken.
  eauto.
  simpl.
  intros a; case a; simpl; auto.
Qed.

Lemma feasible_def: forall (St A: Type) (sp: K St A) st,
 (feasible sp st)
    -> [< st >] skip >==> (bind sp (fun _ => set st)).
Proof.
  unfold feasible; auto.
Qed.

Lemma feasible_wp: forall (St A: Type) (sp: K St A) st,
  ((pre (sp st)) -> (angel (sp st) (fun _ => True)))
   -> feasible sp st.
Proof.
  unfold feasible; intros; SSL_simplify_all. 
  intros;
  eapply angel_weaken.
  apply H.
  eapply sdemon_weaken.
  apply H0.
  auto.
  intros a; case a; simpl; auto.
Qed.

Lemma magic_unfeasible: forall (St A: Type) st, ~(feasible (magic (St:=St) A) st).
Proof.
  intros St A st H; generalize (feasible_unfold H).
  simpl; auto.
Qed.

Lemma refined_preserves_feasible: forall (St A B: Type) (sp1: K St A) (sp2:K St B) st,
   (feasible sp1 st) -> ([< st >] (bind sp1 (fun _ => set st)) >==> (bind sp2 (fun _ => set st))) -> (feasible sp2 st).
Proof.
  unfold feasible; intros.
  SSL_trans; eauto.
Qed.

Lemma feasible_bind_split: forall (St A B: Type) (sp1: K St A) (sp2: A->K St B) st,
   (feasible sp1 st) -> (forall a st', feasible (sp2 a) st') -> (feasible (bind sp1 sp2) st).
Proof.
  unfold feasible; intros.
  SSL_unassoc_right.
  SSL_trans.
  apply H.
  SSL_monotony.
  intros st' a; apply refInEnv_trans with  (sp2:= skip -; set st).
  SSL_Trivial.
  SSL_enter_left.
  apply H0.
  SSL_unassoc_left.
  SSL_monotony.
  intros; SSL_Trivial.
Qed.

Lemma feasible_bind_simplify: forall (St A B: Type) (sp1: K St A) (sp2: A->K St B) st,
   (pre (sp1 st) ->
      angel (sp1 st)
          (fun p : St * A =>
           let (st', a) := p in feasible (sp2 a) st')) 
  -> feasible (bind sp1 sp2) st.
Proof.
 unfold feasible; intros.
 SSL_unassoc_right.
 SSL_simplify_right.
 intros; eapply angel_weaken.
 apply H; auto.
 intros x; case x; simpl; intros st' a.
 intros X; apply refInEnv_trans with  (sp2:= skip -; set st).
  SSL_Trivial.
  SSL_enter_left.
  apply X.
  SSL_unassoc_left.
  SSL_monotony.
  intros; SSL_Trivial.
Qed.

Section WHILE.

Variable St : Type.

Variable cond : K St bool.
Variable body : K St unit.

Section While.

Definition while: (K St unit) :=
  (Bind n <- (any nat)
   In (forN n
          (Bind b <- cond
           In (ensure (b = true) -; body)))
       -;  Bind b <- cond 
           In ensure (b = false))%SSL.

Lemma while_refines_invspec :
 forall (invspec : K St unit),
 (forall (st:St), 
    ([< st >] (If cond Then (body -; invspec) Else skip) >==> invspec)%SSL) -> 
  forall (st:St), 
    ([< st >] while >==> invspec)%SSL.
Proof.
  unfold while ; intros invspec H st.
  SSL_simplify_left; intros n.
  SSL_simplify_left; generalize st; elim n; clear n st; simpl.
  (* cas de base *)
  intros st; SSL_trans.
  2: eapply H.
  SSL_simplify_left.
  SSL_monotony.
  intros st' b; case b; simpl.
  SSL_simplify_all; intros; discriminate.
  SSL_Trivial.
  (* HR*)
  intros n HR st.
  SSL_unassoc_left.
  SSL_monotony_left; simpl.
  apply HR.
  SSL_unassoc_left.
  SSL_trans.
  2:apply H.
  SSL_monotony.
  intros st' b; SSL_unassoc_left.
  SSL_simplify_left.
  intros; subst; simpl; SSL_simplify_left.
  auto with isl.
Qed.


Lemma ifwhile_refines_while:
  forall (st:St), ([< st >] (If cond Then (body -; while) Else Skip) >==> while)%SSL.
Proof.
  unfold while; intros.
  apply if_elim. 
  apply refInEnv_trans with (sp2:=(
   (Bind b <- cond In ensure (b = true) -; body) -; 
       Bind n <- any nat
       In forN n (Bind b0 <- cond In ensure (b0 = true) -;  body) -; 
          Bind b0 <- cond In ensure (b0 = false))%SSL).
  SSL_unassoc_right; auto with isl.
  SSL_monotony.
  intros. SSL_unassoc_right. auto with isl.
  SSL_trans.
  apply any_permut_bind with 
   (f:= (fun (_:unit) (n:nat) =>forN n (Bind b0 <- cond In ensure (b0 = true) -;  body) -; 
    Bind b0 <- cond In ensure (b0 = false))%SSL).
  auto with nonempty.
  SSL_simplify_left; intros n.
  SSL_simplify_right.
  intros X; clear X.
  constructor 1 with (x:=(S n)).
  simpl.
  SSL_unassoc_right; auto with isl.
  SSL_simplify_right.
  intros X; clear X.
  constructor 1 with (x:=O); simpl.
  SSL_simplify_right.
  intros; SSL_simplify_left; SSL_monotony.
  intros; SSL_Trivial. 
Qed.

Lemma while_refines_ifwhile:
  forall (st:St), ([< st >] while >==> (If cond Then (body -; while) Else Skip))%SSL.
Proof.
  unfold while; intros.
  SSL_simplify_left; intros n.
  case n; simpl; clear n.
  SSL_simplify_left.
  SSL_simplify_left.
  SSL_monotony.
  intros st2 x; case x; simpl; SSL_simplify_all; (auto || intros; discriminate).
  intros n; SSL_simplify_left; SSL_unassoc_left.
  SSL_unassoc_left.
  SSL_monotony.
  intros st2 x; SSL_unassoc_left.
  SSL_simplify_left; intros; subst; simpl.
  SSL_simplify_left; SSL_monotony.
  intros; SSL_simplify_right.
  intros; constructor 1 with (x:=n).
  auto with isl.
Qed.

End While.

(** notion "classique" d'invariant **) 

Section WhileAbs.

Variable inv: St -> Prop.

Definition inv_preservation: Prop :=
   forall (st:St), (inv st) 
      -> ([< st >] If cond Then body Else Skip >==> 
    (absPost (fun new_st result => (inv new_st))))%SSL.

Lemma inv_preservation_pre:
  inv_preservation -> forall (st:St), (inv st) 
   -> (pre ((If cond Then body Else Skip)%SSL st)).
Proof.
  unfold inv_preservation, refInEnv, refines; intros.
  abstract firstorder auto.
Qed. 

Definition while_abs: (K St unit) :=
 (( Bind st <- (get St) In require (inv st)) -; 
   (Bind tmp_st <- (any St) In set tmp_st) -;
      Bind b <- cond
      In Bind new_st <- (get St)
           In ensure ((inv new_st) /\ b=false))%SSL.

Lemma while_refines_while_abs: inv_preservation -> forall (st:St), ([< st >] while >==> while_abs)%SSL.
Proof.
  intros H1 st; apply while_refines_invspec; clear st.
  intros st;  unfold while_abs at 2.
  SSL_simplify_right.
  intuition auto.
  apply if_elim.
  (* cas boucle *)
  apply refInEnv_trans with
   (sp2:=((If cond Then body Else Skip) -; while_abs)%SSL).
  SSL_unassoc_right.
  SSL_monotony.
  intros st0 b; SSL_simplify_left.
  intros H; subst; simpl; auto with isl.
  SSL_enter_left.
  apply H1; auto.
  SSL_simplify_left. 
  unfold while_abs.
  intros; SSL_simplify_left.
  SSL_simplify_left.
  intuition auto.
  SSL_simplify_left.
  SSL_simplify_left; intros st1.
  SSL_simplify_right.
  intros; SSL_simplify_right.
  intros; constructor 1 with (x:=st1).
  SSL_assoc_left.
  SSL_assoc_left.
  SSL_simplify_left.
  auto with isl.
  (* cas arret *)
  SSL_simplify_right; simpl.
  intros H; clear H.
  constructor 1 with (x:=st).
  SSL_simplify_left.
  apply refInEnv_trans with
   (sp2:=(Bind b <- cond In ensure (b = false))%SSL).
  SSL_monotony.
  intros; SSL_Trivial.
  eapply intro_abspost.
  SSL_trans.
  2:apply H1; auto.
  SSL_monotony.
  intros; SSL_simplify_all; intros; subst; simpl; intuition auto.
  simpl.
  SSL_unassoc_left; auto with isl.
  SSL_monotony.
  intros; SSL_Trivial.
Qed.

End WhileAbs.


Section While2.

Variable inv: St -> Prop.

Hypothesis ipH: (inv_preservation inv).

Variable R: St -> St -> Prop.
Hypothesis Rwf : (well_founded R).
  
Definition variant_decreasing: Prop :=
  forall (st:St), (inv st) -> ([< st >] (Bind b <- cond In ensure (b=true) -; body) >==> 
    (absPost (fun new_st result => (R new_st st))))%SSL.

Hypothesis vdH: variant_decreasing.

Definition while_loop (wh:(K St unit)): (K St unit) 
  := ((Bind st <- (get St) In (require (inv st))) -; If cond Then body -; wh Else Skip)%SSL.

Lemma while2_decreasing: (call_decreasing while_loop R).
Proof.
  unfold call_decreasing, while_loop.
  intros wh st; apply bind_simpl_both; simpl; auto with isl.
  intuition auto.
  apply if_elim.
  apply refInEnv_trans with (sp2:=((Bind b <- cond In ensure (b = true) -;  body)-; wh)%SSL).
  SSL_unassoc_right; SSL_monotony.
  intros; SSL_unassoc_right; auto with isl.
  eapply intro_bind_abspost.
  apply vdH; auto.
  SSL_unassoc_left.
  SSL_monotony.
  intros; SSL_unassoc_left.
  SSL_simplify_left.
  intros; subst; simpl.
  SSL_simplify_left; SSL_monotony.
  intros; SSL_simplify_left.
  intros; SSL_simplify_left.
  unfold ensure, refInEnv.
  apply FSL.bind_simpl_right; simpl.
  intuition auto with wprefines.
  SSL_monotony.
  intros; SSL_simplify_left.
  intros; subst; simpl.
  SSL_Trivial.
Qed.

Lemma while2_monotony:
 forall (sp1 sp2: (K St unit)),
  (forall x : St, refines (sp1 x) (sp2 x))
  -> forall (st:St),
    refines (while_loop sp1 st) (while_loop sp2 st).
Proof.
  intros sp1 sp2 H1 st.
  unfold while_loop.
  SSL_monotony.
  intros; apply if_monotony; auto with isl.
  intros; SSL_monotony.
  intros st3 dum; apply (H1 st3).
Qed.

Definition while2 := (fixF while_loop Rwf).

Hint Resolve while2_decreasing while2_monotony: isl.

Lemma while2_refines_while:
  forall (st:St), ([< st >] while2 >==> ( (Bind st <- (get St) In (require (inv st))) -; while))%SSL.
Proof. 
  unfold while2.  
  intros st; unfold refInEnv; apply fixF_mostmagic;
  auto with isl.
  clear st; intros st; 
  change ([< st >] (while_loop ((Bind st <- (get St) In (require (inv st))) -; while)) 
    >==> (((Bind st <- (get St) In (require (inv st)))) -; while))%SSL.
  unfold while_loop.
  SSL_simplify_both.
  intuition auto.
  SSL_trans.
  2: apply ifwhile_refines_while.
  apply if_elim.
  apply refInEnv_trans with
   (sp2:=
    ((Bind b <- cond
      In ensure (b = true) -; 
       body)-; 
      (Bind st0 <- (get St) In require (inv st0)) -; while)%SSL).
   repeat (intros; SSL_unassoc_right; SSL_monotony).
   eapply intro_bind_abspost.
   SSL_trans.
   2: apply ipH.
   SSL_monotony.
   intros; SSL_simplify_left; intros; subst; simpl; auto with isl.
   auto.
   simpl; SSL_unassoc_left.
   SSL_monotony.
   intros.
   SSL_unassoc_left.
   SSL_simplify_left; intros; subst; simpl.
   SSL_simplify_left; SSL_monotony.
   intros.
   SSL_simplify_left.
   intros; SSL_simplify_left.
   SSL_simplify_left.
   intuition auto.
   SSL_simplify_left.
   auto with isl.
   SSL_monotony.
   intros; SSL_simplify_left; intros; subst.
   simpl; auto with isl.
Qed.

Lemma while2_decaps:
  forall (st:St), (refInEnv st (while_loop while2) while2).
Proof.
  unfold while2.  
  intros st; unfold refInEnv; apply fixF_prefix; auto with isl.
Qed.

Section While2_rect.

Variable P: (K St unit) -> St -> Type.

Hypothesis P_compat: forall (st:St) (sp1 sp2: (K St unit)),
  (P sp1 st)
    -> (refInEnv st sp1 sp2) 
      -> (refInEnv st sp2 sp1)
     ->  (P sp2 st).

Lemma P_while_require:
   (forall (st:St) (Q: St -> Prop) (sp:(K St unit)),
      ((Q st) -> (P sp st)) -> (P ((require (Q st)) -; sp)%SSL) st)
   -> (forall (wh: (K St unit)), (forall (st:St), (P wh st))
             -> forall (st:St), P (while_loop wh) st)
   -> forall (st:St), P while2 st.
Proof.
  unfold while2.
  intros P_require P_preserv st.
  apply inv_fixF_require; eauto with isl.
  clear st; intros st Q sp H; eapply P_compat.
  apply P_require with (Q:=Q) (sp:=sp); auto.
  unfold refInEnv, require, bind, lift.
  intros; apply FSL.require_refined.
  intros; apply FSL.bind_simpl_left; simpl.
  intuition auto with wprefines.
  SSL_simplify_right.
  intuition auto; unfold refInEnv; auto with wprefines.
Qed.

Lemma P_while_ensure:
   (forall (st:St) (Q: St -> Prop) (sp:(K St unit)),
      ((Q st) -> (P sp st)) -> (P ((ensure (Q st)) -; sp)%SSL) st)
   -> (forall (wh: (K St unit)), (forall (st:St), (P wh st))
             -> forall (st:St), P (while_loop wh) st)
   -> forall (st:St), P while2 st.
Proof.
  unfold while2.
  intros P_require P_preserv st.
  apply inv_fixF_ensure; eauto with isl.
  clear st; intros st Q sp H; eapply P_compat.
  apply P_require with (Q:=Q) (sp:=sp); auto.
  SSL_simplify_left; simpl.
  intros; SSL_simplify_left; unfold refInEnv; auto with wprefines.
  unfold refInEnv, ensure, bind, lift.
  intros; apply FSL.ensure_refines.
  intros; apply FSL.bind_simpl_right; simpl.
  intuition auto with wprefines.
Qed.

End While2_rect.

Lemma while_refines_while2:
  forall (st:St), ([< st >] ((Bind st <- (get St) In (require (inv st))) -; while) >==> while2)%SSL.
Proof.
  intros st; pattern while2, st.
  apply P_while_require; clear st.
  intros; SSL_trans; eauto.
  intros; SSL_simplify_right; intuition auto.
  intros wh H st.
  unfold while_loop.
  SSL_simplify_right.
  intuition auto.
  clear H2.
  SSL_simplify_left; intuition auto.
  SSL_simplify_left.
  SSL_trans.
  apply while_refines_ifwhile.
  repeat (intros; SSL_monotony).
  intros.
  SSL_trans. 2:eauto.
  SSL_simplify_right; auto with isl.
Qed.

Lemma while2_refines_while_abs:
  forall (st:St), ([< st >] while2 >==> (while_abs inv))%SSL.
Proof.
  intros; apply refInEnv_pre_intro; simpl.
  intros H1; intuition auto.
  clear H0.
  eapply refInEnv_trans.
  apply while2_refines_while.
  SSL_simplify_left; intuition auto.
  SSL_simplify_left.
  apply while_refines_while_abs; auto.
Qed.

End While2.

Section WhileWF.

Variable inv: St -> Prop.

Variable R: St -> St -> Prop.

Definition whileWF:=
 ((Bind st <- (get St) In
   require ((inv_preservation inv) /\ (variant_decreasing inv R) /\ (well_founded R) /\ (inv st))) -;
   while)%SSL.

Definition whileWF2:=
 ((require ((inv_preservation inv) /\ (variant_decreasing inv R) /\ (well_founded R))) -;
  Bind Rwf <- (any (well_founded R)) In
     (while2 inv Rwf))%SSL.

Lemma whileWF_ref1: forall (st:St), (refInEnv st whileWF whileWF2).
Proof.
  intros; unfold whileWF, whileWF2; SSL_simplify_right.
  intuition auto.
  SSL_simplify_right.
  intros; constructor 1 with (x:=H3).
  SSL_simplify_left; SSL_trans.
  2: apply while_refines_while2; auto.
  SSL_monotony.
  SSL_Trivial.
Qed.

Lemma whileWF_ref2: forall (st:St), (refInEnv st whileWF2 whileWF).
Proof.
  intros; unfold whileWF, whileWF2. SSL_simplify_right.
  intuition auto.
  SSL_simplify_left.
  intuition auto.
  SSL_simplify_left.
  SSL_simplify_left.
  intros; SSL_simplify_left; SSL_trans.
  apply while2_refines_while; auto.
  SSL_simplify_left.
  intuition auto with isl.
Qed.

Section WhileWF_ind.

Variable P: (K St unit) -> St -> Prop.

Hypothesis P_compat: forall (st:St) (sp1 sp2: (K St unit)),
  (P sp1 st)
    -> (refInEnv st sp1 sp2) 
      -> (refInEnv st sp2 sp1)
     ->  (P sp2 st).

Lemma P_whileWF_require:
   (forall (st:St) (Q: St -> Prop) (sp:(K St unit)),
      ((Q st) -> (P sp st)) -> (P ((require (Q st))  -; sp)%SSL) st)
   -> (forall (while: (K St unit)), (forall (st:St), (P while st))
             -> forall (st:St), P (while_loop inv while) st)
   -> forall (st:St), P whileWF st.
Proof.
  intros P_require P_preserv st.
  eapply P_compat.
  2: apply whileWF_ref2.
  unfold whileWF2.
  apply P_require with (Q:=(fun (_:St) => inv_preservation inv /\ variant_decreasing inv R /\ well_founded R)).
  intros H; decompose [Logic.and] H.
  eapply P_compat.
  apply P_while_require with (inv:=inv) (Rwf:=H3); auto.
  SSL_simplify_right; simpl; intros.
  constructor 1 with (x:=H3); auto with isl.
  SSL_simplify_left; simpl; intros; SSL_simplify_left.
  unfold while2, refInEnv.
  apply fixF1_refines_fixF2; auto with wprefines.
  intros; apply while2_monotony; auto.
  intros; apply while2_monotony; auto.
  intros; apply while2_decreasing; auto.
  apply whileWF_ref1.
Qed.

End WhileWF_ind.

Lemma whileWF_feasible:
  (forall (st:St), (inv st) -> feasible (If cond Then body Else Skip) st)
  -> forall (st:St), feasible whileWF st.
Proof.
  intros; apply P_whileWF_require.
  intros; eapply refined_preserves_feasible; eauto.
  SSL_monotony.
  intros; eapply feasible_bind_simplify; simpl; intuition.
  intros; unfold while_loop, feasible.
  SSL_unassoc_right.
  SSL_simplify_right.
  intuition.
  apply feasible_def.
  intros; eapply refined_preserves_feasible.
  apply H; auto.
  SSL_unassoc_left.
  SSL_unassoc_right.
  SSL_monotony.
  intros st2 x; case x; auto with isl.
  SSL_unassoc_right.
  SSL_monotony.
  intros; eapply refInEnv_trans with (sp2:=skip -; set st0).
  SSL_Trivial.
  SSL_enter_left.
  apply H0.
  SSL_unassoc_left.
  SSL_monotony.
  SSL_Trivial.
Qed.

Lemma whileWF_spec: 
   forall (st:St), ([< st >] whileWF >==> 
     (require ((inv_preservation inv) /\ (variant_decreasing inv R) /\ (well_founded R))) -;
      (while_abs inv))%SSL.
Proof.
  intros st; SSL_simplify_right; intuition auto.
  unfold whileWF.
  apply refInEnv_pre_intro; simpl.
  intros.
  SSL_simplify_left; intuition auto.
  SSL_simplify_left. 
  apply while_refines_while_abs; auto.
Qed.

Lemma whileWF_unfold_left:
 forall (st:St), ([< st >] (If cond Then body -; whileWF Else Skip) >==> whileWF)%SSL.
Proof.
  intros; SSL_trans. 
  2: apply whileWF_ref2.
  unfold whileWF2; SSL_simplify_right.
  intuition auto.
  SSL_simplify_right.
  intros D; clear D.
  constructor 1 with (x:=H3).
  SSL_simplify_left; SSL_trans. 2: apply while2_decaps.
  unfold while_loop.
  SSL_simplify_right.
  intros; SSL_monotony.
  intros; SSL_monotony.
  intros. SSL_trans.
  apply whileWF_ref1.
  unfold whileWF2; SSL_simplify_left; intuition auto with nonempty.
  SSL_simplify_left.
  SSL_simplify_left.
  intros;  SSL_simplify_left; unfold while2, refInEnv.
  apply fixF1_refines_fixF2; auto with wprefines.
  intros; apply while2_monotony; auto.
  intros; apply while2_monotony; auto.
  intros; apply while2_decreasing; auto.
  auto.
Qed.

Lemma invspec_refines_whileWF :
 forall (invspec : K St unit),
 (forall (st:St), 
    ([< st >] invspec >==> (while_loop inv invspec))%SSL) -> 
  forall (st:St), ([< st >] invspec >==> whileWF)%SSL.
Proof.
  intros invspec H st.
  pattern whileWF, st. apply P_whileWF_require; clear st; auto with isl.
  intros; SSL_trans; eauto.
  intros; SSL_simplify_right; intuition auto.
  intros; SSL_trans.
  apply H.
  unfold while_loop.
  SSL_monotony.
Qed.

End WhileWF.

End WHILE.


Lemma while_monotonic_prec: 
  forall (St:Type) (cond1 cond2: K St bool) (body1 body2: K St unit),
    (forall (st:St), [< st >] (bind cond1 (fun b => ensure (b=false))) >==> 
                              (bind cond2 (fun b => ensure (b=false))))
    -> (forall (st:St), [< st >] (bind cond1 (fun b => ensure (b=true) -; body1)) >==> 
                                 (bind cond2 (fun b => ensure (b=true) -; body2)))
     -> forall (st:St), [< st >] (while cond1 body1) >==> (while cond2 body2).
Proof.
  intros; unfold while.
  SSL_monotony.
Qed.

Lemma while_monotonic: 
  forall (St:Type) (cond1 cond2: K St bool) (body1 body2: K St unit),
    (forall (st:St), [< st >] cond1 >==> cond2)
    -> (forall (st:St), [< st >] body1 >==> body2)
     -> forall (st:St), [< st >] (while cond1 body1) >==> (while cond2 body2).
Proof.
  intros; apply while_monotonic_prec; intros; SSL_monotony.
Qed.

Lemma while_morph: 
  forall (St:Type) (cond1 cond2: K St bool) (body1 body2: K St unit) (G: K St unit -> K St unit),
    (forall k1 k2, (forall st, [< st >] k1 >==> k2) -> forall st, [< st >] G k1 >==> G k2)
    -> (forall st, [< st >] (bind cond1 (fun b => ensure (b=false))) >==> 
                                 (G (bind cond2 (fun b => ensure (b=false)))))
    -> (forall k st, [< st >] (bind cond1 (fun b => ensure (b=true) -; body1 -; G k)) >==> 
                              (G (bind cond2 (fun b => ensure (b=true) -; body2 -; k))))
      -> forall (st:St), [< st >] (while cond1 body1) >==> (G (while cond2 body2)).
Proof.
  intros St cond1 cond2 body1 body2 G Gmon H1 H2; apply while_refines_invspec.
  intros st; apply if_elim.
  SSL_trans. apply H2.
  apply Gmon.
  intros; SSL_trans. 2:apply ifwhile_refines_while.
  SSL_monotony.
  intros; SSL_simplify_left.
  intros; subst.
  SSL_simplify_left.
  auto with isl.
  intros;   SSL_assoc_left.
  apply bind_val_r_ref1.
  SSL_trans. apply H1.
  apply Gmon.
  intros; SSL_trans. 2:apply ifwhile_refines_while.
  SSL_monotony.
  SSL_simplify_all; intuition subst; simpl; auto.
  intros a; case a; auto with isl.
Qed.


Unset Implicit Arguments.
