(* Below, a little "hack" to deal with a coqdoc bug *)
(** printing [(*]*) %{\tt[}%*)
(** printing ] %{\tt]}%*)
(* begin hide *)
Require Export SSL.
(* end hide *)

(** ** Module [SSLfriendly]: Macros and tactics in the prover 
*)

Hint Resolve refInEnv_refl bind_val_r_ref1 bind_val_r_ref2: isl.
Hint Resolve bind_monotony if_monotony : isl.
Hint Resolve set_skip_left set_skip_right: isl.

(** *** Simple tactics to simplify [refInEnv] proofs using wp-calculus 
Below, tactics [SSL_simplify_all] and [SSL_Trivial] apply wp to the whole goal. *)
Ltac SSL_simplify_all := intros; apply refInEnv2wp; simpl.
Ltac SSL_Trivial := try (SSL_simplify_all; intuition eauto (*with wipi*); fail).
(** Tactic [SSL_pre_intro] introduces the precondition of the right-side of the goal into hypothesis *)
Ltac SSL_pre_intro := apply refInEnv_pre_intro; simpl.
(** Tactic [SSL_simplify_both] corresponds to [bind_simpl_both] lemma *)
Ltac SSL_simplify_both :=
  apply bind_simpl_both; simpl in |- *; auto with isl.
(** Roughly speaking, [SSL_simplify_left] corresponds to
[bind_simpl_left] lemma.  But, when the first part of the left
expression is a set expression, it tries also to eliminate it.  If it
fails, then it uses [set_swap_left_right] to move the set expression
on the right side of the goal.  
*)
Ltac SSL_skip_set := apply set_skip_left ; [(*]*) simpl; reflexivity | idtac ].
Ltac SSL_simplify_left :=
  (SSL_skip_set ||
    (apply set_simplify_left; [(*]*) SSL_Trivial; fail | idtac ] ) ||
      (apply set_swap_leftright; 
         try (apply set_simplify_right; [(*]*) SSL_Trivial; fail | idtac ] )) ||
        (apply bind_simpl_left; simpl in |- *;
          try SSL_skip_set)); simpl.
(** Tactic [SSL_simplify_right] is quite symmetric to [SSL_simplify_left], except that,
when the right expression is a set expression, it is never moved on the left.
*)
Ltac SSL_simplify_right :=
  (apply set_simplify_right ; [(*]*) SSL_Trivial; fail | idtac ] ) ||
      (apply bind_simpl_right; simpl in |- *;
        try (let z:= fresh "H" in (intro z; SSL_skip_set; generalize z; clear z))); simpl.

(** *** Tactics using monotony to decompose the goal *)
Ltac SSL_enter_left := eapply bind_enter_left; simpl.
Ltac SSL_enter_right := eapply bind_enter_right; simpl.
Ltac SSL_monotony_left :=
  eapply seq_monotony_left; simpl (* || eapply bind_monotony_left *).
Ltac SSL_monotony_right :=
  eapply seq_monotony_right; simpl (* || eapply bind_monotony_right *).
Ltac SSL_monotony :=
  (apply if_monotony || apply bind_monotony); simpl; auto with isl.
(* begin hide *)
Ltac SSL_trans := eapply refInEnv_trans.
(* end hide *)

(** *** Tactics changing the goal using associativity *)
Ltac SSL_unassoc_left := apply bind_left_unassoc.
Ltac SSL_unassoc_right := apply bind_right_unassoc.
Ltac SSL_assoc_left := 
  apply seq_left_assoc || 
  apply bind_left_assoc.
Ltac SSL_assoc_right := 
  apply seq_right_assoc || 
  apply bind_right_assoc.


(** *** Syntactic Sugar *)

Delimit Scope SSL_scope with SSL. 
Bind Scope SSL_scope with K.

Notation "'Skip'" := (val tt) (at level 90): SSL_scope.
Notation "'skip'" := (val tt) (at level 90): SSL_scope.

(* Notations for binders *)

Notation "'If' pb 'Then' sp1 'Else' sp2" :=
  (bind pb (fun b => if b then sp1 else sp2))
  (at level 20, pb, sp1, sp2 at level 90): SSL_scope.

Notation "'Bind' x <- sp1 'In' sp2" := (bind sp1 (fun x => sp2))
  (at level 95, sp1 at level 90, right associativity): SSL_scope.

Notation "sp1  -;  sp2" := (bind sp1 (fun _ : unit => sp2))
  (at level 95, right associativity): SSL_scope.


(* begin hide *)
Notation " [<  st  >]  sp1  >==>  sp2" := (refInEnv st sp1 sp2) 
 (at level 200): SSL_scope.
(* end hide *)

(* A coqdoc bug again... *)
(** [Notation "]%{\tt[<}%[ st ]%{\tt>]}%[  sp1  >==>  sp2" := (refInEnv st sp1 sp2)] *)
(** %~~%[ (at level 200): SSL_scope.] *)

