Set Implicit Arguments.

Require Export List.
Require Export TheoryList.
 
Hint Resolve allS_nil allS_cons: lists.
 
Lemma allS_in_ext: forall (A:Set) (P: A -> Prop) (l:list A),
  (AllS P l) -> forall (x:A), (List.In x l) -> (P x).
Proof.
  induction 1; simpl; intuition auto.
  subst; auto.
Qed.

Lemma in_allS_ext: forall (A:Set) (P: A -> Prop) (l:list A),
  (forall (x:A), (List.In x l) -> (P x)) -> (AllS P l).
Proof.
  induction l; simpl; intuition auto with lists.
Qed.
                                                                                
Lemma allS_not_in: forall (A:Set) (l:list A) (x:A),
  (AllS (fun y => y<>x) l) -> ~(List.In x l).
Proof.
  intros A l x H; generalize (allS_in_ext H).
  unfold not; firstorder auto.
Qed.
                   Lemma not_in_allS: forall (A:Set) (l:list A) (x:A),
  ~(List.In x l) -> (AllS (fun y => y<>x) l).
Proof.
  intros; apply in_allS_ext.
  unfold not; intros; subst.
  auto.
Qed.
                                                                                   
Lemma allS_app_intro: forall (A:Set) (P: A -> Prop) (l1 l2:list A),
  (AllS P l1) -> (AllS P l2) -> (AllS P (l1 ++ l2)).
Proof.
  induction 1; simpl; intuition auto with lists.
Qed.
                                                                                   
Lemma allS_weaken: forall (A:Set) (P Q: A -> Prop) (l:list A),
  (AllS P l) -> (forall (x:A), (P x) -> (Q x)) -> (AllS Q l).
Proof.
  induction 1; simpl; intuition auto with lists.
Qed.
                                                                                   
Hint Resolve allS_app_intro in_allS_ext allS_weaken:lists.
                                                                                   
Inductive doublefree (A:Set): (list A) -> Prop :=
  | doublefree_nil: (doublefree nil)
  | doublefree_hd: forall (x:A) (l:(list A)),
       ~(List.In x l) -> (doublefree l) -> (doublefree (x::l)).
                                                                                   
Hint Resolve doublefree_nil doublefree_hd:lists.
                                                                                   
                                                             
Lemma doublefree_ext1: forall (A:Set) (l:list A),
  (forall (x:A), (List.In x l) ->
      exists l1:list A, exists l2:list A,
        (l=(app l1 (cons x l2))
          /\ ~(List.In x l1) /\ ~(List.In x l2)))
  -> (doublefree l).
Proof.
  induction l; simpl; auto with lists.
  intros; apply doublefree_hd.
  (* ~ In *)
  intros; case (H a); auto.
  intros l1 H1; case H1; clear H1.
  case l1; simpl;
  intros; intuition auto;
  inversion H2; subst; tauto.
  (* doublefree *)
  apply IHl.
  intros b H1.
  case (H b); auto.
  intros l1 H2; case H2; clear H2.
  intros l2; case l1; simpl; clear l1.
  intuition auto.
  inversion H2; subst.
  intuition auto.
  intros c l1; intuition auto.
  inversion H2. subst a. clear H2.
  constructor 1 with (x:=l1).
  constructor 1 with (x:=l2).
  intuition auto.
Qed.

Lemma in_decompose: forall (A:Set) (l:list A) (x:A),
  (List.In x l) ->
      exists l1:list A, exists l2:list A, l=l1++(x::l2).
Proof.
  induction l; simpl.
  tauto.
  intros y H; case H.
  intros; subst.
  constructor 1 with (x:=(nil (A:=A))).
  constructor 1 with (x:=l).
  simpl; auto.
  intros H0; case (IHl _ H0).
  intros l1 H1; case H1; clear H1.
  intros l2 H1; subst.
  constructor 1 with (x:=(cons a l1)).
  constructor 1 with (x:=l2).
  simpl; auto.
Qed.
                                                                                   
Lemma doublefree_app_notin1: forall (A:Set) (l1 l2:list A) (x:A),
  (doublefree (l1 ++ (x::l2)))
    -> ~(List.In x l1).
Proof.
  induction l1; simpl; auto with lists.
  intros l2 x H; inversion_clear H.
  intros X; case X.
  intros; subst.
  apply H0.
  apply in_or_app.
  simpl; auto.
  apply IHl1 with (l2:=l2) (x:=x).
  auto.
Qed.

Lemma doublefree_app_notin2: forall (A:Set) (l1 l2:list A) (x:A),
  (doublefree (l1 ++ (x::l2)))
    -> ~(List.In x l2).
Proof.
  induction l1; simpl; auto with lists.
  intros l2 x H; inversion_clear H; auto.
  intros l2 x H; inversion_clear H; auto.
Qed.
                                                                                   
Inductive enumlist (A:Set): (A -> Prop) -> (list A) -> Prop :=
| enumlist_nil:  forall (cond:A -> Prop),
    (forall (x:A), ~(cond x)) -> (enumlist cond nil)
| enumlist_cons: forall (cond: A -> Prop) (x:A) (l:list A),
   (forall (y:A), (cond y) -> x=y \/ x<>y)
    -> (cond x) ->
        (enumlist (fun y => x<>y /\ (cond y)) l) ->
          (enumlist cond (x::l)).
                                                                                   


Lemma enumlist_ext1: forall (A:Set) (l:list A) (cond: A -> Prop),
 (AllS cond l)
   -> (forall (x:A), (cond x) ->
       exists l1:list A, exists l2:list A,
         (l= l1 ++ (x::l2))
           /\ ~(List.In x l1) /\ ~(List.In x l2))
     -> (enumlist cond l).
Proof.
  induction l; simpl.
  intros; apply enumlist_nil.
  intros x H1; case (H0 x H1).
  intros l H2; case H2; clear H2; intros m.
  intros H2; case H2; clear H2; case l; simpl.
  intros H3; inversion H3.
  intros a l0 H3; inversion H3.
  intros cond H H0.
  inversion_clear H.
  apply enumlist_cons; auto.
  intros y H.
  case (H0 y H); clear H0.
  intros l1 H0; case H0; clear H0; intros l2.
  intros H0; case H0; clear H0.
  case l1; simpl; clear l1.
  intros H3; inversion H3; subst; auto.
  intros b l1 H3; inversion H3; subst; intuition auto.
  apply IHl; intuition auto.
  apply in_allS_ext.
  intros y H3; generalize (allS_in_ext H2 y H3); intros H4.
  intuition auto.
  subst. case (H0 y H4); clear H0.
  intros l1 H0; case H0; clear H0; intros l2.
  intros H0; case H0; clear H0.
  case l1; simpl; clear l1.
  intros H0; inversion H0; subst; intuition auto.
  intros b l1 H0; inversion H0; subst; intuition auto.
  case (H0 x H4); clear H0.
  intros l1 H0; case H0; clear H0; intros l2.
  intros H0; case H0; clear H0.
  case l1; simpl; clear l1.
  intros H0; inversion H0; subst; intuition auto.
  intros b l1 H0; inversion H0; subst; intuition auto.
  constructor 1 with (x:=l1).
  constructor 1 with (x:=l2).
  intuition auto.
Qed.
                                                                                   
Lemma enumlist_AllScond: forall (A:Set) (cond: A -> Prop) (l:list A),
 (enumlist cond l) -> (AllS cond l).
Proof.
 induction 1; auto with lists.
 apply allS_cons; auto with lists.
 eapply allS_weaken. eauto.
 simpl; intuition auto.
Qed.
                                                                                   
Hint Resolve enumlist_AllScond: lists.
                                                                                   
Lemma enumlist_ext2: forall (A:Set) (cond: A -> Prop) (l:list A),
 (enumlist cond l) -> forall (x:A), (cond x) ->
       exists l1:list A, exists l2:list A,
         (l= l1 ++ (x::l2))
           /\ ~(List.In x l1) /\ ~(List.In x l2).
Proof.
 induction 1.
 firstorder auto.
 intros y H2.
 case (H y H2).
 intros; subst.
 constructor 1 with (x:=(nil(A:=A))).
 constructor 1 with (x:=l); simpl; intuition auto.
 generalize (allS_in_ext (enumlist_AllScond H1) y H3).
 intuition auto.
 intros H3; case (IHenumlist y).
 intuition auto.
 intros l1 H4; case H4; clear H4; intros l2.
 intros. intuition subst. constructor 1 with (x:=x::l1).
 constructor 1 with (x:=l2); simpl; intuition auto.
Qed.
                                                                                   
Lemma enumlist_def1: forall (A:Set) (cond: A -> Prop) (l:list A),
  (forall (x:A), (List.In x l) <-> (cond x))
   -> (doublefree l)
      -> (enumlist cond l).
Proof.
  intros; apply enumlist_ext1.
  apply in_allS_ext.
  firstorder auto.
  intros x H1.
  assert (H2: List.In x l).
  firstorder auto.
  case (in_decompose _ _ H2).
  intros l1 H3; case H3; intro l2; intros; subst.
  constructor 1 with (x:=l1).
  constructor 1 with (x:=l2).
  constructor 1; auto.
  constructor 1.
  apply doublefree_app_notin1 with (l2:=l2).
  firstorder auto.
  apply doublefree_app_notin2 with (l1:=l1).
  firstorder auto.
Qed.
                                                                                   
Lemma enumlist_in_cond: forall (A:Set) (cond: A -> Prop) (l:list A),
  (enumlist cond l) -> forall (x:A), (List.In x l) -> (cond x).
Proof.
  intros; eapply allS_in_ext.
  2: eauto.
  auto with lists.
Qed.
                                                                                   
Lemma enumlist_cond_in: forall (A:Set) (cond: A -> Prop) (l:list A),
  (enumlist cond l) -> forall (x:A), (cond x) -> (List.In x l).
Proof.
  intros A cond l H x H0; case (enumlist_ext2 H x H0).
  intros l1 H2; case H2; clear H2.
  intros l2 H2; intuition subst.
  apply in_or_app.
  constructor 2; simpl; auto.
Qed.
                                                                                   
Lemma enumlist_doublefree: forall (A:Set) (cond: A -> Prop) (l:list A),
  (enumlist cond l) -> (doublefree l).
Proof.
  intros; apply doublefree_ext1.
  intros; apply (enumlist_ext2 H).
  eapply enumlist_in_cond; eauto.
Qed.
                                                                                   
Hint Resolve enumlist_nil: lists.
                                                                                   
Lemma enumlist_compatible: forall (A:Set) (l:list A) (cond1 cond2: A -> Prop),
  (forall (x:A), (cond1 x) <-> (cond2 x)) ->
    (enumlist cond1 l) -> (enumlist cond2 l).
Proof.
  intros; apply enumlist_def1.
  intros x; case (H x); intuition auto.
  apply H1; eapply enumlist_in_cond; eauto.
  eapply enumlist_cond_in; eauto.
  eapply enumlist_doublefree; eauto.
Qed.
                                                                                   
Lemma enumlist_app: forall (A:Set) (l1 l2:list A) (cond1: A -> Prop),
  (enumlist cond1 l1) ->
   forall (cond2: A -> Prop),
    (forall (x:A), (cond1 x) -> (cond2 x)) ->
    (forall (x:A), (cond2 x) -> (cond1 x) \/ ~(cond1 x)) ->
       (enumlist (fun x => ~(cond1 x) /\ (cond2 x)) l2) ->
         (enumlist cond2 (l1++l2)).
Proof.
  induction 1; simpl.
  intros; eapply enumlist_compatible.
  2: eauto.
  firstorder auto.
  intros; apply enumlist_cons; auto.
  intros y H5; case (H3 y H5); auto.
  intros; constructor 2.  unfold not; intros; subst.
  intuition auto.
  apply IHenumlist; clear IHenumlist.
  intuition auto.
  intros y H5; case (H3 y); intuition auto.
  intros; eapply enumlist_compatible.
  2: apply H4.
  intuition subst; auto.
Qed.
                                                                                   
Lemma fold_right_app_ext:
  forall (A B:Set) (f: A -> list B) (l1 l2:list A),
      (fold_right (fun x m => (f x) ++ m) nil (l1 ++ l2)) =
       (fold_right (fun x m => (f x) ++ m) nil l1) ++ (fold_right (fun x m => (f x) ++ m) nil l2).
Proof.
  induction l1; simpl; auto.
  intros; rewrite IHl1.
  rewrite app_ass.
  auto.
Qed.

Lemma fold_right_monotony:
  forall (A B:Set) (f g: A -> B -> B),
     (forall (x:A) (y:B), (f x y)=(g x y))
      ->  forall (c:B) (l:list A),
            (fold_right f c l) = (fold_right g c l).
Proof.
  induction l; simpl; auto.
  rewrite IHl.
  rewrite H.
  auto.
Qed.
                                                                                   
Lemma fold_right_cons_ext:
  forall (A B:Set) (f: A -> B) (l1 l2:list A),
      (fold_right (fun x m => (f x)::m) nil (l1 ++ l2)) =
       (fold_right (fun x m => (f x)::m) nil l1) ++ (fold_right (fun x m => (f x)::m) nil l2).
Proof.
  intros; rewrite fold_right_monotony with (g:=(fun x m => ((f x)::nil)++m)).
  intros. rewrite fold_right_app_ext with (f:=(fun x => (f x)::nil)).
  simpl; auto.
  simpl; auto.
Qed.
                                                                                   
Lemma fold_right_allS:
  forall (A B:Set) (f: A -> list B) (P: A -> Prop) (Q: B -> Prop) (l:list A),
      (forall (x:A), (P x) -> AllS Q (f x))
       -> (AllS P l)
           -> (AllS Q (fold_right (fun x m => (f x) ++ m) nil l)).
Proof.
  induction 2; simpl; auto with lists.
Qed.
                                                                                   
Lemma fold_right_allS_cons:
  forall (A B:Set) (f: A -> B) (P: A -> Prop) (Q: B -> Prop) (l:list A),
      (forall (x:A), (P x) -> Q (f x))
       -> (AllS P l)
           -> (AllS Q (fold_right (fun x m => (f x) :: m) nil l)).
Proof.
  induction 2; simpl; auto with lists.
Qed.


Lemma fold_right_app_enumlist:
  forall (A B:Set) (f: A -> list B) (P: A -> Prop) (l:list A),
   (enumlist P l) ->
      forall (Q: B -> Prop),
        (forall (y:B), (Q y) -> exists x:A, (P x) /\ List.In y (f x)) ->
          (forall (x1 x2:A), (P x1) -> (P x2) -> (exists y:B, (List.In y (f x1)) /\ (List.In y (f x2))) -> x1=x2) ->
          (forall (x:A), (P x) -> (AllS Q (f x)) /\ doublefree (f x)) ->
             (enumlist Q (fold_right (fun x m => (f x) ++ m) nil l)).
Proof.
  induction 1; simpl.
  intros; apply enumlist_nil.
  intros x H3; case (H0 x H3).
  firstorder auto.
  intros Q H2 H3 H4.
  apply enumlist_app with
   (cond1:=(fun y:B => (Q y) /\ (List.In y (f x)))).
  case (H4 x H0).
  intros; apply enumlist_def1; auto.
  intros y; intuition auto.
  eapply allS_in_ext; eauto.
  intuition auto.
  intros y H5; case (H2 y H5).
  intros x0; intuition auto.
  case (H x0); intuition subst; auto.
  constructor 2; intuition eauto.
  apply IHenumlist; try ((intuition auto); fail).
  intros y H5; case (H2 y).
  intuition auto.
  intros x0 H6.
  constructor 1 with (x:=x0).
  intuition subst; eauto.
  intros x0 H5; case (H4 x0); intuition auto.
  apply in_allS_ext.
  intros y; intuition eauto.
  eapply allS_in_ext; eauto.
Qed.
                                                                                   
Lemma fold_right_cons_enumlist:
  forall (A B:Set) (f: A -> B) (P: A -> Prop) (l:list A),
   (enumlist P l) ->
      forall (Q: B -> Prop),
      (forall (y:B), (Q y) -> exists x:A, (P x) /\ y=(f x))
       -> (forall (x1 x2:A), (P x1) -> (P x2) -> (f x1)=(f x2) -> x1=x2)
         -> (forall (x:A), (P x) -> Q (f x)) ->
             (enumlist Q (fold_right (fun x m => (f x) :: m) nil l)).
Proof.
  intros. eapply fold_right_app_enumlist with (f:=(fun x => (f x)::nil)).
  eauto.
  simpl; firstorder auto.
  simpl; intros.
  decompose [ex Logic.and] H5.
  intuition auto.
  firstorder subst; auto.
  intuition auto with lists.
Qed.
  
Unset Implicit Arguments.                        
                                                                                 
