(* library for compatibility before v8.1 *)

Set Implicit Arguments.



(*** ***)
Lemma Acc_no_infinite_seq:
  forall (A:Type) (R: A -> A -> Prop) (x:A),
       (Acc R x) ->
       forall (f: nat -> A), (f O)=x
          -> ~(forall (n:nat), (R (f (S n)) (f n))).
Proof.
  induction 1.
  intros f H1 H2.
  apply H0 with (y:=(f (S 0))) (f:=(fun n => (f (S n)))); simpl; subst; auto.
Qed.

Theorem wf_no_infinite_seq:
  forall (A:Type) (R: A -> A -> Prop) (f: nat -> A),
     (well_founded R) 
          -> ~(forall (n:nat), (R (f (S n)) (f n))).
Proof.
  intros A R f H; apply Acc_no_infinite_seq with (x:=(f 0)); auto.
Qed.  

(*******)
Lemma Acc_enforce: forall (A:Type) (R1 R2: A -> A -> Prop) (x:A),
 (Acc R1 x) -> (forall (a b: A), (R2 a b)->(R1 a b)) ->
   (Acc R2 x).
Proof.
  induction 1.
  intros; constructor 1.
  auto.
Qed.

Lemma Acc_remove_invariant:
  forall (A:Type) (R: A -> A -> Prop) (P: A -> Prop),
     (forall (x y:A), (P x) -> (R y x) -> (P y))
      -> forall (x:A), (Acc (fun (y x:A) => (P x) /\ (R y x)) x)
                             -> (P x) -> (Acc R x).
Proof.
  induction 2.
  intros; constructor 1.
  intuition eauto.
Qed.

Lemma Acc_enforce_invariant:
  forall (A:Type) (R: A -> A -> Prop) (P: A -> Prop),
   (forall (x y:A), (P x) -> (R y x) -> (P y))
      -> forall (x:A), (Acc R x) -> (P x) -> 
             (Acc (fun (y x:A) => (P x) /\ (R y x)) x).
Proof.
  induction 2.
  constructor 1.
  intuition eauto.
Qed.

(*** left lexicographic order ***)

Definition lexico_l (A B:Type) (R1:A->A->Prop) (R2:B->B->Prop) (c1 c2: (prodT A B)) :=
  (R1 (fstT c1) (fstT c2)) /\ ((fstT c1)=(fstT c2) /\ (R2 (sndT c1) (sndT c2))).
 
Lemma well_founded_lexico_l:
 forall (A B:Type) (R1:A->A->Prop) (R2:B->B->Prop),
  (well_founded R1) -> (well_founded R2) ->
   (well_founded (lexico_l R1 R2)).
Proof.
  unfold well_founded, lexico_l; intros A B R1 R2 H1 H2 a; case a; intros x y; clear a.
  generalize y; elim (H1 x); clear y x; intros x H3 HR1 y.
  elim (H2 y); clear y; intros y H4 HR2.
  constructor 1; intros a; case a; clear a; simpl; intros x' y' H5; case H5; auto.
Qed.

(** Copy of standard library ("Set" in "Type") *)

Section Inverse_Image.

  Variables A B : Type.
  Variable R : B -> B -> Prop.
  Variable f : A -> B.

  Let Rof (x y:A) : Prop := R (f x) (f y).

  Remark Acc_lemma : forall y:B, Acc R y -> forall x:A, y = f x -> Acc Rof x.
  Proof.
    induction 1 as [y _ IHAcc]; intros x H.
    apply Acc_intro; intros y0 H1.
    apply (IHAcc (f y0)); try trivial.
    rewrite H; trivial.
  Qed.

  Lemma Acc_inverse_image : forall x:A, Acc R (f x) -> Acc Rof x.
  Proof.
    intros; apply (Acc_lemma (y:=(f x))); trivial.
  Qed.

  Theorem wf_inverse_image : well_founded R -> well_founded Rof.
  Proof.
    red in |- *; intros; apply Acc_inverse_image; auto.
  Qed.

  Variable F : A -> B -> Prop.
  Let RoF (x y:A) : Prop :=
    exists2 b : B, F x b & (forall c:B, F y c -> R b c).

  Lemma Acc_inverse_rel : forall b:B, Acc R b -> forall x:A, F x b -> Acc RoF x.
  Proof.
    induction 1 as [x _ IHAcc]; intros x0 H2.
    constructor; intros y H3.
    destruct H3.
    apply (IHAcc x1); auto.
  Qed.


  Theorem wf_inverse_rel : well_founded R -> well_founded RoF.
  Proof.
    red in |- *; constructor; intros.
    case H0; intros.
    apply (Acc_inverse_rel (b:=x)); auto.
  Qed.

End Inverse_Image.

Unset Implicit Arguments.
