Require Export ZArith.
Require Export ZArith_dec.
Require Export Zbool.
Require Export Bool.

Lemma Zle_le_bool : forall n m:Z, (n <= m)%Z -> Zle_bool n m = true.
Proof.
  unfold Zle, Zle_bool in |- *. 
  intros x y. case (x ?= y)%Z; trivial. 
  intro H; elim (H (refl_equal _)).
Qed.

Lemma Zgt_le_bool : forall n m:Z, (n > m)%Z -> Zle_bool n m = false.
Proof.
  unfold Zgt, Zle_bool in |- *. 
  intros x y. case (x ?= y)%Z; trivial || intros; discriminate.
Qed.

Lemma Zle_bool_le: forall n m:Z, Zle_bool n m = true -> (n <= m)%Z.
Proof.
  unfold Zle_bool, Zle in |- *. intros x y. unfold not in |- *.
  case (x ?= y)%Z; (trivial || (intros; discriminate)).
Qed.

Lemma Zle_bool_gt: forall n m:Z, Zle_bool n m = false -> (n > m)%Z.
Proof.
  unfold Zle_bool, Zgt in |- *. intros x y. unfold not in |- *.
  case (x ?= y)%Z; (trivial || (intros; discriminate)).
Qed.

Lemma Zlt_lt_bool : forall n m:Z, (n < m)%Z -> Zlt_bool n m = true.
Proof.
  unfold Zlt, Zlt_bool in |- *. 
  intros x y. case (x ?= y)%Z; trivial || (intros; discriminate).
Qed.

Lemma Zge_lt_bool : forall n m:Z, (n >= m)%Z -> Zlt_bool n m = false.
Proof.
  unfold Zge, Zlt_bool in |- *. 
  intros x y. case (x ?= y)%Z; trivial. 
  intro H; elim (H (refl_equal _)).
Qed.

Lemma Zlt_bool_lt: forall n m:Z, Zlt_bool n m = true -> (n < m)%Z.
Proof.
  unfold Zlt_bool, Zlt in |- *. intros x y. unfold not in |- *.
  case (x ?= y)%Z; (trivial || (intros; discriminate)).
Qed.

Lemma Zlt_bool_ge: forall n m:Z, Zlt_bool n m = false -> (n >= m)%Z.
Proof.
  unfold Zlt_bool, Zge in |- *. intros x y. unfold not in |- *.
  case (x ?= y)%Z; (trivial || (intros; discriminate)).
Qed.

Lemma Zeq_eq_bool: forall n m:Z, (n = m)%Z -> Zeq_bool n m = true.
Proof.
  unfold Zeq_bool.
  intros; subst; rewrite Zcompare_refl; auto.
Qed.

Lemma Zdiff_eq_bool: forall n m:Z, (n <> m)%Z -> Zeq_bool n m = false.
Proof.
  unfold Zeq_bool.
  intros x y; generalize (Zcompare_Eq_iff_eq x y); case (x ?= y)%Z;
  (trivial || intuition auto).
Qed.


Lemma Zeq_cases :
 forall n m:Z, if Zeq_bool n m then (n = m)%Z else (n <> m)%Z.
Proof.
intros x y; unfold Zeq_bool; generalize (Zcompare_Eq_iff_eq x y).
case (x ?= y)%Z; (intuition auto || discriminate).
Qed.

Ltac zlt_case n m := generalize (Zlt_cases n m);  case (Zlt_bool n m); simpl.
Ltac zeq_case n m := generalize (Zeq_cases n m);  case (Zeq_bool n m); simpl.
