Set Implicit Arguments.

Require Export ZBoolAux.

Open Scope Z_scope.

Hint Resolve Z_mod_lt Z_mod_zero_opp: zarith.

Lemma Z_div_zero: forall (a b: Z), b > 0 -> 0 <= a < b -> a/b=0.
Proof.
   intros; generalize (Z_div_ge0 a b); intuition.
   zeq_case (a/b) 0; auto.
   intros; absurd (b <= a).
   intuition.
   generalize (Z_div_mod_eq a b).
  generalize (Z_mod_lt a b).
  intuition.
  absurd (b<=a).
  omega.
  cutrewrite (b=b*1+0).
  rewrite H6.
  auto with zarith.
  ring.
Qed.

Lemma Z_mod_unicity: forall (a b: Z), b > 0 -> 0 <= a < b -> a=(a mod b).
Proof.
  intros; cutrewrite (a mod b = b*(a/b) + a mod b).
  generalize (Z_div_mod_eq a b); intuition.
  rewrite Z_div_zero; auto.
  ring.
Qed.

Lemma Z_div_unicity: forall (a b c: Z), b > 0 -> 0 <= c < b -> a=(b*a+c)/b.
Proof.
  intros; cutrewrite (b * a + c = c + a * b).
  rewrite Z_div_plus; auto.
  rewrite Z_div_zero; auto.
  ring.
Qed.

Lemma Z_mod_mod_id: forall (a b: Z), b > 0 -> (a mod b) mod b = a mod b.
Proof.
  intros; rewrite <- Z_mod_unicity; auto with zarith.
Qed.

Hint Resolve Z_mod_mod_id: zarith.

Lemma Z_mod_plus_mod: forall (a b c: Z), c > 0 -> (a + b) mod c = ((a mod c) + (b mod c)) mod c.
Proof.
  intros a b c H; cutrewrite ((a mod c) + (b mod c) = (a + b) + -(a/c+b/c)*c).
  rewrite Z_mod_plus; auto.
  generalize (Z_div_mod_eq a c);
  generalize (Z_div_mod_eq b c);
  intuition.
  cutrewrite (a mod c=a-c*(a/c)).
  cutrewrite (b mod c=b-c*(b/c)).
  ring.
  omega.
  omega.
Qed.


Lemma Z_mod_minus_mod: forall (a b c: Z), c > 0 -> (a - b) mod c = ((a mod c) - (b mod c)) mod c.
Proof.
  intros a b c H;
  cutrewrite ((a mod c) - (b mod c) = (a mod c) + - (b mod c)); try ring.
  rewrite Z_mod_plus_mod; auto.
  rewrite Z_mod_mod_id; auto.
  rewrite <- Z_mod_plus_mod; auto.
  cutrewrite (a + - (b mod c) = (a - b) + (b/c)*c).
  rewrite Z_mod_plus; auto.
  cutrewrite (b mod c = b - c*(b/c)).
  ring.
  generalize (Z_div_mod_eq b c); omega.
Qed.

Lemma Z_mod_eq_intro_aux: forall (a b c: Z), c > 0 -> a mod c > b mod c -> (a - b) mod c <> 0.
Proof.
  intros a b c H; rewrite Z_mod_minus_mod; auto.
  intros H1; rewrite <- Z_mod_unicity; try omega.
  intuition; try omega.
  generalize (Z_mod_lt a c); auto with zarith.
  generalize (Z_mod_lt b c); auto with zarith.
Qed.

Lemma Z_mod_eq_intro: forall (a b c: Z), c > 0 -> (a - b) mod c = 0 -> a mod c = b mod c.
Proof.
  intros a b c H.
  zlt_case (a mod c) (b mod c); auto.
  intros H1 H2; case (Z_mod_eq_intro_aux b a (c:=c)); try omega.
  cutrewrite (b - a = - (a - b)); auto with zarith.
  zeq_case (a mod c) (b mod c); auto.
  intros; case (Z_mod_eq_intro_aux a b (c:=c)); try omega.
Qed.

Lemma Z_mod_le: forall (a b: Z), a >= 0 -> b > 0 -> (a mod b) <= a.
Proof.
  intros ;
  cutrewrite (a mod b = a - b * (a / b)).
  generalize (Z_div_ge0 a b).
  intuition.
  assert (b*(a/b) >= 0); auto with zarith.
  cutrewrite (0 = 0 * 0); auto with zarith.
  generalize (Z_div_mod_eq a b); omega.
Qed.
